/**
 *  Class for showing coordinates in and outside of a centered image in a coordinate system.
 *
 *  Class:       CoordinateImagePanel - Work in Progress Implementation
 *  Task:        Only Display an Image, and Coordinates of Endpoints of Lines
 *  Docs:	     Only here in the Source Code
 *  Author:      Frankie
 *  Copyright:   2013, Frank
 *  Start:       October 2013
 *  Last Change: October 2013
 *  Version:     0.2 - if this works? Yeah it will! Now it does.
 */

import java.awt.*;
import java.awt.event.*;
import java.awt.image.*;
import java.awt.geom.*;
import java.awt.geom.AffineTransform;
import java.io.*;
import javax.swing.*;
import java.util.Locale;
import javax.imageio.ImageIO;
import javax.imageio.*;
import javax.imageio.plugins.jpeg.JPEGImageWriteParam;
import javax.imageio.stream.ImageOutputStream;
import java.awt.event.MouseListener;
import java.awt.event.MouseMotionListener;
import java.awt.event.MouseEvent;

public class CoordinateImagePanel extends JPanel implements MouseListener, MouseMotionListener {
	boolean 		debugMode 		= false;	// Debug output on Console if true
	String 			theImagePath	= "";
	CoordinateFrame papaFrame;

	int originalWidth	= 1920;	// will be overwritten in loadImage method
	int originalHeight	= 1080; // others will be recalculated then
	double scaleFactor	= 0.3;
	double areaFactor	= 0.7;
	int imageWidth		= (int)((double)originalWidth*scaleFactor);		// Image Size
	int imageHeight		= (int)((double)originalHeight*scaleFactor);	// Image Size
	int prevWidth		= (int)((double)originalWidth*areaFactor);		// Drawing Area Size
	int prevHeight		= (int)((double)originalHeight*areaFactor);		// Drawing Area Size
	int translX 		= (int)((prevWidth)/2);		// Translation Value for centric coordinate system in drawing area
	int translY 		= (int)((prevHeight)/2);	// Translation Value for centric coordinate system in drawing area

	BufferedImage	imageBuffer;	// The preview Image from stills folder

	int		handleSize				= 2;	// Endpoint handles of the lines, and sensitive area for mouseclicks
	int		hotArea					= 5;

	// Lines - coordinates are NOT in scaled and translated style!
	int		line1X1 	= prevWidth/4;	// relative coordinates, so it always shows up
	int		line1Y1 	= prevHeight/4;
	int		line1X2 	= prevWidth/4*3-10;
	int		line1Y2		= prevHeight/4;

	int		line2X1 	= prevWidth/4;
	int		line2Y1 	= prevHeight/4*3;
	int		line2X2 	= prevWidth/4*3-10;
	int		line2Y2		= prevHeight/4*3;

	int		line3X1 	= line2X1;
	int		line3Y1 	= line2Y1;
	int		line3X2 	= prevWidth/4*3;
	int		line3Y2		= (int)((double)prevHeight/4*3.5);

	// Mouse Control, Area Detection
	int usedHandle = 0;
	int startX = 0;
	int startY = 0;
	int lastX = 0;
	int lastY = 0;
	final static int NO_HANDLE			= 0;
	final static int LINE1_HANDLE1		= 1;
	final static int LINE1_HANDLE2		= 2;
	final static int LINE2_HANDLE1		= 3;
	final static int LINE2_HANDLE2		= 4;
	final static int LINE3_HANDLE1		= 5;
	final static int LINE3_HANDLE2		= 6;

	boolean drawCoordinates = true;	// coordinates and cross in panel area?

	public CoordinateImagePanel(String pathToImage, CoordinateFrame papaFrame) {						// *** Constructor 1
		super();
		this.papaFrame = papaFrame;	// needed to call setter methods for coordinate display in labels
		theImagePath = pathToImage;

		this.setPreferredSize(new Dimension(prevWidth,prevHeight));
		this.setSize(prevWidth, prevHeight);
		this.setMinimumSize(new Dimension(prevWidth, prevHeight));
		this.setMaximumSize(new Dimension(prevWidth, prevHeight));

		loadOriginalImage();

		this.addMouseListener(this);
		this.addMouseMotionListener(this);

	} // Constructor end

	// *** overwritten function for displaying the scaled image and other stuff
	public void paintComponent(Graphics g) {	// *** Overwritten Paint methode
		super.paintComponent(g);
		Graphics2D g2 = (Graphics2D)g;
		g2.translate(translX, translY);	// Translate for centered coordinate System

		// **** Draw the loaded Image
		if(imageBuffer != null) {
			g2.drawImage(imageBuffer, (-1*imageWidth/2), (-1*imageHeight/2), imageWidth, imageHeight, null);
		}

		// **** Set general font for strings in the image
		int fSize = 10;
        g2.setFont(new Font("SansSerif", Font.BOLD, fSize));

		// *** Draw the coordinate system
		if(drawCoordinates) {
			g2.setColor(Color.black);
			g2.drawLine(0, prevHeight/2, 0, prevHeight/2*-1);
			g2.drawLine(prevWidth/2*-1, 0, prevWidth/2, 0);
			g2.setColor(new Color(150,150,150));
			g2.drawLine(prevWidth/2*-1, imageHeight/2, prevWidth/2, imageHeight/2);
			g2.drawLine(prevWidth/2*-1, imageHeight/2*-1, prevWidth/2, imageHeight/2*-1);
			g2.drawLine(imageWidth/2*-1, prevHeight/2, imageWidth/2*-1, prevHeight/2*-1);
			g2.drawLine(imageWidth/2, prevHeight/2, imageWidth/2, prevHeight/2*-1);
			// position values
			g2.drawString(""+(int)((1/scaleFactor)*prevHeight/2*-1), fSize, prevHeight/2-fSize);
			g2.drawString(""+(int)((1/scaleFactor)*prevHeight/2), 	fSize, (prevHeight/2*-1)+fSize);
			g2.drawString(""+(int)((1/scaleFactor)*prevWidth/2*-1),	prevWidth/2*-1+fSize, fSize);
			g2.drawString(""+(int)((1/scaleFactor)*prevWidth/2), 	prevWidth/2-(fSize*3), fSize);

			g2.drawString("0,0", fSize/2, fSize);
			g2.drawString(""+(int)((1/scaleFactor)*imageWidth/2*-1),	imageWidth/2*-1-(3*fSize), fSize);
			g2.drawString(""+(int)((1/scaleFactor)*imageWidth/2),	imageWidth/2+fSize, fSize);

			g2.drawString(""+(int)((1/scaleFactor)*imageHeight/2), fSize, (imageHeight/2*-1)-fSize);
			g2.drawString(""+(int)((1/scaleFactor)*imageHeight/2*-1), fSize, (imageHeight/2)+(2*fSize));
		}

		g2.translate(-translX, -translY);	// Translate for centered coordinate System

		// *** Draw the three lines
		g2.setColor(Color.red);
		g2.drawLine(line1X1, line1Y1, line1X2, line1Y2);
		g2.drawRect(line1X1-handleSize, line1Y1-handleSize, handleSize*2, handleSize*2);
		g2.drawRect(line1X2-handleSize, line1Y2-handleSize, handleSize*2, handleSize*2);
		g2.drawString("L1-2",line1X2, line1Y2+15);
		g2.drawString("L1-1",line1X1, line1Y1+15);

		g2.setColor(Color.blue);
		g2.drawLine(line2X1, line2Y1, line2X2, line2Y2);
		g2.drawRect(line2X1-handleSize, line2Y1-handleSize, handleSize*2, handleSize*2);
		g2.drawRect(line2X2-handleSize, line2Y2-handleSize, handleSize*2, handleSize*2);
		g2.drawString("L2-2",line2X2,line2Y2+15);
		g2.drawString("L2-1",line2X1,line2Y1+15);

		g2.setColor(Color.green);
		g2.drawLine(line3X1, line3Y1, line3X2, line3Y2);
		g2.drawRect(line3X1-handleSize, line3Y1-handleSize, handleSize*2, handleSize*2);
		g2.drawRect(line3X2-handleSize, line3Y2-handleSize, handleSize*2, handleSize*2);
		g2.drawString("L3-2",line3X2,line3Y2+15);
		g2.drawString("L3-1",line3X1,line3Y1+15);
	}

	public void zoomOneIn() {
		scaleFactor	= scaleFactor + 0.1;

		loadOriginalImage();
		this.repaint();
	}
	public void zoomOneOut() {
		if((scaleFactor - 0.1)>0.0) {	// must not be 0 or lower
			scaleFactor	= scaleFactor - 0.1;
		}
		loadOriginalImage();
		this.repaint();
	}

	public void recalcDimensions() {	// recalc when reloaded image
		imageWidth		= (int)((double)originalWidth*scaleFactor);		// Image Size
		imageHeight		= (int)((double)originalHeight*scaleFactor);	// Image Size
		prevWidth		= (int)((double)originalWidth*areaFactor);		// Drawing Area Size
		prevHeight		= (int)((double)originalHeight*areaFactor);		// Drawing Area Size
		translX 		= (int)((prevWidth)/2);		// Translation Value for centric coordinate system in drawing area
		translY 		= (int)((prevHeight)/2);	// Translation Value for centric coordinate system in drawing area
	}

	public void loadOriginalImage() {
		BufferedImage tmpBuffer;
		File imageFile;
		imageFile		= new File(theImagePath);
		if(imageFile.exists()) {	// if this exists ... use it
			try {
				// *** load image
				imageBuffer			= ImageIO.read(imageFile);			// load original image to Buffer
				int originalWidth 	= imageBuffer.getWidth();			// Get original size informations
				int originalHeight 	= imageBuffer.getHeight();
				recalcDimensions();

				// *** scale image
				Object hint = RenderingHints.VALUE_INTERPOLATION_BILINEAR;
				int type = (imageBuffer.getTransparency() == Transparency.OPAQUE) ? BufferedImage.TYPE_INT_RGB : BufferedImage.TYPE_INT_ARGB;
				int w,h;
				w = originalWidth;		// use multi-step scale
				h = originalHeight;
				int scaleDirection = 0;
				if(imageWidth>w) scaleDirection = 1;
				do {
					if(scaleDirection == 0) {
						w >>= 1;
						h >>= 1;

						if(h < imageHeight || w < imageWidth) {
							w = imageWidth;
							h = imageHeight;
						}
					}
					if(scaleDirection == 1) {
						w = imageWidth;
						h = imageHeight;
					}

					tmpBuffer = new BufferedImage(w,h, type);
					Graphics2D g2 = tmpBuffer.createGraphics();
					g2.setRenderingHint(RenderingHints.KEY_INTERPOLATION, hint);
					g2.drawImage(imageBuffer,0,0,w,h,null);
					g2.dispose();
					imageBuffer = tmpBuffer;
				} while (w != imageWidth || h != imageHeight);
				// *** scale image end
				// ** free ressources
				tmpBuffer.flush();
				tmpBuffer = null;
				prevWidth	= (int)((double)originalWidth*areaFactor);		// Drawing Area Size
				prevHeight	= (int)((double)originalHeight*areaFactor);		// Drawing Area Size
			} catch( IOException ex ) {
				System.out.println("ERROR LOADING IMAGE IN CAss!!!!");
			}
		}
	}

	// MOUSE FUNCTIONS ************************************************************************
	// *** Check if given coordinates are at line handles
	public int getMouseClickedArea(int theMouseX, int theMouseY) {
		int clickArea = NO_HANDLE;

		// **************** LINE HANDLES
		// Line 1 handels
		if(theMouseX > line1X1 - hotArea && theMouseX < line1X1 + hotArea) {
			if(theMouseY > line1Y1 - hotArea && theMouseY < line1Y1 + hotArea) {
				clickArea = LINE1_HANDLE1;
			}
		}
		if(theMouseX > line1X2 - hotArea && theMouseX < line1X2 + hotArea) {
			if(theMouseY > line1Y2 - hotArea && theMouseY < line1Y2 + hotArea) {
				clickArea = LINE1_HANDLE2;
			}
		}

		// Line 2 handels
		if(theMouseX > line2X1 - hotArea && theMouseX < line2X1 + hotArea) {
			if(theMouseY > line2Y1 - hotArea && theMouseY < line2Y1 + hotArea) {
				clickArea = LINE2_HANDLE1;
			}
		}
		if(theMouseX > line2X2 - hotArea && theMouseX < line2X2 + hotArea) {
			if(theMouseY > line2Y2 - hotArea && theMouseY < line2Y2 + hotArea) {
				clickArea = LINE2_HANDLE2;
			}
		}

		// Line 3 handels
		if(theMouseX > line3X1 - hotArea && theMouseX < line3X1 + hotArea) {
			if(theMouseY > line3Y1 - hotArea && theMouseY < line3Y1 + hotArea) {
				clickArea = LINE3_HANDLE1;
			}
		}
		if(theMouseX > line3X2 - hotArea && theMouseX < line3X2 + hotArea) {
			if(theMouseY > line3Y2 - hotArea && theMouseY < line3Y2 + hotArea) {
				clickArea = LINE3_HANDLE2;
			}
		}

		// Debug Mouse and Return
		if(debugMode) {
			System.out.println("Mouse X: " + theMouseX);
			System.out.println("Mouse Y: " + theMouseY);
			System.out.println("L1 x: "+(int)(scaleFactor*line1X1)+" - L1 y: "+(int)(scaleFactor*line1Y1));
			System.out.println("Clicked Area:" + clickArea);
		}
		return clickArea;
	}

	// left mouse
	public void mouseClicked(MouseEvent e) {;}

	// unused but for listener
  	public void mouseExited(MouseEvent e) {;}
	public void mouseEntered(MouseEvent e) {;}

	public void mousePressed(MouseEvent e) {
		startX = e.getX();
		startY = e.getY();
		lastX = startX;
		lastY = startY;
		usedHandle = getMouseClickedArea(startX, startY);
	}

	public void mouseReleased(MouseEvent e) {
		if(debugMode) {
			System.out.println("Mouse Released at "+e.getX()+"  "+e.getY());
		}

		Graphics g = getGraphics();
		g.setXORMode(Color.lightGray);

		if(usedHandle == LINE1_HANDLE1) {
			g.drawLine(lastX, lastY, line1X2, line1Y2);
			line1X1 = lastX;
			line1Y1 = lastY;
			repaint();
			usedHandle = NO_HANDLE;
		} else if(usedHandle == LINE1_HANDLE2) {
			g.drawLine(line1X1, line1Y1, lastX, lastY);
			line1X2 = lastX;
			line1Y2 = lastY;
			repaint();
			usedHandle = NO_HANDLE;
		} else if(usedHandle == LINE2_HANDLE1) {
			g.drawLine(lastX, lastY, line2X2, line2Y2);	// do two handles since they are connected
			line2X1 = lastX;
			line2Y1 = lastY;
			g.drawLine(lastX, lastY, line3X2, line3Y2);
			line3X1 = lastX;
			line3Y1 = lastY;
			repaint();
			usedHandle = NO_HANDLE;
		} else if(usedHandle == LINE2_HANDLE2) {
			g.drawLine(line2X1, line2Y1, lastX, lastY);
			line2X2 = lastX;
			line2Y2 = lastY;
			repaint();
			usedHandle = NO_HANDLE;
		} else if(usedHandle == LINE3_HANDLE1) {
			g.drawLine(lastX, lastY, line3X2, line3Y2);	// do two handles since they are connected
			line3X1 = lastX;
			line3Y1 = lastY;
			g.drawLine(lastX, lastY, line2X2, line2Y2);
			line2X1 = lastX;
			line2Y1 = lastY;
			repaint();
			usedHandle = NO_HANDLE;
		} else if(usedHandle == LINE3_HANDLE2) {
			g.drawLine(line3X1, line3Y1, lastX, lastY);
			line3X2 = lastX;
			line3Y2 = lastY;
			repaint();
			usedHandle = NO_HANDLE;
		}

		updateCoordinateValuesExtern();
	}

	// MouseMotionListener
	public void mouseMoved(MouseEvent e) {;}
	public void mouseDragged(MouseEvent e) {
		int theX = e.getX();
		int theY = e.getY();
		//int theDeltaX = -(lastX-theX);
		//int theDeltaY = -(lastY-theY);

		//int theWidth = this.getWidth();
		//int theHeight = prevHeight;

		Graphics g = getGraphics();
		g.setXORMode(Color.lightGray);

		if(usedHandle == LINE1_HANDLE1) {
			g.drawLine(lastX, lastY, line1X2, line1Y2);
			g.drawLine(theX, theY, line1X2, line1Y2);
			lastX = theX;
			lastY = theY;
		} else if(usedHandle == LINE1_HANDLE2) {
			g.drawLine(line1X1, line1Y1, lastX, lastY);
			g.drawLine(line1X1, line1Y1, theX, theY);
			lastX = theX;
			lastY = theY;
		} else if(usedHandle == LINE2_HANDLE1) {
			g.drawLine(lastX, lastY, line2X2, line2Y2);	// do two handles since they are connected
			g.drawLine(theX, theY, line2X2, line2Y2);
			g.drawLine(lastX, lastY, line3X2, line3Y2);
			g.drawLine(theX, theY, line3X2, line3Y2);
			lastX = theX;
			lastY = theY;
		} else if(usedHandle == LINE2_HANDLE2) {
			g.drawLine(line2X1, line2Y1, lastX, lastY);
			g.drawLine(line2X1, line2Y1, theX, theY);
			lastX = theX;
			lastY = theY;
		} else if(usedHandle == LINE3_HANDLE1) {
			g.drawLine(lastX, lastY, line2X2, line2Y2);	// do two handles since they are connected
			g.drawLine(theX, theY, line2X2, line2Y2);
			g.drawLine(lastX, lastY, line3X2, line3Y2);
			g.drawLine(theX, theY, line3X2, line3Y2);
			lastX = theX;
			lastY = theY;
		} else if(usedHandle == LINE3_HANDLE2) {
			g.drawLine(line3X1, line3Y1, lastX, lastY);
			g.drawLine(line3X1, line3Y1, theX, theY);
			lastX = theX;
			lastY = theY;
		}
	}

	// ****** convert normal screen mouse coordinates to scaled and translated screen coordinates
	public int pointToScreenX(int mX) {
		int sX = (int)((1/scaleFactor) * (double)(mX - (prevWidth/2)) );
		return sX;
	}
	public int pointToScreenY(int mY) {
		int sY = -1*(int)((1/scaleFactor) * (double)(mY - (prevHeight/2)) );
		return sY;
	}
	public void toggleDrawCoordinates() {
		if(drawCoordinates) {
			drawCoordinates = false;
		} else {
			drawCoordinates = true;
		}
	}
	// **** update coordinate values in papa frames labels for coordinates
	public void updateCoordinateValuesExtern() {
		papaFrame.setLine1X1(pointToScreenX(line1X1));
		papaFrame.setLine2X1(pointToScreenX(line2X1));
		papaFrame.setLine3X1(pointToScreenX(line3X1));
		papaFrame.setLine1Y1(pointToScreenY(line1Y1));
		papaFrame.setLine2Y1(pointToScreenY(line2Y1));
		papaFrame.setLine3Y1(pointToScreenY(line3Y1));
		papaFrame.setLine1X2(pointToScreenX(line1X2));
		papaFrame.setLine2X2(pointToScreenX(line2X2));
		papaFrame.setLine3X2(pointToScreenX(line3X2));
		papaFrame.setLine1Y2(pointToScreenY(line1Y2));
		papaFrame.setLine2Y2(pointToScreenY(line2Y2));
		papaFrame.setLine3Y2(pointToScreenY(line3Y2));
	}
}