//  Class: TimeReloadedFrame
// started: 21.08.2013 (ca.)
// Version 0.8 development from 23.01.2014
// contact frankie@blindcat.de

import java.awt.*;
import java.awt.image.*;
import java.awt.event.*;
import javax.swing.*;
import javax.swing.event.*;
//import javax.swing.ImageIcon;
import javax.swing.border.*;

import java.util.Timer;		// Timer, Arrays etc.
import java.util.TimerTask;
import java.util.*;
import java.lang.Math.*;

import java.io.*;
import javax.imageio.ImageIO;
import javax.imageio.*;

public class TimeReloadedFrame extends JFrame {

	// *** GUI Components
	JPanel				previewPanel, timeLinePanel;
	JLabel				labStatus;
	JButton				buttPreview,	buttCreateMask;
	JButton				buttOriginal,	buttExit;
	JButton				buttPause, 		buttBack;
	JButton				buttInfo,		buttCass;
	JPanel				timeDimPanel;
	GridBagConstraints	gbConstraints;

	// *** Global User Parameter *************************** ALWAYS SET THEM RIGHT FOR PROJECT
	int 	layerCount		= 19;		// layerCount starts at 0 during code useage
	boolean lineAlias		= true;		// should be set to true. Can be changed if wanted
	boolean	motionBlur		= false;	// Experimental. Not implemented yet
	int		originalWidth	= 1920;		// Set the width of the original Videomaterial in Pixel
	int		originalHeight	= 1080;		// Set the height of the original Videomaterial in Pixel
	double	scaleFactor		= 0.5;		// Scale Preview to fit on Monitor. 0.5 is good
	double	framesPerSecond	= 25.0;		// Frames per second for fps of mask sequence output
	int		beatsPerMinute	= 140;		// Enter beats per minute to align movement to music
	int		projectFrames	= (int)(180.0 * framesPerSecond); // Enter Project Length in seconds
	private static int DELAY = 40;	// **** MUST BE CHANGED TO 1000/fps (e.g. 1000/25 = 40)

	String		imageFileName		= "MVI_0682.jpg";					// New mechanism. Only write the name of the file, not the path
	String		absoluteImagePath 	= "";
	//String	maskFileFormat		= "BMP";	// Uncomment to use this format
	String		maskFileFormat		= "jpg";	// Uncomment to use this format
	String		originalMovieEnding	= ".MOV";
	//String	originalMovieEnding	= ".avi";	// Uncomment for automatic file detection in mov folder if avi

	// *** running Variables/Console Feedback
	boolean showTimingInfo			= false;	// Shows timing Info for calc on the console in the method printTiming()
	boolean showMaskCreateFeedback	= true;	// show console feedback during mask creation (because its a long process)

	// *** Global Programmatical Parameter ***************** Never change them
	// *** calc Preview Size
	double prevDWidth	= (double)originalWidth * scaleFactor;	// Value for preview Size
	double prevDHeight	= (double)originalHeight * scaleFactor;	// Value for preview Size
	int prevWidth		= (int)prevDWidth;						// Value for preview Size
	int prevHeight		= (int)prevDHeight;						// Value for preview Size
	int prevTimeHeight	= 90;									// Height of Timeline (width is prevWidth)
	int translX = (int)((originalWidth*scaleFactor)/2);			// Translation Value for centric coordinate system
	int translY = (int)((originalHeight*scaleFactor)/2);		// Translation Value for centric coordinate system

	int maskTranslX = (int)((originalWidth)/2);					// Translation Value for centric coordinate system
	int maskTranslY = (int)((originalHeight)/2);				// Translation Value for centric coordinate system

	// *** Image Buffers
	BufferedImage oddImage;	// odd(ungerade)
	BufferedImage evenImage;
	BufferedImage maskBuffer;
	Graphics2D mask2D;
	Graphics2D oddG2;
	Graphics2D evenG2;
	int 	currBuffer = 0;		// Main counter for playback
	boolean infoOn = true;		// Show psition data in image
	int 	timinginfoX = 5;	// where to paint timing info into the image
	int 	timinginfoY = ((int)((double)originalHeight*scaleFactor))-5;
	int currentFrame = 0;	    // used for calculation progress

	// *** Color Palette for Preview Purposes
	Color[] greys;
	Color colBlack;
	Color colWhite;
	Color colAlias;
	Color colYellow;

	// *** some constants
	double framesPerMinute = framesPerSecond * 60;
	double framesPerBeat = framesPerMinute / (double)beatsPerMinute;

	// *** file access data for orginal files and where to store the masks
	String fullWorkingPath = "";	// This is the path of TimeReloaded -"trmc/TimeReloaded.class" with trailing /
	String movFolder = "mov"+File.separator;		// the folder where the original movs are stored
	String maskFolder = "masks"+File.separator;	// the folder where the masks folders will be stored
	String maskSubs = "-Mask-Lay";	// Substring for Mask Foldernames and Filenames
	String[]	movieNames;			// Will be filled with File.list() function for all MOV files.

	// *** Sequencer Data Arrays and control parameter
	int[]		shapeType	= new int[projectFrames];		// 0 not set. 1 and up different shape renderer
	double[] 	kx1 		= new double[projectFrames];
	double[] 	ky1 		= new double[projectFrames];
	double[] 	kx2 		= new double[projectFrames];
	double[] 	ky2 		= new double[projectFrames];
	Double[] 	wide1 		= new Double[projectFrames];
	Double[] 	wide2 		= new Double[projectFrames];
	Double[] 	wide3 		= new Double[projectFrames];
	Double[] 	wide4 		= new Double[projectFrames];
	Double[] 	pulse 		= new Double[projectFrames];	// pulse multiplikator if is
	boolean[] 	reverse 	= new boolean[projectFrames];	// should mask be painted reverse
	boolean[]	dominance	= new boolean[projectFrames];	// Is one Layer more dominant than the others?
	int[]		domLayerNr	= new int[projectFrames];		// Which layer is it if?
	Double[]	domStrength	= new Double[projectFrames];	// How much more dominant it is 0 to 1
	Double[]	domFalloff	= new Double[projectFrames];	// are all other layers same size or is fall off from dominant to normal? 0 to 1
	int[]		directionX	= new int[projectFrames];		// Mirror Direction of Effect on X Axis 1=normal, -1 = mirrored
	int[]		directionY	= new int[projectFrames];		// Mirror Direction of Effect on Y Axis 1=normal, -1 = mirrored
	boolean[] 	mode 		= new boolean[projectFrames];	// if the renderer has a mode e.g. straightGrow

	Timer timer;
	int currFrame = 0;	// Main frame counter for sequencer
															// *** END Variables *************************
	public TimeReloadedFrame() {									// *** Constructor Start - Create the Frame
		super("Time Reloaded Mask Creator");	// FrameTitel
		setupFileData();

		setupComponentInstances();
		setupListeningComponents();
		setupComponentInitContent();
		setupComponentStyles();

		createImageBuffers();
		createColors();

		calcControl(); 			//***  calculate all animations here!

		makeLayout();

		getContentPane().add(timeDimPanel);

		this.enableEvents(AWTEvent.WINDOW_EVENT_MASK);		// Close Win
		setAbsoluteImagePath();
	}														// *** End Constructor

	// ********************* Window Control Methods *****************
	protected void processWindowEvent(WindowEvent e) {		// Window Events
		if (e.getID()==WindowEvent.WINDOW_CLOSING) {
			quitProgramm();
		}
	}

	public void centerWindow() {							// Method to center window
		Dimension screenSize = Toolkit.getDefaultToolkit ().getScreenSize();
		Dimension frameSize = this.getSize();
		if (frameSize.height > screenSize.height) {
			frameSize.height = screenSize.height;
		}
		if (frameSize.width > screenSize.width) {
			frameSize.width = screenSize.width;
		}
		this.setLocation((screenSize.width - frameSize.width) / 2,
						(screenSize.height - frameSize.height) / 2);
	}

	public void quitProgramm() {	// call to quit
		if(timer != null) {
			timer.cancel();			// Stop timer if its still running
			timer.purge();
		}
		dispose();					// free ressources
		System.exit(0);
	}
	// *** helper Method startUp
	public void setAbsoluteImagePath() {
		File imageFile;
		imageFile		= new File(System.getProperty("user.dir")+File.separator+".."+File.separator+"stills"+File.separator+movieNames[0]+".jpg");
		if(imageFile.exists()) {	// if this exists ... use it
			absoluteImagePath = System.getProperty("user.dir")+File.separator+".."+File.separator+"stills"+File.separator+movieNames[0]+".jpg";
		} else {
			absoluteImagePath = System.getProperty("user.dir")+File.separator+imageFileName;
		}
	}

	// *** Programm dependent code start
	// *** Create Colors for use in the Software
	public void createColors() {
		int colorStep = 255 / layerCount;
		for(int i = 0;i<layerCount;i++) {
			int theCol = colorStep + i * colorStep;
			greys[i] = new Color(theCol, theCol, theCol);
		}
		colBlack = new Color(0,0,0);
		colWhite = new Color(255,255,255);
		colAlias = new Color(127,127,127);
		colYellow = new Color(249,231,33);
	}

	// *** Buffer Management
	public void createImageBuffers() {		// create two buffers for animation painting and set aliasing on/off
		oddImage = new BufferedImage(prevWidth, prevHeight, BufferedImage.TYPE_INT_ARGB);
		evenImage = new BufferedImage(prevWidth, prevHeight, BufferedImage.TYPE_INT_ARGB);
		oddG2 = oddImage.createGraphics();
		evenG2 = evenImage.createGraphics();
		if(lineAlias) {
			oddG2.setRenderingHint( RenderingHints.KEY_TEXT_ANTIALIASING, RenderingHints.VALUE_TEXT_ANTIALIAS_ON);
			oddG2.setRenderingHint(RenderingHints.KEY_ANTIALIASING,RenderingHints.VALUE_ANTIALIAS_ON);
			evenG2.setRenderingHint( RenderingHints.KEY_TEXT_ANTIALIASING, RenderingHints.VALUE_TEXT_ANTIALIAS_ON);
			evenG2.setRenderingHint(RenderingHints.KEY_ANTIALIASING,RenderingHints.VALUE_ANTIALIAS_ON);
		}
	}

	public Graphics2D getTheGraphics() {	// get Graphics for current painting buffer
		if(currBuffer == 0) {
			return evenG2;
		} else {
			return oddG2;
		}
	}
	public BufferedImage getCurrentDisplayBuffer() {
		if(currBuffer == 1) {
			return evenImage;
		} else {
			return oddImage;
		}
	}

	public Graphics2D getCurrentDisplayGraphics() {	// get Graphics for current painting buffer
		if(currBuffer == 1) {
			return evenG2;
		} else {
			return oddG2;
		}
	}

	public void switchBuffer() {			// switch buffer for painting
		if(currBuffer == 0) {
			paintABuffer(evenImage);
			currBuffer = 1;
		} else {
			paintABuffer(oddImage);
			currBuffer = 0;
		}
	}

	public void paintABuffer(BufferedImage theImage) {	// paint the current buffer to the previewPanel
		Graphics2D panelGraphics = (Graphics2D)previewPanel.getGraphics();
		panelGraphics.drawImage(theImage, 0, 0, prevWidth, prevHeight, null);
	}

	// **** Timer Thread for playing the Preview/Sequence Player
	public void playPreview() {
		paintTimeline(); // can't paint it at startup. dont know why. beside it needs a repaint sometimes
		TimerTask task = new TimerTask() {
			public void run() {
				renderControl(currFrame);
				currFrame++;
			}
		};
		timer = new Timer();
		timer.schedule(task, 0, DELAY);
	}

	public void stopPreview() {
		timer.cancel();
		timer.purge();
		buttPreview.setEnabled(true);
		buttPause.setEnabled(false);
		doButtPauseAction();
		resetPreview();
	}

	public void resetPreview() {
		// do reset
		currFrame = 0;
		updateTimeline();
		// Paint it black
		Graphics2D g2 = getTheGraphics();
		g2.setColor(colBlack);
		g2.fillRect(0, 0, prevWidth, prevHeight);
		switchBuffer();
	}

	// **** Render the single Frames. Control and do painting methods
	public void renderControl(int frameNr) {
		boolean stopit = false;
		if(frameNr == projectFrames) {
			if(false) {
				currFrame = 0;
			} else {
				stopit = true;
				stopPreview();
			}
		}
		if(!stopit) {
			if(shapeType[frameNr]==1) {	// if shape type = 1 use poly rect renderer
				shapeRendererPolyRect(frameNr);			// General Poly Rect
			} else if(shapeType[frameNr]==2) {
				fadeRendererHorizontal(frameNr);		// fade Horizontal renderer
			} else if(shapeType[frameNr]==3) {
				shapeRendererWaitInLayer(frameNr);		// Wait in Layer renderer
			} else if(shapeType[frameNr]==4) {
				shapeRendererCircleInOut(frameNr);		// Simple Circle In Out renderer
			} else if(shapeType[frameNr]==5) {
				shapeRendererStartAnimation(frameNr);	// Start Animation Real Time renderer
			} else if(shapeType[frameNr]==6) {
				shapeRendererWaitEmpty(frameNr);		// Wait in Empty Layer renderer
			} else if(shapeType[frameNr]==7) {
				shapeRendererAudioSync(frameNr);		// AudioSync Renderer
			} else if(shapeType[frameNr]==8) {
				shapeRendererCircleInOut(frameNr);		// Simple Circle In Out renderer can render Grow Circle
			} else if(shapeType[frameNr]==9) {
				shapeRendererDualCircle(frameNr);		// Dual growing Circle In Out renderer with move
			} else if(shapeType[frameNr]==10) {
				shapeRendererLayerSpot(frameNr);		// Simple Circle In Out renderer can move on one Layer as a spot
			} else if(shapeType[frameNr]==11) {
				shapeRendererPerspectiveRays(frameNr);	// Perspective Rays Renderer
			} else if(shapeType[frameNr]==12) {
				shapeRendererQuadBox(frameNr);			// Creates a quadratic box renderer
			} else if(shapeType[frameNr]==13) {
				shapeRendererRectBox(frameNr);			// Creates a rectangular box renderer
			} else if(shapeType[frameNr]==14) {
				shapeRendererHorizontThreeStepFader(frameNr);			// Three Step Horizontal Fade
			} else if(shapeType[frameNr]==15) {
				shapeRendererCircleRose(frameNr);		// creates rays that open like a rose in circular geometry
			} else if(shapeType[frameNr]==99) {
				shapeRendererSimpleRect(frameNr);		// Simple Rect Renderer
			} else { // no shape calculated for this frame
				shapeRenderEmpty(frameNr);				// Empty Frame Renderer
			}
			updateTimeline();
		}
	}

	public void shapeRendererCircleRose(int frameNr) {
		// *** shapeType = 15 !!!!
		int startx=0;
		int starty=0;
		int arcwidth = 0;
		int archeight = 0;
		int arcstart = 0;
		int arcrange = 0;
		double darcrange = 0.0;
		double darcstart = 0.0;

		// Delete Background, paint the red door black
		Graphics2D g2 = getTheGraphics();
		if(reverse[frameNr]) {	// select color to paint
			g2.setColor(colWhite);
		} else {
			g2.setColor(colBlack);
		}
		g2.fillRect(0, 0, prevWidth, prevHeight);

		// Translate and paint layers
		g2.translate(translX, translY);		// is calculated by int translX = (int)((originalWidth*scaleFactor)/2); and so on
		for(int lay = 0;lay<layerCount;lay++) {
			arcwidth	= (int)(scaleFactor * wide2[frameNr]);
			archeight	= (int)(scaleFactor * wide2[frameNr]);
			startx		= (int)((scaleFactor * kx1[frameNr])-((double)arcwidth/2.0));
			starty		= (int)((scaleFactor * (ky1[frameNr]*-1.0)) - ((double)archeight/2.0));
			//arcrange	= (int)(((wide1[frameNr]/(double)layerCount)*((double)layerCount-(double)lay))*-1);
			//arcstart	= (int)(((double)arcrange/2.0*-1.0)+90.0);	// both lines uncomment if you want to use java fillArc
			darcrange = (((wide1[frameNr]/(double)layerCount)*((double)layerCount-(double)lay)));
			darcstart = (double)darcrange/2.0+90.0;

			if(reverse[frameNr]) {	// select color to paint
				g2.setColor(greys[layerCount-lay-1]);
			} else {
				g2.setColor(greys[lay]);
			}
			//g2.fillArc(startx, starty, arcwidth, archeight, arcstart, arcrange);
			fillPolyArc(g2, startx, starty, arcwidth, archeight, darcstart, darcrange);
		}
		g2.translate(-translX, -translY);	// reset before paint other stuff that has origin at 0,0

		// do the rest
		if(infoOn) {
			paintTimingInfo(g2);
		}
		switchBuffer();
	}

	public void fillPolyArc(Graphics2D g2, int x, int y, int w, int h, double arcstart, double arc) {
		// Paints a Polygon like the fillArc method. But has not a REAL circle outline, but 10 vertices
		// Color must be set before
		// xp[0] is the center point of circle. Other vertices are on circle outline.

		double kante = w/2;
		int xp[] = new int[10];	// coordinate arrays for a polygon object
		int yp[] = new int[10];

		xp[0]=x+(int)kante;	// Center Point
		yp[0]=y+(int)kante;

		xp[1]= xp[0]+(int)(Math.cos(Math.toRadians(arcstart-(arc*(0.0/9.0))))*kante);	// First left vertex
		yp[1]= yp[0]+(int)(Math.sin(Math.toRadians(arcstart-(arc*(0.0/9.0))))*kante)*-1;

		xp[2]= xp[0]+(int)(Math.cos(Math.toRadians(arcstart-(arc*(1.125/9.0))))*kante);
		yp[2]= yp[0]+(int)(Math.sin(Math.toRadians(arcstart-(arc*(1.125/9.0))))*kante)*-1;

		xp[3]= xp[0]+(int)(Math.cos(Math.toRadians(arcstart-(arc*(2.25/9.0))))*kante);
		yp[3]= yp[0]+(int)(Math.sin(Math.toRadians(arcstart-(arc*(2.25/9.0))))*kante)*-1;


		xp[4]= xp[0]+(int)(Math.cos(Math.toRadians(arcstart-(arc*(3.375/9.0))))*kante);
		yp[4]= yp[0]+(int)(Math.sin(Math.toRadians(arcstart-(arc*(3.375/9.0))))*kante)*-1;

		xp[5]= xp[0]+(int)(Math.cos(Math.toRadians(arcstart-(arc*(4.5/9.0))))*kante);
		yp[5]= yp[0]+(int)(Math.sin(Math.toRadians(arcstart-(arc*(4.5/9.0))))*kante)*-1;	// Middle

		xp[6]= xp[0]+(int)(Math.cos(Math.toRadians(arcstart-(arc*(5.625/9.0))))*kante);
		yp[6]= yp[0]+(int)(Math.sin(Math.toRadians(arcstart-(arc*(5.625/9.0))))*kante)*-1;

		xp[7]= xp[0]+(int)(Math.cos(Math.toRadians(arcstart-(arc*(6.75/9.0))))*kante);
		yp[7]= yp[0]+(int)(Math.sin(Math.toRadians(arcstart-(arc*(6.75/9.0))))*kante)*-1;

		xp[8]= xp[0]+(int)(Math.cos(Math.toRadians(arcstart-(arc*(7.875/9.0))))*kante);
		yp[8]= yp[0]+(int)(Math.sin(Math.toRadians(arcstart-(arc*(7.875/9.0))))*kante)*-1;

		xp[9]= xp[0]+(int)(Math.cos(Math.toRadians(arcstart-(arc*(9.0/9.0))))*kante);	// Last right vertex
		yp[9]= yp[0]+(int)(Math.sin(Math.toRadians(arcstart-(arc*(9.0/9.0))))*kante)*-1;

		int n = 10;
		Polygon p = new Polygon(xp, yp, n);  // This polygon represents a triangle with the above vertices.

		g2.fillPolygon(p);     // Fills the triangle above.
	}

	public void shapeRendererHorizontThreeStepFader(int frameNr) {
		// *** shapeType = 14 !!!!
		int lx = 0;
		int ly = 0;
		int lw = 0;
		int lh = 0;
		int cv = (int)((originalHeight/2)*scaleFactor*-1); 	//compare value if last frame is lower than screen height

		// Delete Background, paint the red door black
		Graphics2D g2 = getTheGraphics();
		if(reverse[frameNr]) {	// select color to paint
			g2.setColor(colWhite);
		} else {
			g2.setColor(colBlack);
		}
		g2.fillRect(0, 0, prevWidth, prevHeight);

		// Translate and paint layers
		g2.translate(translX, translY);		// is calculated by int translX = (int)((originalWidth*scaleFactor)/2); and so on
		for(int lay = 0;lay<layerCount;lay++) {
			lx = (int)(scaleFactor * kx1[frameNr]);
			ly = (int)((scaleFactor *(ky1[frameNr]+(double)originalHeight) - (layerCount*wide1[frameNr]/2) + ((double)lay * wide1[frameNr])))*-1;
			//System.out.println("ly: "+ly+ " cv: "+cv );
			if(ly>cv)
				ly = cv;
			lw = (int)(scaleFactor * (double)originalWidth);
			lh = (int)(scaleFactor * ((double)originalHeight));
			if(reverse[frameNr]) {	// select color to paint
				g2.setColor(greys[layerCount-lay-1]);
			} else {
				g2.setColor(greys[lay]);
			}
			g2.fillRect(lx, ly, lw, lh);
		}
		g2.translate(-translX, -translY);	// reset before paint other stuff that has origin at 0,0

		// do the rest
		if(infoOn) {
			paintTimingInfo(g2);
		}
		switchBuffer();
	}

	public void fadeRendererHorizontal(int frameNr) {
		// *** shapeType = 2 !!!!
		int lx = 0;
		int ly = 0;
		int lw = 0;
		int lh = 0;
		int startLayer 		= wide3[frameNr].intValue();
		int endLayer 		= wide4[frameNr].intValue();
		int localLayerCount = endLayer - startLayer;

		// Delete Background, paint the red door black
		Graphics2D g2 = getTheGraphics();
		if(reverse[frameNr]) {	// select color to paint
			if(startLayer == 0) {
				g2.setColor(colWhite);
			} else {
				g2.setColor(greys[startLayer]);
			}
		} else {
			if(startLayer == 0) {
				g2.setColor(colBlack);
			} else {
				g2.setColor(greys[startLayer]);
			}
		}
		g2.fillRect(0, 0, prevWidth, prevHeight);
//System.out.println("KY Wert: "+ky1[frameNr]);
		// Translate and paint layers
		g2.translate(translX, translY);		// is calculated by int translX = (int)((originalWidth*scaleFactor)/2); and so on
		for(int lay = startLayer;lay<endLayer;lay++) {

			// A simple rect .
			lx = (int)(scaleFactor * kx1[frameNr]);
			ly = (int)(scaleFactor *((ky1[frameNr]+(double)originalHeight+wide1[frameNr]) + ((double)lay * wide1[frameNr])))*-1;
			if(directionY[frameNr]==-1) {
				ly = (int)(scaleFactor *((ky1[frameNr]*directionY[frameNr]-wide1[frameNr]) - ((double)lay * wide1[frameNr])))*-1;
			}
			lw = (int)(scaleFactor * (double)originalWidth);
			lh = (int)(scaleFactor * (double)originalHeight);
			if(reverse[frameNr]) {	// select color to paint
				g2.setColor(greys[endLayer-lay-1]);
				//g2.setColor(greys[startLayer+lay-1]);
			} else {
				g2.setColor(greys[lay]);
				//g2.setColor(greys[endLayer-lay-1]);
			}
			g2.fillRect(lx, ly, lw, lh);
		}
		g2.translate(-translX, -translY);	// reset before paint other stuff that has origin at 0,0

		// do the rest
		if(infoOn) {
			paintTimingInfo(g2);
		}
		switchBuffer();
	}

//*******
	public void shapeRendererPolyRect(int frameNr) {	// paints one figure per frame. A figure is a polygon with 4 points
		// *** shapeType = 1 !!!!
		int xp[] = new int[4];	// coordinate arrays for a polygon object
		int yp[] = new int[4];
		int domVal1 = 0;
		int domVal2 = 0;
		int dominanceWidth1 = 0;
		int dominanceWidth2 = 0;
		if(dominance[frameNr]) {
			dominanceWidth1 = (int)(wide1[frameNr] * domStrength[frameNr]);
			dominanceWidth2 = (int)(wide2[frameNr] * domStrength[frameNr]);
		}
		// Delete Background, paint the red door black
		Graphics2D g2 = getTheGraphics();
		if(reverse[frameNr]) {	// select color to paint
			g2.setColor(colWhite);
		} else {
			g2.setColor(colBlack);
		}
		g2.fillRect(0, 0, prevWidth, prevHeight);

		g2.translate(translX, translY);		// is calculated by int translX = (int)((originalWidth*scaleFactor)/2); and so on
		for(int lay = layerCount;lay>0;lay--) {
			// A simple rect polygon.
			if(dominance[frameNr]) {
				if(lay>domLayerNr[frameNr]) {
					domVal1 = dominanceWidth1;
					domVal2 = dominanceWidth2;
				} else {
					domVal1 = 0;
					domVal2 = 0;
				}
			}
			xp[0]=((int)(scaleFactor * kx1[frameNr]) - (int)(scaleFactor * lay * (domVal1+wide1[frameNr])))*directionX[frameNr];
			xp[1]=((int)(scaleFactor * kx1[frameNr]) + (int)(scaleFactor * lay * (domVal1+wide1[frameNr])))*directionX[frameNr];
			xp[2]=((int)(scaleFactor * kx2[frameNr]) + (int)(scaleFactor * lay * (domVal2+wide2[frameNr])))*directionX[frameNr];
			xp[3]=((int)(scaleFactor * kx2[frameNr]) - (int)(scaleFactor * lay * (domVal2+wide2[frameNr])))*directionX[frameNr];

			yp[0]=((int)(scaleFactor * ky1[frameNr])*-1)*directionY[frameNr];
			yp[1]=((int)(scaleFactor * ky1[frameNr])*-1)*directionY[frameNr];
			yp[2]=((int)(scaleFactor * ky2[frameNr])*-1)*directionY[frameNr];
			yp[3]=((int)(scaleFactor * ky2[frameNr])*-1)*directionY[frameNr];
			int n = 4;

			Polygon p = new Polygon(xp, yp, n);  // This polygon represents a triangle with the above vertices.

			if(reverse[frameNr]) {	// select color to paint
				g2.setColor(greys[lay-1]);
			} else {
				g2.setColor(greys[layerCount-lay]);
			}

			g2.fillPolygon(p);     // Fills the triangle above.

		}
		g2.translate(-translX, -translY);	// reset before paint other stuff that has origin at 0,0

		if(infoOn) {
			paintTimingInfo(g2);
		}
		switchBuffer();
	}

	public void shapeRendererPerspectiveRays(int frameNr) {
		// *** shapeType = 11 !!!!
		int xp[] = new int[3];	// coordinate arrays for a polygon object
		int yp[] = new int[3];

		// Delete Background, paint the red door black
		Graphics2D g2 = getTheGraphics();
		if(reverse[frameNr]) {	// select color to paint
			g2.setColor(colWhite);
		} else {
			g2.setColor(colBlack);
		}
		g2.fillRect(0, 0, prevWidth, prevHeight);

		g2.translate(translX, translY);		// is calculated by int translX = (int)((originalWidth*scaleFactor)/2); and so on

		for(int lay = layerCount;lay>0;lay--) {
//System.out.println("fr: "+frameNr+" - kx2: "+kx2[frameNr]+" - wide1: "+wide1[frameNr]+" - direction: "+directionX[frameNr]);
			// 	vanishing point
			xp[0]=(int)(scaleFactor * kx1[frameNr]) * directionX[frameNr];
			yp[0]=((int)(scaleFactor * ky1[frameNr])*-1)*directionY[frameNr];

			if(wide3[frameNr]>0) {	// Horizontal Rays
				xp[1]=(int)(scaleFactor * kx2[frameNr]) *directionX[frameNr];
				xp[2]=(int)(scaleFactor * kx2[frameNr]) *directionX[frameNr];

				yp[1]=(((int)(scaleFactor * ky2[frameNr])+ (int)(scaleFactor * lay * (wide1[frameNr])))*-1)*directionY[frameNr];
				yp[2]=(((int)(scaleFactor * ky2[frameNr])- (int)(scaleFactor * lay * (wide2[frameNr])))*-1)*directionY[frameNr];
			} else {				// Vertical Rays
				xp[1]=((int)(scaleFactor * kx2[frameNr]) + (int)(scaleFactor * lay * (wide1[frameNr])))*directionX[frameNr];
				xp[2]=((int)(scaleFactor * kx2[frameNr]) - (int)(scaleFactor * lay * (wide2[frameNr])))*directionX[frameNr];

				yp[1]=((int)(scaleFactor * ky2[frameNr])*-1)*directionY[frameNr];
				yp[2]=((int)(scaleFactor * ky2[frameNr])*-1)*directionY[frameNr];
			}
			int n = 3;

			Polygon p = new Polygon(xp, yp, n);  // This polygon represents a triangle with the above vertices.

			if(reverse[frameNr]) {	// select color to paint
				g2.setColor(greys[lay-1]);
			} else {
				g2.setColor(greys[layerCount-lay]);
			}

			g2.fillPolygon(p);     // Fills the triangle above.
		}

		g2.translate(-translX, -translY);	// reset before paint other stuff that has origin at 0,0
		if(infoOn) {
			paintTimingInfo(g2);
		}
		switchBuffer();
	}

	public void shapeRendererCircleInOut(int frameNr) {
		// *** shapeType = 4 !!!!
		int xp = 0;
		int yp = 0;
		int radp = 0;
		int domVal1 = 0;
		int dominanceWidth1 = 0;
		if(dominance[frameNr]) {
			dominanceWidth1 = (int)(wide1[frameNr] * domStrength[frameNr]);
		}
		// Delete Background, paint the red door black
		Graphics2D g2 = getTheGraphics();
		if(reverse[frameNr]) {	// select color to paint
			g2.setColor(colWhite);
		} else {
			g2.setColor(colBlack);
		}
		g2.fillRect(0, 0, prevWidth, prevHeight);

		g2.translate(translX, translY);		// is calculated by int translX = (int)((originalWidth*scaleFactor)/2); and so on

		if(mode[frameNr]) {	// if grow mode, do this. if not, do the other
			int biggestRadius = (int)(wide1[frameNr] * (double)layerCount);
			int growLayCount = (int)((double)biggestRadius / wide2[frameNr]);	// Must have grown minimum the circle width to be painted
			if(growLayCount>layerCount) growLayCount = layerCount;
			for(int lay = 0;lay<growLayCount;lay++) {
				// No Dominance Layer possible in grow Mode
				radp = (int)((double)biggestRadius - ((double)lay * wide2[frameNr]));
				xp = (int)(scaleFactor * kx1[frameNr]-(radp/2));		// fillOval Coordinates are in the left upper corner, not the center
				yp = ((int)(scaleFactor * ky1[frameNr]+(radp/2))*-1);

				if(reverse[frameNr]) {	// select color to paint
					g2.setColor(greys[layerCount-lay-1]);
				} else {
					g2.setColor(greys[lay]);
				}
				g2.fillOval(xp, yp, radp, radp);     // Fills the circle
			}
		} else {
			for(int lay = layerCount;lay>0;lay--) {
				if(dominance[frameNr]) {
					if(lay>domLayerNr[frameNr]) {
						domVal1 = dominanceWidth1;
					} else {
						domVal1 = 0;
					}
				}
				radp = (int)(scaleFactor * lay * (domVal1+wide1[frameNr]));
				xp = (int)(scaleFactor * kx1[frameNr]-(radp/2));		// fillOval Coordinates are in the left upper corner, not the center
				yp = ((int)(scaleFactor * ky1[frameNr]+(radp/2))*-1);

				if(reverse[frameNr]) {	// select color to paint
					g2.setColor(greys[lay-1]);
				} else {
					g2.setColor(greys[layerCount-lay]);
				}
				g2.fillOval(xp, yp, radp, radp);     // Fills the circle
			}
		}


		g2.translate(-translX, -translY);	// reset before paint other stuff that has origin at 0,0
		if(infoOn) {
			paintTimingInfo(g2);
		}
		switchBuffer();
	}

	public void shapeRendererQuadBox(int frameNr) {
		// *** shapeType = 12 !!!!
		int xp = 0;
		int yp = 0;
		int radp = 0;
		int domVal1 = 0;
		int dominanceWidth1 = 0;
		if(dominance[frameNr]) {
			dominanceWidth1 = (int)(wide1[frameNr] * domStrength[frameNr]);
		}
		// Delete Background, paint the red door black
		Graphics2D g2 = getTheGraphics();
		if(reverse[frameNr]) {	// select color to paint
			g2.setColor(colWhite);
		} else {
			g2.setColor(colBlack);
		}
		g2.fillRect(0, 0, prevWidth, prevHeight);

		g2.translate(translX, translY);		// is calculated by int translX = (int)((originalWidth*scaleFactor)/2); and so on

		if(mode[frameNr]) {	// if grow mode, do this. if not, do the other
			int biggestRadius = (int)(wide1[frameNr] * (double)layerCount);
			int growLayCount = (int)((double)biggestRadius / wide2[frameNr]);	// Must have grown minimum the circle width to be painted
			if(growLayCount>layerCount) growLayCount = layerCount;
			for(int lay = 0;lay<growLayCount;lay++) {
				// No Dominance Layer possible in grow Mode
				radp = (int)((double)biggestRadius - ((double)lay * wide2[frameNr]));
				xp = (int)(scaleFactor * kx1[frameNr]-(radp/2));		// fillOval Coordinates are in the left upper corner, not the center
				yp = ((int)(scaleFactor * ky1[frameNr]+(radp/2))*-1);

				if(reverse[frameNr]) {	// select color to paint
					g2.setColor(greys[layerCount-lay-1]);
				} else {
					g2.setColor(greys[lay]);
				}
				//g2.fillOval(xp, yp, radp, radp);     // Fills the circle
				g2.fillRect(xp,yp,radp,radp);
			}
		} else {
			for(int lay = layerCount;lay>0;lay--) {
				if(dominance[frameNr]) {
					if(lay>domLayerNr[frameNr]) {
						domVal1 = dominanceWidth1;
					} else {
						domVal1 = 0;
					}
				}
				radp = (int)(scaleFactor * lay * (domVal1+wide1[frameNr]));
				xp = (int)(scaleFactor * kx1[frameNr]-(radp/2));		// fillOval Coordinates are in the left upper corner, not the center
				yp = ((int)(scaleFactor * ky1[frameNr]+(radp/2))*-1);

				if(reverse[frameNr]) {	// select color to paint
					g2.setColor(greys[lay-1]);
				} else {
					g2.setColor(greys[layerCount-lay]);
				}
				//g2.fillOval(xp, yp, radp, radp);     // Fills the circle
				g2.fillRect(xp,yp,radp,radp);
			}
		}


		g2.translate(-translX, -translY);	// reset before paint other stuff that has origin at 0,0
		if(infoOn) {
			paintTimingInfo(g2);
		}
		switchBuffer();
	}

	public void shapeRendererRectBox(int frameNr) {
		// *** shapeType = 12 !!!!
		int xp = 0;
		int yp = 0;
		int radp = 0;
		int domVal1 = 0;
		int dominanceWidth1 = 0;
		if(dominance[frameNr]) {
			dominanceWidth1 = (int)(wide1[frameNr] * domStrength[frameNr]);
		}
		// Delete Background, paint the red door black
		Graphics2D g2 = getTheGraphics();
		if(reverse[frameNr]) {	// select color to paint
			g2.setColor(colWhite);
		} else {
			g2.setColor(colBlack);
		}
		g2.fillRect(0, 0, prevWidth, prevHeight);

		g2.translate(translX, translY);		// is calculated by int translX = (int)((originalWidth*scaleFactor)/2); and so on

		if(mode[frameNr]) {	// if grow mode, do this. if not, do the other
			int biggestRadius = (int)(wide1[frameNr] * (double)layerCount);
			int growLayCount = (int)((double)biggestRadius / wide2[frameNr]);	// Must have grown minimum the circle width to be painted
			if(growLayCount>layerCount) growLayCount = layerCount;
			for(int lay = 0;lay<growLayCount;lay++) {
				// No Dominance Layer possible in grow Mode
				radp = (int)((double)biggestRadius - ((double)lay * wide2[frameNr]));
				xp = (int)(scaleFactor * kx1[frameNr]-(radp/2));		// fillOval Coordinates are in the left upper corner, not the center
				yp = ((int)(scaleFactor * ky1[frameNr]+(radp/2))*-1);

				if(reverse[frameNr]) {	// select color to paint
					g2.setColor(greys[layerCount-lay-1]);
				} else {
					g2.setColor(greys[lay]);
				}
				//g2.fillOval(xp, yp, radp, radp);     // Fills the circle
				g2.fillRect(xp-(radp/2),yp,2*radp,radp);
			}
		} else {
			for(int lay = layerCount;lay>0;lay--) {
				if(dominance[frameNr]) {
					if(lay>domLayerNr[frameNr]) {
						domVal1 = dominanceWidth1;
					} else {
						domVal1 = 0;
					}
				}
				radp = (int)(scaleFactor * lay * (domVal1+wide1[frameNr]));
				xp = (int)(scaleFactor * kx1[frameNr]-(radp/2));		// fillOval Coordinates are in the left upper corner, not the center
				yp = ((int)(scaleFactor * ky1[frameNr]+(radp/2))*-1);

				if(reverse[frameNr]) {	// select color to paint
					g2.setColor(greys[lay-1]);
				} else {
					g2.setColor(greys[layerCount-lay]);
				}
				//g2.fillOval(xp, yp, radp, radp);     // Fills the circle
				g2.fillRect(xp-(radp/2),yp,2*radp,radp);
			}
		}


		g2.translate(-translX, -translY);	// reset before paint other stuff that has origin at 0,0
		if(infoOn) {
			paintTimingInfo(g2);
		}
		switchBuffer();
	}

	public void shapeRendererLayerSpot(int frameNr) {
		// *** shapeType = 10 !!!!
		int xp = 0;
		int yp = 0;
		int radp = 0;
		int spottedLayer = domLayerNr[frameNr];

		// Delete Background, paint the red door black
		Graphics2D g2 = getTheGraphics();
		if(reverse[frameNr]) {	// select color to paint
			g2.setColor(colWhite);
		} else {
			g2.setColor(colBlack);
		}
		g2.fillRect(0, 0, prevWidth, prevHeight);

		g2.translate(translX, translY);		// is calculated by int translX = (int)((originalWidth*scaleFactor)/2); and so on

		//for(int lay = layerCount;lay>0;lay--) {
		for(int lay = layerCount;lay>spottedLayer;lay--) {
			radp = (int)(scaleFactor * (lay-spottedLayer) * (wide1[frameNr]));
			xp = (int)(scaleFactor * kx1[frameNr]-(radp/2));		// fillOval Coordinates are in the left upper corner, not the center
			yp = ((int)(scaleFactor * ky1[frameNr]+(radp/2))*-1);

			if(reverse[frameNr]) {	// select color to paint
				g2.setColor(greys[lay-1]);
			} else {
				g2.setColor(greys[layerCount-lay]);
			}
			g2.fillOval(xp, yp, radp, radp);     // Fills the circle
		}

		g2.translate(-translX, -translY);	// reset before paint other stuff that has origin at 0,0
		if(infoOn) {
			paintTimingInfo(g2);
		}
		switchBuffer();
	}

	public void shapeRendererDualCircle(int frameNr) {
		// *** shapeType = 9 !!!!
		int xp = 0;
		int yp = 0;
		int radp = 0;
		int domVal1 = 0;
		int dominanceWidth1 = 0;
		if(dominance[frameNr]) {
			dominanceWidth1 = (int)(wide1[frameNr] * domStrength[frameNr]);
		}
		// same for second circle:
		int xp2 = 0;
		int yp2 = 0;
		int radp2 = 0;
		int domVal2 = 0;
		int dominanceWidth2 = 0;
		if(dominance[frameNr]) {
			dominanceWidth2 = (int)(wide2[frameNr] * domStrength[frameNr]);
		}

		// Delete Background, paint the red door black
		Graphics2D g2 = getTheGraphics();
		if(reverse[frameNr]) {	// select color to paint
			g2.setColor(colWhite);
		} else {
			g2.setColor(colBlack);
		}
		g2.fillRect(0, 0, prevWidth, prevHeight);

		g2.translate(translX, translY);		// is calculated by int translX = (int)((originalWidth*scaleFactor)/2); and so on


		for(int lay = layerCount;lay>0;lay--) {
			// Paint circle 1
			if(dominance[frameNr]) {
				if(lay>domLayerNr[frameNr]) {
					domVal1 = dominanceWidth1;
				} else {
					domVal1 = 0;
				}
			}
			radp = (int)(scaleFactor * lay * (domVal1+wide1[frameNr]));
			xp = (int)(scaleFactor * kx1[frameNr]-(radp/2));		// fillOval Coordinates are in the left upper corner, not the center
			yp = ((int)(scaleFactor * ky1[frameNr]+(radp/2))*-1);

			if(reverse[frameNr]) {	// select color to paint
				g2.setColor(greys[lay-1]);
			} else {
				g2.setColor(greys[layerCount-lay]);
			}
			g2.fillOval(xp, yp, radp, radp);     // Fills the circle
			// Paint circle 2
			if(dominance[frameNr]) {
				if(lay>domLayerNr[frameNr]) {
					domVal2 = dominanceWidth2;
				} else {
					domVal2 = 0;
				}
			}
			radp2 = (int)(scaleFactor * lay * (domVal2+wide2[frameNr]));
			xp2 = (int)(scaleFactor * kx2[frameNr]-(radp2/2));		// fillOval Coordinates are in the left upper corner, not the center
			yp2 = ((int)(scaleFactor * ky2[frameNr]+(radp2/2))*-1);

			// same color as before. No change
			g2.fillOval(xp2, yp2, radp2, radp2);     // Fills the circle
		}



		g2.translate(-translX, -translY);	// reset before paint other stuff that has origin at 0,0
		if(infoOn) {
			paintTimingInfo(g2);
		}
		switchBuffer();
	}

	public void shapeRendererAudioSync(int frameNr) {
		// *** shapeType = 7 !!!!
		// Delete Background, paint the red door black
		Graphics2D g2 = getTheGraphics();
		g2.setColor(colBlack);
		g2.fillRect(0, 0, prevWidth, prevHeight);

		if(dominance[frameNr]) {
			g2.setColor(colWhite);
			g2.fillRect(0, 0, prevWidth, prevHeight);
		}

		//g2.setColor(greys[domLayerNr[frameNr]]);		// layer nr to wait in is stored in the domLayer field
		//g2.fillRect(0, 0, prevWidth, prevHeight);dominance[currentFrame]
		if(infoOn) {
			paintTimingInfo(g2);
		}
		switchBuffer();
	}

	public void shapeRendererWaitInLayer(int frameNr) {
		// *** shapeType = 3 !!!!
		// Delete Background, paint the red door black
		Graphics2D g2 = getTheGraphics();
		g2.setColor(colBlack);
		g2.fillRect(0, 0, prevWidth, prevHeight);
		g2.setColor(greys[domLayerNr[frameNr]]);		// layer nr to wait in is stored in the domLayer field
		g2.fillRect(0, 0, prevWidth, prevHeight);
		if(infoOn) {
			paintTimingInfo(g2);
		}
		switchBuffer();
	}

	public void shapeRendererWaitEmpty(int frameNr) {	// All Layers visible (upper layer visible alone)
		// *** shapeType = 6 !!!!
		// Delete Background, paint the red door black
		Graphics2D g2 = getTheGraphics();
		//g2.setColor(colWhite);
		if(reverse[frameNr]) {	// if reverse all layers are transparent (lowest layer shown)
			g2.setColor(colBlack);
		} else {
			g2.setColor(colWhite);
		}
		g2.fillRect(0, 0, prevWidth, prevHeight);
		if(infoOn) {
			paintTimingInfo(g2);
		}
		switchBuffer();
	}

	public void shapeRendererSimpleRect(int frameNr) {	// ** Only for testing and debugging
		// ** Standard Start start
		// *** shapeType = 99 !!!!
		Graphics2D g2 = getTheGraphics();	// Delete Background
		g2.setColor(colBlack);
		g2.fillRect(0, 0, prevWidth, prevHeight);
		// ** Standard Start end
		// ***************************************

		int px1 = (int)(kx1[frameNr]*scaleFactor);
		int py1 = (int)(ky1[frameNr]*scaleFactor*-1);	// *-1 because coordinate system is headup. Only coordinates need this. Height and Width dont.
		int pwid= (int)((Math.abs(kx2[frameNr])+Math.abs(kx1[frameNr]))*scaleFactor); // In this case negative and positive value must be added up
		int phei= (int)((Math.abs(ky2[frameNr])+Math.abs(ky1[frameNr]))*scaleFactor);
		g2.setColor(colWhite);
		g2.translate(translX, translY);		// is calculated by int translX = (int)((originalWidth*scaleFactor)/2); and so on
		g2.drawRect(px1, py1, pwid, phei);
		g2.translate(-translX, -translY);

		// ***************************************
		// *** Standard ending start
		if(infoOn) {
			paintTimingInfo(g2);
		}
		switchBuffer();
	}

	public void shapeRenderEmpty(int frameNr) {		// Renders empty Frames Black
		// *** shapeType = 0 !!!!
		// Delete Background, paint the red door black
		Graphics2D g2 = getTheGraphics();
		g2.setColor(colBlack);
		g2.fillRect(0, 0, prevWidth, prevHeight);
		g2.setColor(colWhite);
		g2.drawLine(0,0,(int)(scaleFactor*originalWidth),(int)(scaleFactor*originalHeight));
		g2.drawLine(0,(int)(scaleFactor*originalHeight),(int)(scaleFactor*originalWidth),0);
		if(infoOn) {
			paintTimingInfo(g2);
		}
		switchBuffer();
	}
//**************************************
	public void calcControl() {
		//Preset shape type with 0
		for(int i = 0;i<projectFrames;i++) {	// renderer 0 default - rendererEmpty
			shapeType[i] = 0;
		}

		printTiming(currentFrame);
		calcAudioSync(7, 4.0);		// renderer type, duration (always 4 in this figure)

		printTiming(currentFrame);
		calcWaitEmpty(6, 4.0, false);	// wait for 8 beats without layer from frame 1009 with renderer 3

		printTiming(currentFrame);
		calcInOutRays( 1, 1.0, 2.0, 1.0,					//
						1340, 600, 1000, -600, 10,  2,
						-1200, 600, 1000, -600, 300, 10,	//x1,y1,x2,y2,wide1,wide2
						1340, 600, 1000, -600, 10,  2,
						1, 1, 1.0, 1,		// mirrorX, mirrorY(-1), pulse, pulseStrength
						true, false,		// reverse, ramp
						false, 0, 2.0, 0);	// dom, domLayer, domStrength, domFalloff

		printTiming(currentFrame);
		calcHorizontalFader(2, 4.0,				// shape type, duration
						0, layerCount,			// startLayer, endLayer
						-1, 1.0, 1.0,			// mirrorY(-1), pulse strength, pulse speed
						true, false);			// reverse, ramp

		// Render Type 3 is Wait in Layer

		printTiming(currentFrame);
		calcSimpleInOutCircle(4, 1.0, 2.0, 1.0,		// shape type, duration in stay out
						406.0, 206.0,				// centered coordinates
						0, 170, 0,					// three wide parameter in mid out
						1.0, 1.0,					// pulse strength and speed
						false, false, false,		// reverse, ramp, straight grow
						false, 7, 0.5, 1.0);		// dom on, dom layer, strength, falloff

		// Render Type 5 is Intro Animation. This is Special and only programmed for the beginning of TimeReloaded
		// Render Type 6 is WaitEmpty (see first figure)
		// Render Type 7 is AudioSync Blinking on the 1st frame of each quarter note

		printTiming(currentFrame);
		calcMovedGrowCircle(8, 2.0, 1.0, 2.0, 1.0, 2.0,		// shape type, duration in stay1 move stay2 out
						-383.0, 140.0,					// centered start coordinates
						-3.0, -263.0,				// centered end coordinates
						0, 150, 220, 180, 0,		// five wide parameter in stay move stay out
						1.0, 1.0,					// pulse strength and speed
						false, false,					// reverse, ramp
						false, 7, 0.5, 1.0);		// dom on, dom layer, strength, falloff

		printTiming(currentFrame);
		calcDualCircle(9, 1.0, 1.0, 1.0, 1.0, 1.0,		// shape type, duration in stay1 move stay2 out
						-960.0, -540.0,					// centered start coordinates for circle 1
						-446.0, -373.0,					// centered stay1 coordinates for circle 1
						-380.0, 146.0,					// centered stay2 coordinates for circle 1
						-960.0, 540.0,					// centered end coordinates for circle 1
						0, 100, 200, 180, 150, 0,		// six wide parameter in start, stay1, move, stay2, end for circle 1
						960.0, 540.0,					// centered start coordinates for circle 2
						403.0, 220.0,					// centered stay1 coordinates for circle 2
						456.0, -363.0,					// centered stay2 coordinates for circle 2
						960.0, -540.0,				// centered end coordinates for circle 2
						0, 100, 200, 180, 150, 0,		// six wide parameter in start, stay1, move, stay2, end for circle 2
						1.0, 1.0,						// pulse strength and speed
						false, false, true,				// reverse, ramp, Synchronity
						false, 7, 0.5, 1.0);			// dom on, dom layer, strength, falloff

		printTiming(currentFrame);
		calcLayerSpot(10, 2.0, 2.0, 2.0,				// shape type, duration in move out
						5,								// The Layer that is spottet!
						-796.0, -510.0,					// centered end coordinates for circle 1
						-783.0, -360.0,					// centered start coordinates for circle 1
						0, 210, 140, 0,					// wide parameter in start, startend, move, end for circle 1
						1.0, 1.0,						// pulse strength and speed
						false, false);					// reverse, ramp (does not work linear!! ramp always true!)

		printTiming(currentFrame);
		calcPerspectiveRays(11, 1, 1.0, 1.0, 1.0, 1.0, 1.0,		// shape type, orientation(1=horizontal) duration in stay1 move stay2 out
						-6380.0, -260.0,				// vanishing point coordinates
						1000.0, -170.0,					// centered start coordinates for Point 2
						1000.0, -350.0,					// centered stay1 coordinates for Point 2
						1000.0, -170.0,					// centered stay2 coordinates for Point 2
						1000.0, -3000.0,				// centered end coordinates for Point 2
						0, 100, 120, 100, 90, 80,		// six upper wide parameter in start, endOfstartWide, stay1, move, stay2, end for circle 1
						0, 20, 40, 20, 15, 5,		// six lower wide parameter in start, endOfstartWide, stay1, move, stay2, end for Point 2
						1.0, 1.0,						// pulse strength and speed
						false, false,					// reverse, ramp,
						1, 1);								// direction x and y

		printTiming(currentFrame);
		calcSimpleInOutCircle(12, 1.0, 2.0, 1.0,		// shape type, duration in stay out
						26.0, 33.0,				// centered coordinates
						0, 130, 0,					// three wide parameter in mid out
						1.0, 1.0,					// pulse strength and speed
						false, false, false,			// reverse, ramp, straight grow
						false, 7, 0.5, 1.0);		// dom on, dom layer, strength, falloff

		printTiming(currentFrame);
		calcSimpleInOutCircle(13, 1.0, 2.0, 1.0,		// shape type, duration in stay out
						-10.0, -280.0,				// centered coordinates
						0, 130, 0,					// three wide parameter in mid out
						1.0, 1.0,					// pulse strength and speed
						false, false, false,			// reverse, ramp, straight grow
						false, 7, 0.5, 1.0);		// dom on, dom layer, strength, falloff

		calcHorizontThreeStepFader(14, 1.0, 2.0, 1.0, 2.0, 1.0, 2.0, 1.0,		// shape type, duration startMove, stay, move, stay ...
						540.0, 							// start Y coordinate
						170.0, -176.0, -376.0,			// Y coordinates for stay1, stay2, stay3
						-640.0,							// end Y coordinate
						0, 11, 7, 11, 1,					// five wide parameter in start, stay1, move, stay2, move, stay3, end
						1.0, 1.0,						// pulse strength and speed
						true, false,						// reverse, ramp,
						1);								// moving direction up/down DIRECTION DONT WORK

		printTiming(currentFrame);
		calcCircleRose(15, 1.0, 2.0, 1.0, 2.0, 1.0, 2.0,		// shape type, duration startMove, stay, move, stay ...
						originalWidth*2, 				// rayLength
						90.0, -453.0,						// start position x,y
						90.0, -20.0,						// midPosition x,y
						90.0, -453.0,						// endPosition x,y
						0.0, 180.0,						// start and stay1 angle
						200.0, 350.0,					// move1 and stay 2 angle
						340.0, 180.0,					// move2 and stay 3 angle
						0.0,							// end angle
						0.1, 1.0,						// pulse strength and speed PULSE STRENGTH IS USED HERE
						false, false);					// reverse and ramp

		printTiming(currentFrame);
		calcWaitEmpty(6, 16.0, true);	// wait for 8 beats without layer from frame 1009 with renderer 3

		// print status on calculated frames/time/beats
		int statFrames = currentFrame;
		int tempSecond = (int)((double)statFrames / framesPerSecond)+1;
		int statMinute = (int)((double)tempSecond/(double)60);
		int statSecond = (int)((double)tempSecond - ((double)statMinute*(double)60));
		int statBeats = (int)((double)statFrames/(double)framesPerBeat);
		String dura = makeLeadingZeros(statMinute, 2)+":"+makeLeadingZeros(statSecond, 2);

		printStatus("Calculated "+statFrames+" frames for "+dura+" (min:sec). That's "+statBeats+" beats");
	}

/* Sample call:
		calcCircleRose(15, 4.0, 8.0, 4.0, 8.0, 4.0, 8.0,		// shape type, duration startMove, stay, move, stay ...
						originalWidth, 					// rayLength
						0.0,-540.0,						// start position x,y
						0.0,0.0,						// midPosition x,y
						0.0,-540.0,						// endPosition x,y
						0.0, 180.0,						// start and stay1 angle
						220.0, 270.0,					// move1 and stay 2 angle
						220.0, 180.0,					// move2 and stay 3 angle
						0.0,							// end angle
						1.0, 1.0,						// pulse strength and speed
						false, false);					// reverse and ramp
*/
	public void calcCircleRose(int shapeTypeP, double startMoveDuration, double stay1Duration,
							double move2Duration, double stay2Duration,  double move3Duration,
							double stay3Duration,

							double rayLength,

							double startX, double startY,
							double stayX, double stayY,
							double endX, double endY,

							double startAng, double stay1Ang,
							double move1Ang, double stay2Ang,
							double move2Ang, double stay3Ang,
							double endAng,

							double pulseStrength, double pulseSpeed,
							boolean reversePaint, boolean useRamp) {

		double ps = pulseStrength;	// pulse strength - 0.0 for no pulse, higher for pulse
		// duration parameters are in score notation. amount in quarter notes
		int frameDurationStart 	= (int)(framesPerBeat * startMoveDuration)+1;
		int frameDurationStay1	= (int)(framesPerBeat * stay1Duration)+1;
		int frameDurationMove2 	= (int)(framesPerBeat * move2Duration)+1;
		int frameDurationStay2 	= (int)(framesPerBeat * stay2Duration)+1;

		int frameDurationMove3 	= (int)(framesPerBeat * move3Duration)+1;
		int frameDurationStay3 	= (int)(framesPerBeat * stay3Duration)+1;

		int frameDurationAll = frameDurationStart+frameDurationStay1+frameDurationMove2+frameDurationStay2+frameDurationMove3+frameDurationStay3;

		double halfFrameDuration = 1.0;
		double distanceX = 1.0;
		double distanceY = 1.0;
		double distanceStepX = 1.0;
		double distanceStepY = 1.0;
		double offsetX = 1.0;
		double offsetY = 1.0;
		double offAng = 0.0;
		double distAng = 1.0;
		double dWideAng = 1.0;

		int savedCurrentFrame = currentFrame;

		for(int phase = 0;phase<6;phase++) {
			if(phase==0) {	// start ... stay at start and angels grow
				halfFrameDuration = (double)frameDurationStart/2;	// calc grow of angels at start position
				distanceX = 0.0;
				distanceY = 0.0;
				distanceStepX = 0.0;
				distanceStepY = 0.0;
				offsetX = startX;
				offsetY = startY;
				offAng = startAng;
				distAng = stay1Ang - startAng;
				dWideAng = distAng/(double)frameDurationStart;
			} else if(phase==1) {	// stay at stay 1 coordinate. Maybe grow.
				halfFrameDuration = (double)frameDurationStay1/2;	// calc parameter for start to stay 1 phase
				distanceX = 0.0;
				distanceY = 0.0;
				distanceStepX = 0.0;
				distanceStepY = 0.0;
				offsetX = startX;
				offsetY = startY;
				offAng = stay1Ang;
				distAng = move1Ang - stay1Ang;
				dWideAng = distAng/(double)frameDurationStay1;
			} else if(phase==2) {	// move from stay 1 coordinate to stay 2 coordinate. Maybe grow
				halfFrameDuration = (double)frameDurationMove2/2;	// calc parameter for start to stay 1 phase
				distanceX = stayX-startX;
				distanceY = stayY-startY;
				distanceStepX = distanceX/(double)frameDurationMove2;
				distanceStepY = distanceY/(double)frameDurationMove2;
				offsetX = startX;
				offsetY = startY;
				offAng = move1Ang;
				distAng = stay2Ang - move1Ang;
				dWideAng = distAng/(double)frameDurationMove2;
			} else if(phase==3) {	// stay at stay 2 coordinate. maybe grow
				halfFrameDuration = (double)frameDurationStay2/2;	// calc parameter for start to stay 1 phase
				distanceX = 0.0;
				distanceY = 0.0;
				distanceStepX = 0.0;
				distanceStepY = 0.0;
				offsetX = stayX;
				offsetY = stayY;
				offAng = stay2Ang;
				distAng = move2Ang - stay2Ang;
				dWideAng = distAng/(double)frameDurationStay2;
			} else if(phase==4) {	// move from stay2 to end coordinate. maybe grow or shrink
				halfFrameDuration = (double)frameDurationMove3/2;	// calc parameter for start to stay 1 phase
				distanceX = endX-stayX;
				distanceY = endY-stayY;
				distanceStepX = distanceX/(double)frameDurationMove3;
				distanceStepY = distanceY/(double)frameDurationMove3;
				offsetX = stayX;
				offsetY = stayX;
				offAng = move2Ang;
				distAng = stay3Ang - move2Ang;
				dWideAng = distAng/(double)frameDurationMove3;
			} else if(phase==5) {	// stay at stay 2 coordinate. maybe grow
				halfFrameDuration = (double)frameDurationStay3/2;	// calc parameter for start to stay 1 phase
				distanceX = 0.0;
				distanceY = 0.0;
				distanceStepX = 0.0;
				distanceStepY = 0.0;
				offsetX = endX;
				offsetY = endY;
				offAng = stay3Ang;
				distAng = endAng - stay3Ang;
				dWideAng = distAng/(double)frameDurationStay3;
			}

			// ****** Now the real move calculations.
			// First half of movement
			for(int fr = 0;fr<halfFrameDuration;fr++) {
				if(useRamp) {																			// ps = pulse strength (look at top of method)
					kx1[currentFrame]		= offsetX + ((distanceX/2) * Math.pow((double)fr / halfFrameDuration, 4.0));
					ky1[currentFrame]		= offsetY + ((distanceY/2) * Math.pow((double)fr / halfFrameDuration, 4.0));
					wide1[currentFrame]		= offAng  + ((distAng/2) * Math.pow((double)fr / halfFrameDuration, 4.0)) + (ps * fr);
				} else {
					kx1[currentFrame]		= offsetX + ((double)fr*distanceStepX);
					ky1[currentFrame]		= offsetY + ((double)fr*distanceStepY);
					wide1[currentFrame]		= offAng + ((double)fr*dWideAng) + (ps * fr);
				}
				currentFrame++;	// problem with this. must be resetet in the next
			}
			// Second half of movement
			for(int fr = 0;fr<halfFrameDuration;fr++) {
				if(useRamp) {
					kx1[currentFrame]		= offsetX + (distanceX/2) + ((distanceX/2) * (1-Math.pow(1-((double)fr / halfFrameDuration), 4.0)));
					ky1[currentFrame]		= offsetY + (distanceY/2) + ((distanceY/2) * (1-Math.pow(1-((double)fr / halfFrameDuration), 4.0)));
					wide1[currentFrame]		= offAng  + (distAng/2)  + ((distAng/2)  * (1-Math.pow(1-((double)fr / halfFrameDuration), 4.0))) + (ps * (halfFrameDuration-fr));
				} else {
					kx1[currentFrame]		= offsetX + (halfFrameDuration*distanceStepX) + ((double)fr*distanceStepX);
					ky1[currentFrame]		= offsetY + (halfFrameDuration*distanceStepY) + ((double)fr*distanceStepY);
					wide1[currentFrame]		= offAng  + (halfFrameDuration*dWideAng) + ((double)fr*dWideAng) + (ps * (halfFrameDuration-fr));
				}
				currentFrame++;	// problem with this. must be resetted in the next
			}
		}

		currentFrame = savedCurrentFrame;
		// Static values in this figure
		for(int fr = 0;fr<frameDurationAll;fr++) {
			shapeType[currentFrame]	= shapeTypeP;

			kx2[currentFrame]		= 0.0;
			ky2[currentFrame]		= 0.0;

			wide2[currentFrame]		= rayLength;
			wide3[currentFrame]		= 0.0;
			wide4[currentFrame]		= 0.0;	// unused in this figure

			pulse[currentFrame]		= 1.0;	// its a multiplicator
			reverse[currentFrame]		= reversePaint;	// set as called
			dominance[currentFrame]		= false;	// set all as called
			domLayerNr[currentFrame]	= 0;
			domStrength[currentFrame]	= 0.0;
			domFalloff[currentFrame]	= 0.0;
			directionX[currentFrame]	= 1;
			directionY[currentFrame]	= 1;
			mode[currentFrame]			= false;
			currentFrame++;
		}
	}

	public void calcWaitInLayer(int shapeTypeP, double waitDuration, int waitLayer) {
		double frameDuration 		= framesPerBeat * waitDuration;

		for(int fr = 0;fr<(int)frameDuration;fr++) {	// FIRST HALF  ****************
			shapeType[currentFrame]	= shapeTypeP; 			// Rendered by waitLayerRenderer
			kx1[currentFrame]		= 1;
			ky1[currentFrame]		= 1;
			kx2[currentFrame]		= 1;
			ky2[currentFrame]		= 1;
			wide1[currentFrame]		= 0.0;
			wide2[currentFrame]		= 0.0;
			wide3[currentFrame]		= 0.0;	// unused in PolyRectRenderer ?
			wide4[currentFrame]		= 0.0;	// unused in PolyRectRenderer ?
			pulse[currentFrame]		= 1.0;	// its a multiplicator
			reverse[currentFrame]		= false;	// set as called
			dominance[currentFrame]		= false;	// set all as called
			domLayerNr[currentFrame]	= waitLayer;
			domStrength[currentFrame]	= 1.0;
			domFalloff[currentFrame]	= 1.0;
			directionX[currentFrame]	= 1;
			directionY[currentFrame]	= 1;
			mode[currentFrame]			= false;

			currentFrame++;
		}
	}
	public void calcWaitEmpty(int shapeTypeP, double waitDuration, boolean ereverse) {
		double frameDuration 		= framesPerBeat * waitDuration;

		for(int fr = 0;fr<(int)frameDuration;fr++) {	// FIRST HALF  ****************
			shapeType[currentFrame]	= shapeTypeP; 			// Rendered by waitLayerRenderer
			kx1[currentFrame]		= 1;
			ky1[currentFrame]		= 1;
			kx2[currentFrame]		= 1;
			ky2[currentFrame]		= 1;
			wide1[currentFrame]		= 0.0;
			wide2[currentFrame]		= 0.0;
			wide3[currentFrame]		= 0.0;	// unused in PolyRectRenderer ?
			wide4[currentFrame]		= 0.0;	// unused in PolyRectRenderer ?
			pulse[currentFrame]		= 1.0;	// its a multiplicator
			reverse[currentFrame]		= ereverse;	// set as called
			dominance[currentFrame]		= false;	// set all as called
			domLayerNr[currentFrame]	= 0;
			domStrength[currentFrame]	= 1.0;
			domFalloff[currentFrame]	= 1.0;
			directionX[currentFrame]	= 1;
			directionY[currentFrame]	= 1;
			mode[currentFrame]			= false;

			currentFrame++;
		}
	}

	public void calcIntroAnimation(int shapeTypeP, double saDuration, boolean sareverse) {
		double frameDuration 		= framesPerBeat * saDuration;

		for(int fr = 0;fr<(int)frameDuration;fr++) {
			shapeType[currentFrame]	= shapeTypeP; 			// Rendered by Intro Animation Renderer
			kx1[currentFrame]		= 1;
			ky1[currentFrame]		= 1;
			kx2[currentFrame]		= 1;
			ky2[currentFrame]		= 1;
			wide1[currentFrame]		= 0.0;
			wide2[currentFrame]		= 0.0;
			wide3[currentFrame]		= 0.0;	// unused in PolyRectRenderer ?
			wide4[currentFrame]		= 0.0;	// unused in PolyRectRenderer ?
			pulse[currentFrame]		= 1.0;	// its a multiplicator
			reverse[currentFrame]		= sareverse;	// set as called
			dominance[currentFrame]		= false;
			domLayerNr[currentFrame]	= fr;		// store the local Timing Frame Nr here
			domStrength[currentFrame]	= 1.0;
			domFalloff[currentFrame]	= 1.0;
			directionX[currentFrame]	= 1;
			directionY[currentFrame]	= 1;
			mode[currentFrame]			= false;

			currentFrame++;
		}
	}

	public void calcAudioSync(int shapeTypeP, double duration) {
		double frameDuration 		= framesPerBeat * 4.0;

		for(int fr = 0;fr<(int)frameDuration;fr++) {
			shapeType[currentFrame]	= shapeTypeP; 			// Rendered by Audio Sync Renderer
			kx1[currentFrame]		= 1;
			ky1[currentFrame]		= 1;
			kx2[currentFrame]		= 1;
			ky2[currentFrame]		= 1;
			wide1[currentFrame]		= 0.0;
			wide2[currentFrame]		= 0.0;
			wide3[currentFrame]		= 0.0;
			wide4[currentFrame]		= 0.0;
			pulse[currentFrame]		= 1.0;
			reverse[currentFrame]		= false;
			dominance[currentFrame]		= false;	// here the marker for the sync
			domLayerNr[currentFrame]	= fr;
			domStrength[currentFrame]	= 1.0;
			domFalloff[currentFrame]	= 1.0;
			directionX[currentFrame]	= 1;
			directionY[currentFrame]	= 1;
			mode[currentFrame]			= false;

			if(fr == (int)((frameDuration/4) - (frameDuration/4))) {
				dominance[currentFrame]		= true;
			}
			if(fr == (int)((frameDuration/4)*2 - (frameDuration/4))) {
				dominance[currentFrame]		= true;
			}
			if(fr == (int)((frameDuration/4)*3 - (frameDuration/4))) {
				dominance[currentFrame]		= true;
			}
			if(fr == (int)((frameDuration/4)*4 - (frameDuration/4))) {
				dominance[currentFrame]		= true;
			}
			currentFrame++;
		}
		dominance[currentFrame-1]		= true;
	}
/* Sample call:
		calcHorizontThreeStepFader(14, 4.0, 8.0, 4.0, 8.0, 4.0, 8.0, 4.0,		// shape type, duration startMove, stay, move, stay ...
						540.0, 							// start Y coordinate
						250.0,
						20.0, -200.0,					// Y coordinates for stay1, stay2, stay3
						-540.0,							// end Y coordinate
						0, 200, 300, 200, 150,			// five wide parameter in start, stay1, move, stay2, move, stay3, end
						1.0, 1.0,						// pulse strength and speed
						false, false,					// reverse, ramp,
						1);								// moving direction up/down
*/
	public void calcHorizontThreeStepFader(int shapeTypeP, double startMoveDuration, double stay1Duration,
							double move2Duration, double stay2Duration,  double move3Duration,
							double stay3Duration, double endMoveDuration,
							double startY1, double stay1,
							double stay2, double stay3,
							double endY1,
							double startWide, double stay1Wide,
							double stay2Wide, double stay3Wide,
							double endWide,
							double pulseStrength, double pulseSpeed,
							boolean reversePaint, boolean useRamp,
							int mdirectionY) {

		double ps = 0.1;	// pulse strength - 0.0 for no pulse, higher for pulse
		// duration parameters are in score notation. amount in quarter notes
		int frameDurationStart 	= (int)(framesPerBeat * startMoveDuration)+1;
		int frameDurationStay1	= (int)(framesPerBeat * stay1Duration)+1;
		int frameDurationMove2 	= (int)(framesPerBeat * move2Duration)+1;
		int frameDurationStay2 	= (int)(framesPerBeat * stay2Duration)+1;

		int frameDurationMove3 	= (int)(framesPerBeat * move3Duration)+1;
		int frameDurationStay3 	= (int)(framesPerBeat * stay3Duration)+1;

		int frameDurationEnd	= (int)(framesPerBeat * endMoveDuration)+1;
		int frameDurationAll = frameDurationStart+frameDurationStay1+frameDurationMove2+frameDurationStay2+frameDurationMove3+frameDurationStay3+frameDurationEnd;

		double halfFrameDuration = 1.0;
		double distanceY = 1.0;
		double distanceStepY = 1.0;
		double offsetY = 1.0;
		double offWide = 0.0;
		double distWide = 1.0;
		double dWideStep = 1.0;
		double ty = 0.0;	// temp Y coordinate

		int savedCurrentFrame = currentFrame;

		for(int phase = 0;phase<7;phase++) {
			if(phase==0) {	// move from start coordinate to stay 1 coordinate. Maybe grow
				halfFrameDuration = (double)frameDurationStart/2;	// calc parameter for start to stay 1 phase
				distanceY = stay1 - startY1;
				distanceStepY = distanceY/(double)frameDurationStart;
				offsetY = startY1;
				offWide = startWide;
				distWide = stay1Wide - startWide;
				dWideStep = distWide/(double)frameDurationStart;
			} else if(phase==1) {	// stay at stay 1 coordinate. Maybe grow.
				halfFrameDuration = (double)frameDurationStay1/2;	// calc parameter for start to stay 1 phase
				distanceY = 0.0;
				distanceStepY = 0.0;
				offsetY = stay1;
				offWide = stay1Wide;
				distWide = 0.0;
				dWideStep = 0.0;
			} else if(phase==2) {	// move from stay 1 coordinate to stay 2 coordinate. Maybe grow
				halfFrameDuration = (double)frameDurationMove2/2;	// calc parameter for start to stay 1 phase
				distanceY = stay2 - stay1;
				distanceStepY = distanceY/(double)frameDurationMove2;
				offsetY = stay1;
				offWide = stay1Wide;
				distWide = stay2Wide - stay1Wide;
				dWideStep = distWide/(double)frameDurationMove2;
			} else if(phase==3) {	// stay at stay 2 coordinate. maybe grow
				halfFrameDuration = (double)frameDurationStay2/2;	// calc parameter for start to stay 1 phase
				distanceY = 0.0;
				distanceStepY = 0.0;
				offsetY = stay2;
				offWide = stay2Wide;
				distWide = 0.0;
				dWideStep = 0.0;
			} else if(phase==4) {	// move from stay2 to end coordinate. maybe grow or shrink
				halfFrameDuration = (double)frameDurationMove3/2;	// calc parameter for start to stay 1 phase
				distanceY = stay3 - stay2;
				distanceStepY = distanceY/(double)frameDurationMove3;
				offsetY = stay2;
				offWide = stay2Wide;
				distWide = stay3Wide - stay2Wide;
				dWideStep = distWide/(double)frameDurationMove3;

			} else if(phase==5) {	// stay at stay 2 coordinate. maybe grow
				halfFrameDuration = (double)frameDurationStay3/2;	// calc parameter for start to stay 1 phase
				distanceY = 0.0;
				distanceStepY = 0.0;
				offsetY = stay3;
				offWide = stay3Wide;
				distWide = 0.0;
				dWideStep = 0.0;
			} else if(phase==6) {	// move from stay2 to end coordinate. maybe grow or shrink
				halfFrameDuration = (double)frameDurationEnd/2;	// calc parameter for start to stay 1 phase
				distanceY = endY1 - stay3;
				distanceStepY = distanceY/(double)frameDurationEnd;
				offsetY = stay3;
				offWide = stay3Wide;
				distWide = endWide - stay3Wide;
				dWideStep = distWide/(double)frameDurationEnd;
			}

			// ****** Now the real move calculations.
			// First half of movement
			for(int fr = 0;fr<halfFrameDuration;fr++) {
				if(useRamp) {																			// ps = pulse strength (look at top of method)
					ty = offsetY + ((distanceY/2) * Math.pow((double)fr / halfFrameDuration, 4.0));
					ky1[currentFrame]		= ty;
					ky2[currentFrame]		= ty;
					wide1[currentFrame]		= offWide  + ((distWide/2) * Math.pow((double)fr / halfFrameDuration, 4.0)) + (ps * fr);
				} else {
					ty = offsetY + ((double)fr*distanceStepY);
					ky1[currentFrame]		= ty;
					ky2[currentFrame]		= ty;
					wide1[currentFrame]		= offWide + ((double)fr*dWideStep) + (ps * fr);
				}
				currentFrame++;	// problem with this. must be resetet in the next
			}
			// Second half of movement
			for(int fr = 0;fr<halfFrameDuration;fr++) {
				if(useRamp) {
					ty = offsetY + (distanceY/2) + ((distanceY/2) * (1-Math.pow(1-((double)fr / halfFrameDuration), 4.0)));
					ky1[currentFrame]		= ty;
					ky2[currentFrame]		= ty;
					wide1[currentFrame]		= offWide  + (distWide/2)  + ((distWide/2)  * (1-Math.pow(1-((double)fr / halfFrameDuration), 4.0))) + (ps * (halfFrameDuration-fr));
				} else {
					ty = offsetY + (halfFrameDuration*distanceStepY) + ((double)fr*distanceStepY);
					ky1[currentFrame]		= ty;
					ky2[currentFrame]		= ty;
					wide1[currentFrame]		= offWide  + (halfFrameDuration*dWideStep) + ((double)fr*dWideStep) + (ps * (halfFrameDuration-fr));
				}
				currentFrame++;	// problem with this. must be resetted in the next
			}
		}

		currentFrame = savedCurrentFrame;
		// Static values in this figure
		for(int fr = 0;fr<frameDurationAll;fr++) {
			shapeType[currentFrame]	= shapeTypeP;

			kx1[currentFrame]		= (originalWidth/2)*-1;
			kx2[currentFrame]		= originalWidth/2;

			wide3[currentFrame]		= 0.0;
			wide3[currentFrame]		= 0.0;
			wide4[currentFrame]		= 0.0;	// unused in this figure

			pulse[currentFrame]		= 1.0;	// its a multiplicator
			reverse[currentFrame]		= reversePaint;	// set as called
			dominance[currentFrame]		= false;	// set all as called
			domLayerNr[currentFrame]	= 0;
			domStrength[currentFrame]	= 0.0;
			domFalloff[currentFrame]	= 0.0;
			directionX[currentFrame]	= 1;
			directionY[currentFrame]	= mdirectionY;
			mode[currentFrame]			= false;
			currentFrame++;
		}
	}
/*
Sample call:
calcHorizontalFader(2, 4.0,					// shape type, duration
					0, layerCount,			// startLayer, endLayer
					1.0, 1.0, 1.0,			// mirrorY, pulse strength, pulse speed
					false, true)			// reverse, ramp
*/
	public void calcHorizontalFader(int shapeTypeP, double duration,
							int startLayer, int endLayer,
							int mdirectionY, double pulseStrength, double pulseSpeed,
							boolean reversePaint, boolean useRamp) {
		// if ramp then we set mode to false. if no ramp mode is true and it is linear
		// shapeType = 2
		double frameDuration		= (framesPerBeat * duration)+2;
		double halfFrameDuration	= frameDuration/2;
		double localLayerCount 		= (double)(endLayer - startLayer);
		double movingDistance 		= (double)originalHeight * 2;
		double singleStep			= movingDistance / frameDuration;
		double rectDistance			= (double)originalHeight / localLayerCount;

		double wideStep = 0;
		double layerDistance = (double)originalHeight / localLayerCount;

		double startY = (double)originalHeight / 2;
		double startX = (double)originalWidth / 2*-1;

		for(int fr = 0;fr<(int)halfFrameDuration;fr++) {		// First Half Ease In
			shapeType[currentFrame]	= shapeTypeP; 			// Rendered by SimpleCircleInOut
			if(useRamp) {
				// *** Ease functions by Robert Penner
				// *** http://www.robertpenner.com/easing/
				//      complete distance in pixels  Where we are t    Where have we to go t    easeStrength  // = easeIn
				//easeXUp = halfDistXUp * Math.pow( (double)fr       /   halfFrameDuration,         4.0);
				kx1[currentFrame]		= startX;	// is static in this figure
				ky1[currentFrame]		= startY-((originalHeight) * Math.pow((double)fr / halfFrameDuration, 4.0));
				kx2[currentFrame]		= 0.0;		// unused by circle
				ky2[currentFrame]		= 0.0;
			} else {
				kx1[currentFrame]		= startX;
				ky1[currentFrame]		= startY-((double)fr*singleStep);	// scroll it out complete
				kx2[currentFrame]		= 0.0;						// unused by this
				ky2[currentFrame]		= 0.0;
			}

			if(useRamp) {
				wide1[currentFrame]		= rectDistance * Math.pow((double)fr / halfFrameDuration, 4.0);
				wide2[currentFrame]		= 0.0;		// unused in this figure
				wide3[currentFrame]		= (double)startLayer;	// unused in this figure
				wide4[currentFrame]		= (double)endLayer;		// unused in this figure
			} else {
				wide1[currentFrame]		= rectDistance;			// Distance between Rectangles in Pixels
				wide2[currentFrame]		= 0.0;		// unused in this figure
				wide3[currentFrame]		= (double)startLayer;	// unused in this figure
				wide4[currentFrame]		= (double)endLayer;		// unused in this figure
			}

			pulse[currentFrame]			= 1.0;	// its a multiplicator
			reverse[currentFrame]		= reversePaint;	// set as called
			dominance[currentFrame]		= false;	// set all as called
			domLayerNr[currentFrame]	= 0;
			domStrength[currentFrame]	= 1.0;
			domFalloff[currentFrame]	= 1.0;
			directionX[currentFrame]	= 1;
			directionY[currentFrame]	= mdirectionY;
			mode[currentFrame]			= false;

			currentFrame++;
		}
		for(int fr = 0;fr<halfFrameDuration;fr++) {		// Second Half Ease Out
			shapeType[currentFrame]	= shapeTypeP; 			// Rendered by SimpleCircleInOut
			if(useRamp) {
				// *** Ease functions by Robert Penner
				// *** http://www.robertpenner.com/easing/
				//      complete distance in pixels  Where we are t    Where have we to go t    easeStrength  // = easeIn
				//easeXUp = halfDistXUp * Math.pow( (double)fr       /   halfFrameDuration,         4.0);
				kx1[currentFrame]		= startX;	// is static in this figure
				ky1[currentFrame]		= startY-(movingDistance/2)-((double)fr*singleStep);	// scroll it out complete
				//ky1[currentFrame]		= startY-(originalHeight)-(originalHeight * (1-Math.pow(1-((double)fr / halfFrameDuration), 4.0)));
				kx2[currentFrame]		= 0.0;		// unused by circle
				ky2[currentFrame]		= 0.0;
			} else {
				kx1[currentFrame]		= startX;
				ky1[currentFrame]		= startY-(movingDistance/2)-((double)fr*singleStep);	// scroll it out complete
				kx2[currentFrame]		= 0.0;						// unused by this
				ky2[currentFrame]		= 0.0;
			}

			if(useRamp) {
				wide1[currentFrame]		= rectDistance;			// Distance between Rectangles in Pixels
				wide2[currentFrame]		= 0.0;		// unused in this figure
				wide3[currentFrame]		= (double)startLayer;	// unused in this figure
				wide4[currentFrame]		= (double)endLayer;		// unused in this figure
			} else {
				wide1[currentFrame]		= rectDistance;			// Distance between Rectangles in Pixels
				wide2[currentFrame]		= 0.0;		// unused in this figure
				wide3[currentFrame]		= (double)startLayer;	// unused in this figure
				wide4[currentFrame]		= (double)endLayer;		// unused in this figure
			}

			pulse[currentFrame]			= 1.0;	// its a multiplicator
			reverse[currentFrame]		= reversePaint;	// set as called
			dominance[currentFrame]		= false;	// set all as called
			domLayerNr[currentFrame]	= 0;
			domStrength[currentFrame]	= 1.0;
			domFalloff[currentFrame]	= 1.0;
			directionX[currentFrame]	= 1;
			directionY[currentFrame]	= mdirectionY;
			mode[currentFrame]			= true;

			currentFrame++;
		}
	}
/*
Sample call:
calcSimpleInOutCircle(4, 1009, 4.0, 2.0, 2.0,		// shape type, start frame, duration in stay out
							480, 360,				// centered coordinates
							0, 100, 0,				// three wide parameter in mid out
							1.0, 1.0,				// pulse strength and speed
							false, true, true,		// reverse, ramp, straight grow
							false, 0, 1.0, 1.0)		// dom on, dom layer, strength, falloff
*/
	public void calcSimpleInOutCircle(int shapeTypeP, double inDuration, double stayDuration, double outDuration,
							double startX1, double startY1,
							double swide, double mwide, double ewide,
							double pulseStrength, double pulseSpeed,
							boolean reversePaint, boolean useRamp, boolean straightGrow,
							boolean dominanceOn, int layerDom, double strengthDom, double falloffDom) {
		// duration parameters are in score notation. amount in quarter notes
		int frameDurationIn 	= (int)(framesPerBeat * inDuration)+1;
		int frameDurationStay	= (int)(framesPerBeat * stayDuration)+1;
		int frameDurationOut	= (int)(framesPerBeat * outDuration)+1;

		//*** In (Forward) berechnen
		double halfFrameDuration = frameDurationIn/2;
		double wideStep = (mwide-swide) / frameDurationIn;
		double halfWideDist = wideStep * halfFrameDuration;
		for(int fr = 0;fr<(int)halfFrameDuration;fr++) {	// FIRST HALF  ****************
			shapeType[currentFrame]	= shapeTypeP; 			// Rendered by SimpleCircleInOut
			kx1[currentFrame]		= startX1;	// is static in this figure
			ky1[currentFrame]		= startY1;
			kx2[currentFrame]		= 0.0;		// unused by circle
			ky2[currentFrame]		= 0.0;

			if(useRamp) {
				// *** Ease functions by Robert Penner
				// *** http://www.robertpenner.com/easing/
				//      complete distance in pixels  Where we are t    Where have we to go t    easeStrength  // = easeIn
				//easeXUp = halfDistXUp * Math.pow( (double)fr       /   halfFrameDuration,         4.0);
				wide1[currentFrame]		= swide	+ (halfWideDist * Math.pow((double)fr / halfFrameDuration, 4.0));
				wide2[currentFrame]		= mwide;	// unused in this figure
				wide3[currentFrame]		= 0.0;	// unused in this figure
				wide4[currentFrame]		= 0.0;	// unused in this figure
			} else {
				wide1[currentFrame]		= swide	+ (fr*wideStep);
				wide2[currentFrame]		= mwide;	// unused in this figure
				wide3[currentFrame]		= 0.0;	// unused in this figure
				wide4[currentFrame]		= 0.0;	// unused in this figure
			}

			pulse[currentFrame]		= 1.0;	// its a multiplicator
			reverse[currentFrame]		= reversePaint;	// set as called
			dominance[currentFrame]		= dominanceOn;	// set all as called
			domLayerNr[currentFrame]	= layerDom;
			domStrength[currentFrame]	= strengthDom;
			domFalloff[currentFrame]	= falloffDom;
			directionX[currentFrame]	= 1;
			directionY[currentFrame]	= 1;
			mode[currentFrame]			= straightGrow;

			currentFrame++;
		}
		for(int fr = 0;fr<(int)halfFrameDuration;fr++) {	// SECOND HALF  ****************
			shapeType[currentFrame]	= shapeTypeP; 			// Rendered by SimpleCircleInOut
			kx1[currentFrame]		= startX1;	// is static in this figure
			ky1[currentFrame]		= startY1;
			kx2[currentFrame]		= 0.0;		// unused by circle
			ky2[currentFrame]		= 0.0;

			if(useRamp) {
				// *** Ease functions by Robert Penner
				// *** http://www.robertpenner.com/easing/
				//      complete distance in pixels  Where we are t    Where have we to go t    easeStrength  // = easeIn
				//easeXUp = (completeDistanceXUp/2) * (1-Math.pow( 1-((double)fr    /   halfFrameDuration),         4.0)); //=easeOut
				wide1[currentFrame]		= swide	+ halfWideDist + (halfWideDist * (1-Math.pow( 1-((double)fr / halfFrameDuration), 4.0)));
				wide2[currentFrame]		= mwide;	// fixed value, needed for straightGrow
				wide3[currentFrame]		= 0.0;	// unused in this figure
				wide4[currentFrame]		= 0.0;	// unused in this figure
			} else {
				wide1[currentFrame]		= swide	+ halfWideDist + (fr*wideStep);
				wide2[currentFrame]		= mwide;	// fixed value, needed for straightGrow
				wide3[currentFrame]		= 0.0;	// unused in this figure
				wide4[currentFrame]		= 0.0;	// unused in this figure
			}

			pulse[currentFrame]		= 1.0;	// its a multiplicator
			reverse[currentFrame]		= reversePaint;	// set as called
			dominance[currentFrame]		= dominanceOn;	// set all as called
			domLayerNr[currentFrame]	= layerDom;
			domStrength[currentFrame]	= strengthDom;
			domFalloff[currentFrame]	= falloffDom;
			directionX[currentFrame]	= 1;
			directionY[currentFrame]	= 1;
			mode[currentFrame]			= straightGrow;

			currentFrame++;
		}

		// ******* Mid (Stay) calculation
		halfFrameDuration = frameDurationStay/2;
		for(int fr = 0;fr<(int)halfFrameDuration;fr++) {	// FIRST HALF  ****************
			shapeType[currentFrame]	= shapeTypeP; 			// Rendered by SimpleCircleInOut
			kx1[currentFrame]		= startX1;	// is static in this figure
			ky1[currentFrame]		= startY1;
			kx2[currentFrame]		= 0.0;		// unused by circle
			ky2[currentFrame]		= 0.0;

			if(useRamp) {	// easy easing only adding 1 or 2 pixels
				wide1[currentFrame]		= mwide	+ (fr*1);
				wide2[currentFrame]		= mwide;	// fixed value, needed for straightGrow
				wide3[currentFrame]		= 0.0;	// unused in this figure
				wide4[currentFrame]		= 0.0;	// unused in this figure
			} else {
				wide1[currentFrame]		= mwide;
				wide2[currentFrame]		= mwide;	// fixed value, needed for straightGrow
				wide3[currentFrame]		= 0.0;	// unused in this figure
				wide4[currentFrame]		= 0.0;	// unused in this figure
			}

			pulse[currentFrame]		= 1.0;	// its a multiplicator
			reverse[currentFrame]		= reversePaint;	// set as called
			dominance[currentFrame]		= dominanceOn;	// set all as called
			domLayerNr[currentFrame]	= layerDom;
			domStrength[currentFrame]	= strengthDom;
			domFalloff[currentFrame]	= falloffDom;
			directionX[currentFrame]	= 1;
			directionY[currentFrame]	= 1;
			mode[currentFrame]			= straightGrow;

			currentFrame++;
		}
		for(int fr = 0;fr<(int)halfFrameDuration;fr++) {	// SECOND HALF  ****************
			shapeType[currentFrame]	= shapeTypeP; 			// Rendered by SimpleCircleInOut
			kx1[currentFrame]		= startX1;	// is static in this figure
			ky1[currentFrame]		= startY1;
			kx2[currentFrame]		= 0.0;		// unused by circle
			ky2[currentFrame]		= 0.0;

			if(useRamp) {	// easy easing only adding 1 or 2 pixels
				wide1[currentFrame]		= mwide	+ ((int)halfFrameDuration*1) - (fr*1);
				wide2[currentFrame]		= mwide;	// fixed value, needed for straightGrow
				wide3[currentFrame]		= 0.0;	// unused in this figure
				wide4[currentFrame]		= 0.0;	// unused in this figure
			} else {
				wide1[currentFrame]		= mwide;
				wide2[currentFrame]		= mwide;	// fixed value, needed for straightGrow
				wide3[currentFrame]		= 0.0;	// unused in this figure
				wide4[currentFrame]		= 0.0;	// unused in this figure
			}

			pulse[currentFrame]		= 1.0;	// its a multiplicator
			reverse[currentFrame]		= reversePaint;	// set as called
			dominance[currentFrame]		= dominanceOn;	// set all as called
			domLayerNr[currentFrame]	= layerDom;
			domStrength[currentFrame]	= strengthDom;
			domFalloff[currentFrame]	= falloffDom;
			directionX[currentFrame]	= 1;
			directionY[currentFrame]	= 1;
			mode[currentFrame]			= straightGrow;

			currentFrame++;
		}
		//*** Out (Backward) berechnen
		halfFrameDuration = frameDurationOut/2;
		wideStep = (mwide-ewide) / frameDurationOut;
		halfWideDist = wideStep * halfFrameDuration;
		for(int fr = 0;fr<(int)halfFrameDuration;fr++) {	// FIRST HALF  ****************
			shapeType[currentFrame]	= shapeTypeP; 			// Rendered by SimpleCircleInOut
			kx1[currentFrame]		= startX1;	// is static in this figure
			ky1[currentFrame]		= startY1;
			kx2[currentFrame]		= 0.0;		// unused by circle
			ky2[currentFrame]		= 0.0;

			if(useRamp) {
				// *** Ease functions by Robert Penner
				// *** http://www.robertpenner.com/easing/
				//      complete distance in pixels  Where we are t    Where have we to go t    easeStrength  // = easeIn
				//easeXUp = halfDistXUp * Math.pow( (double)fr       /   halfFrameDuration,         4.0);
				wide1[currentFrame]		= mwide	- (halfWideDist * Math.pow((double)fr / halfFrameDuration, 4.0));
				wide2[currentFrame]		= mwide;	// fixed value, needed for straightGrow
				wide3[currentFrame]		= 0.0;	// unused in this figure
				wide4[currentFrame]		= 0.0;	// unused in this figure
			} else {
				wide1[currentFrame]		= mwide	- (fr*wideStep);
				wide2[currentFrame]		= mwide;	// fixed value, needed for straightGrow
				wide3[currentFrame]		= 0.0;	// unused in this figure
				wide4[currentFrame]		= 0.0;	// unused in this figure
			}

			pulse[currentFrame]		= 1.0;	// its a multiplicator
			reverse[currentFrame]		= reversePaint;	// set as called
			dominance[currentFrame]		= dominanceOn;	// set all as called
			domLayerNr[currentFrame]	= layerDom;
			domStrength[currentFrame]	= strengthDom;
			domFalloff[currentFrame]	= falloffDom;
			directionX[currentFrame]	= 1;
			directionY[currentFrame]	= 1;
			mode[currentFrame]			= straightGrow;

			currentFrame++;
		}
		//double halfFrameDuration = frameDurationIn/2;
		for(int fr = 0;fr<(int)halfFrameDuration;fr++) {	// SECOND HALF  ****************
			shapeType[currentFrame]	= shapeTypeP; 			// Rendered by SimpleCircleInOut
			kx1[currentFrame]		= startX1;	// is static in this figure
			ky1[currentFrame]		= startY1;
			kx2[currentFrame]		= 0.0;		// unused by circle
			ky2[currentFrame]		= 0.0;

			if(useRamp) {
				// *** Ease functions by Robert Penner
				// *** http://www.robertpenner.com/easing/
				//      complete distance in pixels  Where we are t    Where have we to go t    easeStrength  // = easeIn
				//easeXUp = (completeDistanceXUp/2) * (1-Math.pow( 1-((double)fr    /   halfFrameDuration),         4.0)); //=easeOut
				wide1[currentFrame]		= mwide	- halfWideDist - (halfWideDist * (1-Math.pow( 1-((double)fr / halfFrameDuration), 4.0)));
				wide2[currentFrame]		= mwide;	// fixed value, needed for straightGrow
				wide3[currentFrame]		= 0.0;	// unused in this figure
				wide4[currentFrame]		= 0.0;	// unused in this figure
			} else {
				wide1[currentFrame]		= mwide	- halfWideDist - (fr*wideStep);
				wide2[currentFrame]		= mwide;	// fixed value, needed for straightGrow
				wide3[currentFrame]		= 0.0;	// unused in this figure
				wide4[currentFrame]		= 0.0;	// unused in this figure
			}

			pulse[currentFrame]		= 1.0;	// its a multiplicator
			reverse[currentFrame]		= reversePaint;	// set as called
			dominance[currentFrame]		= dominanceOn;	// set all as called
			domLayerNr[currentFrame]	= layerDom;
			domStrength[currentFrame]	= strengthDom;
			domFalloff[currentFrame]	= falloffDom;
			directionX[currentFrame]	= 1;
			directionY[currentFrame]	= 1;
			mode[currentFrame]			= straightGrow;

			currentFrame++;
		}
	}
/*
Sample Call:
		calcMovedGrowCircle(8, 4.0, 4.0, 2.0, 4.0, 8.0,		// shape type, duration in stay1 move stay2 out
						450.0, 360.0,					// centered start coordinates
						300.0, -280.0,				// centered end coordinates
						0, 200, 300, 200, 0,		// five wide parameter in stay move stay out
						1.0, 1.0,					// pulse strength and speed
						true, true,					// reverse, ramp
						false, 7, 0.5, 1.0);		// dom on, dom layer, strength, falloff
*/
	public void calcMovedGrowCircle(int shapeTypeP, double inDuration, double stayDuration,
							double moveDuration, double stay2Duration,  double outDuration,
							double startX1, double startY1,
							double endX1, double endY1,
							double startWide, double stayWide, double moveWide,
							double stay2Wide, double endWide,
							double pulseStrength, double pulseSpeed,
							boolean reversePaint, boolean useRamp,
							boolean dominanceOn, int layerDom, double strengthDom, double falloffDom) {
		// duration parameters are in score notation. amount in quarter notes
		int frameDurationIn 	= (int)(framesPerBeat * inDuration)+1;
		int frameDurationStay	= (int)(framesPerBeat * stayDuration)+1;
		int frameDurationMove 	= (int)(framesPerBeat * moveDuration)+1;
		int frameDurationStay2 	= (int)(framesPerBeat * stay2Duration)+1;
		int frameDurationOut	= (int)(framesPerBeat * outDuration)+1;

		//*** In (Forward) berechnen
		double halfFrameDuration = frameDurationIn/2;
		double wideStep = (stayWide-startWide) / frameDurationIn;
		double halfWideDist = wideStep * halfFrameDuration;
		for(int fr = 0;fr<(int)halfFrameDuration;fr++) {	// FIRST HALF  ****************
			shapeType[currentFrame]	= shapeTypeP; 			// Rendered by SimpleCircleInOut
			kx1[currentFrame]		= startX1;
			ky1[currentFrame]		= startY1;
			kx2[currentFrame]		= endX1;		// unused by circle
			ky2[currentFrame]		= endY1;

			if(useRamp) {
				// *** Ease functions by Robert Penner
				// *** http://www.robertpenner.com/easing/
				//      complete distance in pixels  Where we are t    Where have we to go t    easeStrength  // = easeIn
				//easeXUp = halfDistXUp * Math.pow( (double)fr       /   halfFrameDuration,         4.0);
				wide1[currentFrame]		= startWide	+ (halfWideDist * Math.pow((double)fr / halfFrameDuration, 4.0));
				wide2[currentFrame]		= stayWide;	// unused in this figure
				wide3[currentFrame]		= 0.0;	// unused in this figure
				wide4[currentFrame]		= 0.0;	// unused in this figure
			} else {
				wide1[currentFrame]		= startWide	+ (fr*wideStep);
				wide2[currentFrame]		= stayWide;	// unused in this figure
				wide3[currentFrame]		= 0.0;	// unused in this figure
				wide4[currentFrame]		= 0.0;	// unused in this figure
			}

			pulse[currentFrame]		= 1.0;	// its a multiplicator
			reverse[currentFrame]		= reversePaint;	// set as called
			dominance[currentFrame]		= dominanceOn;	// set all as called
			domLayerNr[currentFrame]	= layerDom;
			domStrength[currentFrame]	= strengthDom;
			domFalloff[currentFrame]	= falloffDom;
			directionX[currentFrame]	= 1;
			directionY[currentFrame]	= 1;
			mode[currentFrame]			= false;

			currentFrame++;
		}
		//double halfFrameDuration = frameDurationIn/2;
		for(int fr = 0;fr<(int)halfFrameDuration;fr++) {	// SECOND HALF  ****************
			shapeType[currentFrame]	= shapeTypeP; 			// Rendered by SimpleCircleInOut
			kx1[currentFrame]		= startX1;	// is static in this figure
			ky1[currentFrame]		= startY1;
			kx2[currentFrame]		= endX1;		// unused by circle
			ky2[currentFrame]		= endY1;

			if(useRamp) {
				// *** Ease functions by Robert Penner
				// *** http://www.robertpenner.com/easing/
				//      complete distance in pixels  Where we are t    Where have we to go t    easeStrength  // = easeIn
				//easeXUp = (completeDistanceXUp/2) * (1-Math.pow( 1-((double)fr    /   halfFrameDuration),         4.0)); //=easeOut
				wide1[currentFrame]		= startWide	+ halfWideDist + (halfWideDist * (1-Math.pow( 1-((double)fr / halfFrameDuration), 4.0)));
				wide2[currentFrame]		= stayWide;	// fixed value, needed for straightGrow
				wide3[currentFrame]		= 0.0;	// unused in this figure
				wide4[currentFrame]		= 0.0;	// unused in this figure
			} else {
				wide1[currentFrame]		= startWide	+ halfWideDist + (fr*wideStep);
				wide2[currentFrame]		= stayWide;	// fixed value, needed for straightGrow
				wide3[currentFrame]		= 0.0;	// unused in this figure
				wide4[currentFrame]		= 0.0;	// unused in this figure
			}

			pulse[currentFrame]		= 1.0;	// its a multiplicator
			reverse[currentFrame]		= reversePaint;	// set as called
			dominance[currentFrame]		= dominanceOn;	// set all as called
			domLayerNr[currentFrame]	= layerDom;
			domStrength[currentFrame]	= strengthDom;
			domFalloff[currentFrame]	= falloffDom;
			directionX[currentFrame]	= 1;
			directionY[currentFrame]	= 1;
			mode[currentFrame]			= false;

			currentFrame++;
		}

		// ******* First Stay calculation
		halfFrameDuration = frameDurationStay/2;
		wideStep = (moveWide-stayWide) / frameDurationStay;
		halfWideDist = wideStep * halfFrameDuration;
		for(int fr = 0;fr<(int)halfFrameDuration;fr++) {	// FIRST HALF  ****************
			shapeType[currentFrame]	= shapeTypeP; 			// Rendered by SimpleCircleInOut
			kx1[currentFrame]		= startX1;	// is static in this figure
			ky1[currentFrame]		= startY1;
			kx2[currentFrame]		= endX1;		// unused by circle
			ky2[currentFrame]		= endY1;

			if(useRamp) {	// easy easing only adding 1 or 2 pixels
				wide1[currentFrame]		= stayWide	+ (halfWideDist * Math.pow((double)fr / halfFrameDuration, 4.0));
				wide2[currentFrame]		= stayWide;	// unused in this figure
				wide3[currentFrame]		= 0.0;	// unused in this figure
				wide4[currentFrame]		= 0.0;	// unused in this figure
			} else {
				wide1[currentFrame]		= stayWide	+ (fr*wideStep);
				wide2[currentFrame]		= stayWide;	// unused in this figure
				wide3[currentFrame]		= 0.0;	// unused in this figure
				wide4[currentFrame]		= 0.0;	// unused in this figure
			}

			pulse[currentFrame]		= 1.0;	// its a multiplicator
			reverse[currentFrame]		= reversePaint;	// set as called
			dominance[currentFrame]		= dominanceOn;	// set all as called
			domLayerNr[currentFrame]	= layerDom;
			domStrength[currentFrame]	= strengthDom;
			domFalloff[currentFrame]	= falloffDom;
			directionX[currentFrame]	= 1;
			directionY[currentFrame]	= 1;
			mode[currentFrame]			= false;

			currentFrame++;
		}
		for(int fr = 0;fr<(int)halfFrameDuration;fr++) {	// SECOND HALF  ****************
			shapeType[currentFrame]	= shapeTypeP; 			// Rendered by SimpleCircleInOut
			kx1[currentFrame]		= startX1;	// is static in this figure
			ky1[currentFrame]		= startY1;
			kx2[currentFrame]		= 0.0;		// unused by circle
			ky2[currentFrame]		= 0.0;

			if(useRamp) {	// easy easing only adding 1 or 2 pixels
				wide1[currentFrame]		= stayWide	+ halfWideDist + (halfWideDist * (1-Math.pow( 1-((double)fr / halfFrameDuration), 4.0)));
				wide2[currentFrame]		= stayWide;	// fixed value, needed for straightGrow
				wide3[currentFrame]		= 0.0;	// unused in this figure
				wide4[currentFrame]		= 0.0;	// unused in this figure
			} else {
				wide1[currentFrame]		= stayWide	+ halfWideDist + (fr*wideStep);
				wide2[currentFrame]		= stayWide;	// fixed value, needed for straightGrow
				wide3[currentFrame]		= 0.0;	// unused in this figure
				wide4[currentFrame]		= 0.0;	// unused in this figure
			}

			pulse[currentFrame]		= 1.0;	// its a multiplicator
			reverse[currentFrame]		= reversePaint;	// set as called
			dominance[currentFrame]		= dominanceOn;	// set all as called
			domLayerNr[currentFrame]	= layerDom;
			domStrength[currentFrame]	= strengthDom;
			domFalloff[currentFrame]	= falloffDom;
			directionX[currentFrame]	= 1;
			directionY[currentFrame]	= 1;
			mode[currentFrame]			= false;

			currentFrame++;
		}
// *************
		// ******* Move calculation
		halfFrameDuration = frameDurationMove/2;

		double compDistX = startX1 - endX1;	// complete distance of movement
		double compDistY = startY1 - endY1;

		double movStepX = (compDistX/frameDurationMove);
		double movStepY = (compDistY/frameDurationMove);

		wideStep = (stay2Wide-moveWide) / frameDurationMove;
		halfWideDist = wideStep * halfFrameDuration;

		for(int fr = 0;fr<(int)halfFrameDuration;fr++) {	// FIRST HALF  ****************
			shapeType[currentFrame]	= shapeTypeP; 			// Rendered by SimpleCircleInOut

			if(useRamp) {	// easy easing only adding 1 or 2 pixels
				kx1[currentFrame]		= startX1 - ((compDistX/2) * Math.pow((double)fr / halfFrameDuration, 4.0));
				ky1[currentFrame]		= startY1 - ((compDistY/2) * Math.pow((double)fr / halfFrameDuration, 4.0));
				kx2[currentFrame]		= endX1;		// unused by circle
				ky2[currentFrame]		= endY1;

				wide1[currentFrame]		= moveWide	+ (halfWideDist * Math.pow((double)fr / halfFrameDuration, 4.0));
				wide2[currentFrame]		= moveWide;	// unused in this figure
				wide3[currentFrame]		= 0.0;	// unused in this figure
				wide4[currentFrame]		= 0.0;	// unused in this figure
			} else {
				kx1[currentFrame]		= startX1-(int)(fr*movStepX);	// is static in this figure
				ky1[currentFrame]		= startY1-(int)(fr*movStepY);
				kx2[currentFrame]		= endX1;		// unused by circle
				ky2[currentFrame]		= endY1;

				wide1[currentFrame]		= moveWide	+ (fr*wideStep);
				wide2[currentFrame]		= moveWide;	// unused in this figure
				wide3[currentFrame]		= 0.0;	// unused in this figure
				wide4[currentFrame]		= 0.0;	// unused in this figure
			}

			pulse[currentFrame]		= 1.0;	// its a multiplicator
			reverse[currentFrame]		= reversePaint;	// set as called
			dominance[currentFrame]		= dominanceOn;	// set all as called
			domLayerNr[currentFrame]	= layerDom;
			domStrength[currentFrame]	= strengthDom;
			domFalloff[currentFrame]	= falloffDom;
			directionX[currentFrame]	= 1;
			directionY[currentFrame]	= 1;
			mode[currentFrame]			= false;

			currentFrame++;
		}
		for(int fr = 0;fr<(int)halfFrameDuration;fr++) {	// SECOND HALF  ****************
			shapeType[currentFrame]	= shapeTypeP; 			// Rendered by SimpleCircleInOut

			if(useRamp) {	// easy easing only adding 1 or 2 pixels
				kx1[currentFrame]		= startX1 - (compDistX/2) - ((compDistX/2) * (1-Math.pow(1-((double)fr / halfFrameDuration), 4.0)));
				ky1[currentFrame]		= startY1 - (compDistY/2) - ((compDistY/2) * (1-Math.pow(1-((double)fr / halfFrameDuration), 4.0)));
				kx2[currentFrame]		= endX1;		// unused by circle
				ky2[currentFrame]		= endY1;

				wide1[currentFrame]		= moveWide	+ halfWideDist + (halfWideDist * (1-Math.pow( 1-((double)fr / halfFrameDuration), 4.0)));
				wide2[currentFrame]		= moveWide;	// fixed value, needed for straightGrow
				wide3[currentFrame]		= 0.0;	// unused in this figure
				wide4[currentFrame]		= 0.0;	// unused in this figure
			} else {
				kx1[currentFrame]		= startX1-(halfFrameDuration*movStepX) -(int)(fr*movStepX);	// is static in this figure
				ky1[currentFrame]		= startY1-(halfFrameDuration*movStepY) -(int)(fr*movStepY);
				kx2[currentFrame]		= endX1;		// unused by circle
				ky2[currentFrame]		= endY1;

				wide1[currentFrame]		= moveWide	+ halfWideDist + (fr*wideStep);
				wide2[currentFrame]		= moveWide;	// fixed value, needed for straightGrow
				wide3[currentFrame]		= 0.0;	// unused in this figure
				wide4[currentFrame]		= 0.0;	// unused in this figure
			}

			pulse[currentFrame]		= 1.0;	// its a multiplicator
			reverse[currentFrame]		= reversePaint;	// set as called
			dominance[currentFrame]		= dominanceOn;	// set all as called
			domLayerNr[currentFrame]	= layerDom;
			domStrength[currentFrame]	= strengthDom;
			domFalloff[currentFrame]	= falloffDom;
			directionX[currentFrame]	= 1;
			directionY[currentFrame]	= 1;
			mode[currentFrame]			= false;

			currentFrame++;
		}
// *************
		// ******* Second Stay calculation
		halfFrameDuration = frameDurationStay2/2;
		for(int fr = 0;fr<(int)halfFrameDuration;fr++) {	// FIRST HALF  ****************
			shapeType[currentFrame]	= shapeTypeP; 			// Rendered by SimpleCircleInOut
			kx1[currentFrame]		= endX1;	// is static in this figure
			ky1[currentFrame]		= endY1;
			kx2[currentFrame]		= endX1;		// unused by circle
			ky2[currentFrame]		= endY1;

			if(useRamp) {	// easy easing only adding 1 or 2 pixels
				wide1[currentFrame]		= stay2Wide	+ (fr*1);
				wide2[currentFrame]		= stay2Wide;	// fixed value, needed for straightGrow
				wide3[currentFrame]		= 0.0;	// unused in this figure
				wide4[currentFrame]		= 0.0;	// unused in this figure
			} else {
				wide1[currentFrame]		= stay2Wide;
				wide2[currentFrame]		= stay2Wide;	// fixed value, needed for straightGrow
				wide3[currentFrame]		= 0.0;	// unused in this figure
				wide4[currentFrame]		= 0.0;	// unused in this figure
			}

			pulse[currentFrame]		= 1.0;	// its a multiplicator
			reverse[currentFrame]		= reversePaint;	// set as called
			dominance[currentFrame]		= dominanceOn;	// set all as called
			domLayerNr[currentFrame]	= layerDom;
			domStrength[currentFrame]	= strengthDom;
			domFalloff[currentFrame]	= falloffDom;
			directionX[currentFrame]	= 1;
			directionY[currentFrame]	= 1;
			mode[currentFrame]			= false;

			currentFrame++;
		}
		for(int fr = 0;fr<(int)halfFrameDuration;fr++) {	// SECOND HALF  ****************
			shapeType[currentFrame]	= shapeTypeP; 			// Rendered by SimpleCircleInOut
			kx1[currentFrame]		= endX1;	// is static in this figure
			ky1[currentFrame]		= endY1;
			kx2[currentFrame]		= 0.0;		// unused by circle
			ky2[currentFrame]		= 0.0;

			if(useRamp) {	// easy easing only adding 1 or 2 pixels
				wide1[currentFrame]		= stay2Wide	+ ((int)halfFrameDuration*1) - (fr*1);
				wide2[currentFrame]		= stay2Wide;	// fixed value, needed for straightGrow
				wide3[currentFrame]		= 0.0;	// unused in this figure
				wide4[currentFrame]		= 0.0;	// unused in this figure
			} else {
				wide1[currentFrame]		= stay2Wide;
				wide2[currentFrame]		= stay2Wide;	// fixed value, needed for straightGrow
				wide3[currentFrame]		= 0.0;	// unused in this figure
				wide4[currentFrame]		= 0.0;	// unused in this figure
			}

			pulse[currentFrame]		= 1.0;	// its a multiplicator
			reverse[currentFrame]		= reversePaint;	// set as called
			dominance[currentFrame]		= dominanceOn;	// set all as called
			domLayerNr[currentFrame]	= layerDom;
			domStrength[currentFrame]	= strengthDom;
			domFalloff[currentFrame]	= falloffDom;
			directionX[currentFrame]	= 1;
			directionY[currentFrame]	= 1;
			mode[currentFrame]			= false;

			currentFrame++;
		}

		//*** Out (Backward) berechnen
		halfFrameDuration = frameDurationOut/2;
		wideStep = (stay2Wide-endWide) / frameDurationOut;
		halfWideDist = wideStep * halfFrameDuration;
		for(int fr = 0;fr<(int)halfFrameDuration;fr++) {	// FIRST HALF  ****************
			shapeType[currentFrame]	= shapeTypeP; 			// Rendered by SimpleCircleInOut
			if(useRamp) {
				// *** Ease functions by Robert Penner
				// *** http://www.robertpenner.com/easing/
				//      complete distance in pixels  Where we are t    Where have we to go t    easeStrength  // = easeIn
				//easeXUp = halfDistXUp * Math.pow( (double)fr       /   halfFrameDuration,         4.0);
				kx1[currentFrame]		= endX1;	// is static in this figure
				ky1[currentFrame]		= endY1;
				kx2[currentFrame]		= 0.0;		// unused by circle
				ky2[currentFrame]		= 0.0;

				wide1[currentFrame]		= stay2Wide	- (halfWideDist * Math.pow((double)fr / halfFrameDuration, 4.0));
				wide2[currentFrame]		= stay2Wide;	// fixed value, needed for straightGrow
				wide3[currentFrame]		= 0.0;	// unused in this figure
				wide4[currentFrame]		= 0.0;	// unused in this figure
			} else {
				kx1[currentFrame]		= endX1;	// is static in this figure
				ky1[currentFrame]		= endY1;
				kx2[currentFrame]		= 0.0;		// unused by circle
				ky2[currentFrame]		= 0.0;

				wide1[currentFrame]		= stay2Wide	- (fr*wideStep);
				wide2[currentFrame]		= stay2Wide;	// fixed value, needed for straightGrow
				wide3[currentFrame]		= 0.0;	// unused in this figure
				wide4[currentFrame]		= 0.0;	// unused in this figure
			}

			pulse[currentFrame]		= 1.0;	// its a multiplicator
			reverse[currentFrame]		= reversePaint;	// set as called
			dominance[currentFrame]		= dominanceOn;	// set all as called
			domLayerNr[currentFrame]	= layerDom;
			domStrength[currentFrame]	= strengthDom;
			domFalloff[currentFrame]	= falloffDom;
			directionX[currentFrame]	= 1;
			directionY[currentFrame]	= 1;
			mode[currentFrame]			= false;

			currentFrame++;
		}
		for(int fr = 0;fr<(int)halfFrameDuration;fr++) {	// SECOND HALF  ****************
			shapeType[currentFrame]	= shapeTypeP; 			// Rendered by SimpleCircleInOut

			if(useRamp) {
				// *** Ease functions by Robert Penner
				// *** http://www.robertpenner.com/easing/
				//      complete distance in pixels  Where we are t    Where have we to go t    easeStrength  // = easeIn
				//easeXUp = (completeDistanceXUp/2) * (1-Math.pow( 1-((double)fr    /   halfFrameDuration),         4.0)); //=easeOut
				kx1[currentFrame]		= endX1;	// is static in this figure
				ky1[currentFrame]		= endY1;
				kx2[currentFrame]		= 0.0;		// unused by circle
				ky2[currentFrame]		= 0.0;

				wide1[currentFrame]		= stay2Wide	- halfWideDist - (halfWideDist * (1-Math.pow( 1-((double)fr / halfFrameDuration), 4.0)));
				wide2[currentFrame]		= stay2Wide;	// fixed value, needed for straightGrow
				wide3[currentFrame]		= 0.0;	// unused in this figure
				wide4[currentFrame]		= 0.0;	// unused in this figure
			} else {
				kx1[currentFrame]		= endX1;	// is static in this figure
				ky1[currentFrame]		= endY1;
				kx2[currentFrame]		= 0.0;		// unused by circle
				ky2[currentFrame]		= 0.0;

				wide1[currentFrame]		= stay2Wide	- halfWideDist - (fr*wideStep);
				wide2[currentFrame]		= stay2Wide;	// fixed value, needed for straightGrow
				wide3[currentFrame]		= 0.0;	// unused in this figure
				wide4[currentFrame]		= 0.0;	// unused in this figure
			}

			pulse[currentFrame]		= 1.0;	// its a multiplicator
			reverse[currentFrame]		= reversePaint;	// set as called
			dominance[currentFrame]		= dominanceOn;	// set all as called
			domLayerNr[currentFrame]	= layerDom;
			domStrength[currentFrame]	= strengthDom;
			domFalloff[currentFrame]	= falloffDom;
			directionX[currentFrame]	= 1;
			directionY[currentFrame]	= 1;
			mode[currentFrame]			= false;

			currentFrame++;
		}
	}
/* Sample call:
		calcLayerSpot(10, 8.0, 8.0, 8.0,				// shape type, duration in move out
						8,								// the spottetd layer!
						300.0, 360.0,					// centered start coordinates for circle 1
						100.0, -280.0,					// centered end coordinates for circle 1
						0, 200, 300, 0,					// wide parameter in start, endstart, move, end for circle 1
						1.0, 1.0,						// pulse strength and speed
						false, true);					// reverse, ramp
*/
	public void calcLayerSpot(int shapeTypeP, double startDuration, double moveDuration, double endDuration,
							int theSpotLayer,
							double startX1, double startY1,
							double endX1, double endY1,
							double startWide, double endOfstartWide, double moveWide, double endWide,
							double pulseStrength, double pulseSpeed,
							boolean reversePaint, boolean useRamp) {
		// duration parameters are in score notation. amount in quarter notes
		int frameDurationStart 	= (int)(framesPerBeat * startDuration)+1;
		int frameDurationMove 	= (int)(framesPerBeat * moveDuration)+1;
		int frameDurationEnd	= (int)(framesPerBeat * endDuration)+1;
		int frameDurationAll = frameDurationStart+frameDurationMove+frameDurationEnd;

		double halfFrameDuration = 1.0;
		double distanceX1 = 1.0;
		double distanceY1 = 1.0;
		double distanceStepX1 = 1.0;
		double distanceStepY1 = 1.0;
		double offsetX1 = 1.0;
		double offsetY1 = 1.0;
		double offWide = 0.0;
		double distWide1 = 1.0;
		double dWideStep1 = 1.0;

		int savedCurrentFrame = currentFrame;

		for(int phase = 0;phase<3;phase++) {
			if(phase==0) {	// move from start coordinate to move coordinate. Maybe grow
				halfFrameDuration = frameDurationStart/2;	// calc parameter for start to stay 1 phase
				distanceX1 = 0;
				distanceY1 = 0;
				distanceStepX1 = 0;
				distanceStepY1 = 0;
				offsetX1 = startX1;
				offsetY1 = startY1;
				offWide = startWide;
				distWide1 = endOfstartWide - startWide;
				dWideStep1 = distWide1/(double)frameDurationStart;
			} else if(phase==1) {	// move from stay 1 coordinate to stay 2 coordinate. Maybe grow
				halfFrameDuration = frameDurationMove/2;	// calc parameter for move from stay1 to stay 2 phase
				distanceX1 = endX1 - startX1;
				distanceY1 = endY1 - startY1;
				distanceStepX1 = distanceX1/(double)frameDurationMove;
				distanceStepY1 = distanceY1/(double)frameDurationMove;
				offsetX1 = startX1;
				offsetY1 = startY1;
				offWide = endOfstartWide;
				distWide1 = moveWide - endOfstartWide;
				dWideStep1 = distWide1/(double)frameDurationMove;
			} else if(phase==2) {	// move from stay2 to end coordinate. maybe grow or shrink
				halfFrameDuration = frameDurationEnd/2;
				distanceX1 = 0;
				distanceY1 = 0;
				distanceStepX1 = 0;
				distanceStepY1 = 0;
				offsetX1 = endX1;
				offsetY1 = endY1;
				offWide = moveWide;
				distWide1 = endWide - moveWide;
				dWideStep1 = distWide1/(double)frameDurationEnd;
			}
			// ****** Now the real move calculations.
			// First half of movement
			for(int fr = 0;fr<halfFrameDuration;fr++) {
				if(useRamp) {
					kx1[currentFrame]		= offsetX1 + ((distanceX1/2) * Math.pow((double)fr / halfFrameDuration, 4.0));
					ky1[currentFrame]		= offsetY1 + ((distanceY1/2) * Math.pow((double)fr / halfFrameDuration, 4.0));
					wide1[currentFrame]		= offWide  + ((distWide1/2) * Math.pow((double)fr / halfFrameDuration, 4.0));
				} else {
					kx1[currentFrame]		= offsetX1 + ((double)fr*distanceStepX1);
					ky1[currentFrame]		= offsetY1 + ((double)fr*distanceStepY1);
					wide1[currentFrame]		= offWide + ((double)fr*dWideStep1);
				}
				currentFrame++;	// problem with this. must be resetet in the next
			}
			// Second half of movement
			for(int fr = 0;fr<halfFrameDuration;fr++) {
				if(useRamp) {
					kx1[currentFrame]		= offsetX1 + (distanceX1/2) + ((distanceX1/2) * (1-Math.pow(1-((double)fr / halfFrameDuration), 4.0)));
					ky1[currentFrame]		= offsetY1 + (distanceY1/2) + ((distanceY1/2) * (1-Math.pow(1-((double)fr / halfFrameDuration), 4.0)));
					wide1[currentFrame]		= offWide  + (distWide1/2)  + ((distWide1/2)  * (1-Math.pow(1-((double)fr / halfFrameDuration), 4.0)));
				} else {
					kx1[currentFrame]		= offsetX1 + (halfFrameDuration*distanceStepX1) + ((double)fr*distanceStepX1);
					ky1[currentFrame]		= offsetY1 + (halfFrameDuration*distanceStepY1) + ((double)fr*distanceStepY1);
					wide1[currentFrame]		= offWide + (halfFrameDuration*dWideStep1) + ((double)fr*dWideStep1);
				}
				currentFrame++;	// problem with this. must be resetet in the next
			}
		}

		currentFrame = savedCurrentFrame;
		// Static values in this figure
		for(int fr = 0;fr<frameDurationAll;fr++) {
			shapeType[currentFrame]	= shapeTypeP;
			kx2[currentFrame]		= 0.0;	// unused in this figure
			ky2[currentFrame]		= 0.0;	// unused in this figure
			wide2[currentFrame]		= 0.0;	// unused in this figure
			wide3[currentFrame]		= 0.0;	// unused in this figure
			wide4[currentFrame]		= 0.0;	// unused in this figure

			pulse[currentFrame]		= 1.0;	// its a multiplicator
			reverse[currentFrame]		= reversePaint;	// set as called
			dominance[currentFrame]		= false;	// set all as called
			domLayerNr[currentFrame]	= theSpotLayer;
			domStrength[currentFrame]	= 0.0;
			domFalloff[currentFrame]	= 0.0;
			directionX[currentFrame]	= 1;
			directionY[currentFrame]	= 1;
			mode[currentFrame]			= false;
			currentFrame++;
		}
	}

/* Sample call:
		calcDualCircle(9, 8.0, 8.0, 8.0, 8.0, 8.0,		// shape type, duration in stay1 move stay2 out
						1000.0, 360.0,					// centered start coordinates for circle 1
						450.0, 360.0,					// centered stay1 coordinates for circle 1
						300.0, -280.0,					// centered stay2 coordinates for circle 1
						1000.0, -280.0,					// centered end coordinates for circle 1
						0, 200, 300, 200, 150, 0,		// six wide parameter in start, endOfstartWide, stay1, move, stay2, end for circle 1
						-1000.0, 360.0,					// centered start coordinates for circle 2
						-450.0, 360.0,					// centered stay1 coordinates for circle 2
						-300.0, -280.0,					// centered stay2 coordinates for circle 2
						-1000.0, -280.0,				// centered end coordinates for circle 2
						0, 200, 300, 200, 150, 0,		// six wide parameter in start, endOfstartWide, stay1, move, stay2, end for circle 2
						1.0, 1.0,						// pulse strength and speed
						false, false, true,				// reverse, ramp, Synchronity
						false, 7, 0.5, 1.0);			// dom on, dom layer, strength, falloff
*/
	public void calcDualCircle(int shapeTypeP, double startDuration, double stay1Duration,
							double moveDuration, double stay2Duration,  double endDuration,
							double startX1, double startY1,
							double stay1X1, double stay1Y1,
							double stay2X1, double stay2Y1,
							double endX1, double endY1,
							double startWide, double endOfstartWide, double stay1Wide,
							double moveWide, double stay2Wide, double endWide,
							double startX2, double startY2,
							double stay1X2, double stay1Y2,
							double stay2X2, double stay2Y2,
							double endX2, double endY2,
							double startWide2, double endOfstartWide2, double stay1Wide2,
							double moveWide2, double stay2Wide2, double endWide2,
							double pulseStrength, double pulseSpeed,
							boolean reversePaint, boolean useRamp, boolean synchronity,
							boolean dominanceOn, int layerDom, double strengthDom, double falloffDom) {
		// duration parameters are in score notation. amount in quarter notes
		int frameDurationStart 	= (int)(framesPerBeat * startDuration)+1;
		int frameDurationStay1	= (int)(framesPerBeat * stay1Duration)+1;
		int frameDurationMove 	= (int)(framesPerBeat * moveDuration)+1;
		int frameDurationStay2 	= (int)(framesPerBeat * stay2Duration)+1;
		int frameDurationEnd	= (int)(framesPerBeat * endDuration)+1;
		int frameDurationAll = frameDurationStart+frameDurationStay1+frameDurationMove+frameDurationStay2+frameDurationEnd;

		double halfFrameDuration = 1.0;
		double distanceX1 = 1.0;
		double distanceY1 = 1.0;
		double distanceX2 = 1.0;
		double distanceY2 = 1.0;
		double distanceStepX1 = 1.0;
		double distanceStepY1 = 1.0;
		double distanceStepX2 = 1.0;
		double distanceStepY2 = 1.0;
		double offsetX1 = 1.0;
		double offsetY1 = 1.0;
		double offsetX2 = 1.0;
		double offsetY2 = 1.0;
		double offWide = 0.0;
		double offWide2 = 0.0;
		double distWide1 = 1.0;
		double distWide2 = 1.0;
		double dWideStep1 = 1.0;
		double dWideStep2 = 1.0;

		int savedCurrentFrame = currentFrame;

		for(int phase = 0;phase<5;phase++) {
			if(phase==0) {	// move from start coordinate to stay 1 coordinate. Maybe grow
				halfFrameDuration = (double)frameDurationStart/2;	// calc parameter for start to stay 1 phase
				distanceX1 = stay1X1 - startX1;
				distanceY1 = stay1Y1 - startY1;
				distanceX2 = stay1X2 - startX2;
				distanceY2 = stay1Y2 - startY2;
				distanceStepX1 = distanceX1/(double)frameDurationStart;
				distanceStepY1 = distanceY1/(double)frameDurationStart;
				distanceStepX2 = distanceX2/(double)frameDurationStart;
				distanceStepY2 = distanceY2/(double)frameDurationStart;
				offsetX1 = startX1;
				offsetY1 = startY1;
				offsetX2 = startX2;
				offsetY2 = startY2;
				offWide = startWide;
				offWide2 = startWide2;
				distWide1 = endOfstartWide - startWide;
				distWide2 = endOfstartWide2 - startWide2;
				dWideStep1 = distWide1/(double)frameDurationStart;
				dWideStep2 = distWide2/(double)frameDurationStart;
			} else if(phase==1) {	// stay at stay 1 coordinate. Maybe grow.
				halfFrameDuration = (double)frameDurationStay1/2;	// calc parameter for stay1 to end of stay1 phase
				distanceX1 = 0.0;
				distanceY1 = 0.0;
				distanceX2 = 0.0;
				distanceY2 = 0.0;
				distanceStepX1 = 0.0;
				distanceStepY1 = 0.0;
				distanceStepX2 = 0.0;
				distanceStepY2 = 0.0;
				offsetX1 = stay1X1;
				offsetY1 = stay1Y1;
				offsetX2 = stay1X2;
				offsetY2 = stay1Y2;
				offWide = endOfstartWide;
				offWide2 = endOfstartWide2;
				distWide1 = stay1Wide - endOfstartWide;
				distWide2 = stay1Wide2 - endOfstartWide2;
				dWideStep1 = distWide1/(double)frameDurationStay1;
				dWideStep2 = distWide2/(double)frameDurationStay1;
			} else if(phase==2) {	// move from stay 1 coordinate to stay 2 coordinate. Maybe grow
				halfFrameDuration = (double)frameDurationMove/2;	// calc parameter for move from stay1 to stay 2 phase
				distanceX1 = stay2X1 - stay1X1;
				distanceY1 = stay2Y1 - stay1Y1;
				distanceX2 = stay2X2 - stay1X2;
				distanceY2 = stay2Y2 - stay1Y2;
				distanceStepX1 = distanceX1/(double)frameDurationMove;
				distanceStepY1 = distanceY1/(double)frameDurationMove;
				distanceStepX2 = distanceX2/(double)frameDurationMove;
				distanceStepY2 = distanceY2/(double)frameDurationMove;
				offsetX1 = stay1X1;
				offsetY1 = stay1Y1;
				offsetX2 = stay1X2;
				offsetY2 = stay1Y2;
				offWide = stay1Wide;
				offWide2 = stay1Wide2;
				distWide1 = moveWide - stay1Wide;
				distWide2 = moveWide2 - stay1Wide2;
				dWideStep1 = distWide1/(double)frameDurationMove;
				dWideStep2 = distWide2/(double)frameDurationMove;
			} else if(phase==3) {	// stay at stay 2 coordinate. maybe grow
				halfFrameDuration = (double)frameDurationStay2/2;	// calc parameter for stay2 to end of stay2 phase
				distanceX1 = 0.0;
				distanceY1 = 0.0;
				distanceX2 = 0.0;
				distanceY2 = 0.0;
				distanceStepX1 = 0.0;
				distanceStepY1 = 0.0;
				distanceStepX2 = 0.0;
				distanceStepY2 = 0.0;
				offsetX1 = stay2X1;
				offsetY1 = stay2Y1;
				offsetX2 = stay2X2;
				offsetY2 = stay2Y2;
				offWide = moveWide;
				offWide2 = moveWide2;
				distWide1 = stay2Wide - moveWide;
				distWide2 = stay2Wide2 - moveWide2;
				dWideStep1 = distWide1/(double)frameDurationMove;
				dWideStep2 = distWide2/(double)frameDurationMove;
			} else if(phase==4) {	// move from stay2 to end coordinate. maybe grow or shrink
				halfFrameDuration = (double)frameDurationEnd/2;
				distanceX1 = endX1 - stay2X1;
				distanceY1 = endY1 - stay2Y1;
				distanceX2 = endX2 - stay2X2;
				distanceY2 = endY2 - stay2Y2;
				distanceStepX1 = distanceX1/(double)frameDurationEnd;
				distanceStepY1 = distanceY1/(double)frameDurationEnd;
				distanceStepX2 = distanceX2/(double)frameDurationEnd;
				distanceStepY2 = distanceY2/(double)frameDurationEnd;
				offsetX1 = stay2X1;
				offsetY1 = stay2Y1;
				offsetX2 = stay2X2;
				offsetY2 = stay2Y2;
				offWide = stay2Wide;
				offWide2 = stay2Wide2;
				distWide1 = endWide - stay2Wide;
				distWide2 = endWide2 - stay2Wide2;
				dWideStep1 = distWide1/(double)frameDurationMove;
				dWideStep2 = distWide2/(double)frameDurationMove;
			}
			// ****** Now the real move calculations.
			if(synchronity) {
				// First half of movement
				for(int fr = 0;fr<halfFrameDuration;fr++) {
					if(useRamp) {
						kx1[currentFrame]		= offsetX1 + ((distanceX1/2) * Math.pow((double)fr / halfFrameDuration, 4.0));
						ky1[currentFrame]		= offsetY1 + ((distanceY1/2) * Math.pow((double)fr / halfFrameDuration, 4.0));
						kx2[currentFrame]		= offsetX2 + ((distanceX2/2) * Math.pow((double)fr / halfFrameDuration, 4.0));
						ky2[currentFrame]		= offsetY2 + ((distanceY2/2) * Math.pow((double)fr / halfFrameDuration, 4.0));
						wide1[currentFrame]		= offWide  + ((distWide1/2) * Math.pow((double)fr / halfFrameDuration, 4.0));
						wide2[currentFrame]		= offWide2 + ((distWide2/2) * Math.pow((double)fr / halfFrameDuration, 4.0));
					} else {
						kx1[currentFrame]		= offsetX1 + ((double)fr*distanceStepX1);
						ky1[currentFrame]		= offsetY1 + ((double)fr*distanceStepY1);
						kx2[currentFrame]		= offsetX2 + ((double)fr*distanceStepX2);
						ky2[currentFrame]		= offsetY2 + ((double)fr*distanceStepY2);
						wide1[currentFrame]		= offWide + ((double)fr*dWideStep1);
						wide2[currentFrame]		= offWide2 + ((double)fr*dWideStep2);
					}
					currentFrame++;	// problem with this. must be resetet in the next
				}
				// Second half of movement
				for(int fr = 0;fr<halfFrameDuration;fr++) {
					if(useRamp) {
						kx1[currentFrame]		= offsetX1 + (distanceX1/2) + ((distanceX1/2) * (1-Math.pow(1-((double)fr / halfFrameDuration), 4.0)));
						ky1[currentFrame]		= offsetY1 + (distanceY1/2) + ((distanceY1/2) * (1-Math.pow(1-((double)fr / halfFrameDuration), 4.0)));
						kx2[currentFrame]		= offsetX2 + (distanceX2/2) + ((distanceX2/2) * (1-Math.pow(1-((double)fr / halfFrameDuration), 4.0)));
						ky2[currentFrame]		= offsetY2 + (distanceY2/2) + ((distanceY2/2) * (1-Math.pow(1-((double)fr / halfFrameDuration), 4.0)));
						wide1[currentFrame]		= offWide  + (distWide1/2)  + ((distWide1/2)  * (1-Math.pow(1-((double)fr / halfFrameDuration), 4.0)));
						wide2[currentFrame]		= offWide2 + (distWide2/2)  + ((distWide2/2)  * (1-Math.pow(1-((double)fr / halfFrameDuration), 4.0)));
					} else {
						kx1[currentFrame]		= offsetX1 + (halfFrameDuration*distanceStepX1) + ((double)fr*distanceStepX1);
						ky1[currentFrame]		= offsetY1 + (halfFrameDuration*distanceStepY1) + ((double)fr*distanceStepY1);
						kx2[currentFrame]		= offsetX2 + (halfFrameDuration*distanceStepX2) + ((double)fr*distanceStepX2);
						ky2[currentFrame]		= offsetY2 + (halfFrameDuration*distanceStepY2) + ((double)fr*distanceStepY2);
						wide1[currentFrame]		= offWide + (halfFrameDuration*dWideStep1) + ((double)fr*dWideStep1);
						wide2[currentFrame]		= offWide2 + (halfFrameDuration*dWideStep2) + ((double)fr*dWideStep2);
					}
					currentFrame++;	// problem with this. must be resetted in the next
				}
			} else {
				// ***** FirstHalf of Synchronity *******
				for(int fr = 0;fr<(halfFrameDuration/2);fr++) {
					if(useRamp) {
						kx1[currentFrame]		= offsetX1 + ((distanceX1/2) * Math.pow((double)fr / (halfFrameDuration/2), 4.0));
						ky1[currentFrame]		= offsetY1 + ((distanceY1/2) * Math.pow((double)fr / (halfFrameDuration/2), 4.0));
						kx2[currentFrame]		= offsetX2;
						ky2[currentFrame]		= offsetY2;
						wide1[currentFrame]		= offWide  + ((distWide1/2) * Math.pow((double)fr / (halfFrameDuration/2), 4.0));
						wide2[currentFrame]		= offWide2;
					} else {
						kx1[currentFrame]		= offsetX1 + ((double)fr*(distanceStepX1*2));
						ky1[currentFrame]		= offsetY1 + ((double)fr*(distanceStepY1*2));
						kx2[currentFrame]		= offsetX2;
						ky2[currentFrame]		= offsetY2;
						wide1[currentFrame]		= offWide  + ((double)fr*(dWideStep1*2));
						wide2[currentFrame]		= offWide2;
					}
					currentFrame++;	// problem with this. must be resetet in the next
				}
				// Second half of movement
				for(int fr = 0;fr<(halfFrameDuration/2);fr++) {
					shapeType[currentFrame]	= shapeTypeP;
					if(useRamp) {
						kx1[currentFrame]		= offsetX1 + (distanceX1/2) + ((distanceX1/2) * (1-Math.pow(1-((double)fr / (halfFrameDuration/2)), 4.0)));
						ky1[currentFrame]		= offsetY1 + (distanceY1/2) + ((distanceY1/2) * (1-Math.pow(1-((double)fr / (halfFrameDuration/2)), 4.0)));
						kx2[currentFrame]		= offsetX2;
						ky2[currentFrame]		= offsetY2;
						wide1[currentFrame]		= offWide  + (distWide1/2)  + ((distWide1/2)  * (1-Math.pow(1-((double)fr / (halfFrameDuration/2)), 4.0)));
						wide2[currentFrame]		= offWide2;
					} else {
						kx1[currentFrame]		= offsetX1 + ((halfFrameDuration/2)*(distanceStepX1*2)) + ((double)fr*distanceStepX1*2);
						ky1[currentFrame]		= offsetY1 + ((halfFrameDuration/2)*(distanceStepY1*2)) + ((double)fr*distanceStepY1*2);
						kx2[currentFrame]		= offsetX2;
						ky2[currentFrame]		= offsetY2;
						wide1[currentFrame]		= offWide  + ((halfFrameDuration/2)*(dWideStep1*2))     + ((double)fr*dWideStep1*2);
						wide2[currentFrame]		= offWide2;
					}
					currentFrame++;	// problem with this. must be resetet in the next
				}
				// ***** Second of Synchronity *******
				for(int fr = 0;fr<(halfFrameDuration/2);fr++) {
					if(useRamp) {
						kx1[currentFrame]		= offsetX1 + distanceX1;
						ky1[currentFrame]		= offsetY1 + distanceY1;
						kx2[currentFrame]		= offsetX2 + ((distanceX2/2) * Math.pow((double)fr / (halfFrameDuration/2), 4.0));
						ky2[currentFrame]		= offsetY2 + ((distanceY2/2) * Math.pow((double)fr / (halfFrameDuration/2), 4.0));
						wide1[currentFrame]		= offWide  + distWide1;
						wide2[currentFrame]		= offWide2 + ((distWide2/2) * Math.pow((double)fr / (halfFrameDuration/2), 4.0));
					} else {
						kx1[currentFrame]		= offsetX1 + distanceX1;
						ky1[currentFrame]		= offsetY1 + distanceY1;
						kx2[currentFrame]		= offsetX2 + ((double)fr*(distanceStepX2*2));
						ky2[currentFrame]		= offsetY2 + ((double)fr*(distanceStepY2*2));
						wide1[currentFrame]		= offWide  + distWide1;
						wide2[currentFrame]		= offWide2 + ((double)fr*(dWideStep2*2));
					}
					currentFrame++;	// problem with this. must be resetet in the next
				}
				// Second half of movement
				for(int fr = 0;fr<(halfFrameDuration/2);fr++) {
					shapeType[currentFrame]	= shapeTypeP;
					if(useRamp) {
						kx1[currentFrame]		= offsetX1 + distanceX1;
						ky1[currentFrame]		= offsetY1 + distanceY1;
						kx2[currentFrame]		= offsetX2 + (distanceX2/2) + ((distanceX2/2) * (1-Math.pow(1-((double)fr / (halfFrameDuration/2)), 4.0)));
						ky2[currentFrame]		= offsetY2 + (distanceY2/2) + ((distanceY2/2) * (1-Math.pow(1-((double)fr / (halfFrameDuration/2)), 4.0)));
						wide1[currentFrame]		= offWide  + distWide1;
						wide2[currentFrame]		= offWide2 + (distWide2/2)  + ((distWide2/2)  * (1-Math.pow(1-((double)fr / (halfFrameDuration/2)), 4.0)));
					} else {
						kx1[currentFrame]		= offsetX1 + distanceX1;
						ky1[currentFrame]		= offsetY1 + distanceY1;
						kx2[currentFrame]		= offsetX2 + ((halfFrameDuration/2)*(distanceStepX2*2)) + ((double)fr*distanceStepX2*2);
						ky2[currentFrame]		= offsetY2 + ((halfFrameDuration/2)*(distanceStepY2*2)) + ((double)fr*distanceStepY2*2);
						wide1[currentFrame]		= offWide  + distWide1;
						wide2[currentFrame]		= offWide2 + ((halfFrameDuration/2)*(dWideStep2*2))     + ((double)fr*dWideStep2*2);
					}
					currentFrame++;	// problem with this. must be resetet in the next
				}
			}
		}

		currentFrame = savedCurrentFrame;
		// Static values in this figure
		for(int fr = 0;fr<frameDurationAll;fr++) {
			shapeType[currentFrame]	= shapeTypeP;

			wide3[currentFrame]		= 0.0;	// unused in this figure
			wide4[currentFrame]		= 0.0;	// unused in this figure

			pulse[currentFrame]		= 1.0;	// its a multiplicator
			reverse[currentFrame]		= reversePaint;	// set as called
			dominance[currentFrame]		= dominanceOn;	// set all as called
			domLayerNr[currentFrame]	= layerDom;
			domStrength[currentFrame]	= strengthDom;
			domFalloff[currentFrame]	= falloffDom;
			directionX[currentFrame]	= 1;
			directionY[currentFrame]	= 1;
			mode[currentFrame]			= synchronity;
			currentFrame++;
		}
	}

/* Sample call:
		calcPerspectiveRays(11, 1, 8.0, 8.0, 8.0, 8.0, 8.0,		// shape type, orientation(1=horizontal) duration in stay1 move stay2 out
						-3395.0, -160.0,				// vanishing point coordinates
						1000.0, -1500.0,					// centered start coordinates for Point 2
						1000.0, -75.0,					// centered stay1 coordinates for Point 2
						1000.0, 280.0,					// centered stay2 coordinates for Point 2
						-1000.0, -1500.0,				// centered end coordinates for Point 2
						0, 200, 300, 200, 150, 0,		// six upper wide parameter in start, endOfstartWide, stay1, move, stay2, end for circle 1
						0, 200, 300, 200, 150, 0,		// six lower wide parameter in start, endOfstartWide, stay1, move, stay2, end for Point 2
						1.0, 1.0,						// pulse strength and speed
						false, false,					// reverse, ramp,
						1, 1);								// direction
*/
	public void calcPerspectiveRays(int shapeTypeP, int orientation, double startDuration, double stay1Duration,
							double moveDuration, double stay2Duration,  double endDuration,
							double startX1, double startY1,
							double startX2, double startY2,
							double stay1X2, double stay1Y2,
							double stay2X2, double stay2Y2,
							double endX2, double endY2,
							double startWide, double endOfstartWide, double stay1Wide,
							double moveWide, double stay2Wide, double endWide,
							double startWide2, double endOfstartWide2, double stay1Wide2,
							double moveWide2, double stay2Wide2, double endWide2,
							double pulseStrength, double pulseSpeed,
							boolean reversePaint, boolean useRamp,
							int mdirectionX, int mdirectionY) {
		// duration parameters are in score notation. amount in quarter notes
		int frameDurationStart 	= (int)(framesPerBeat * startDuration)+1;
		int frameDurationStay1	= (int)(framesPerBeat * stay1Duration)+1;
		int frameDurationMove 	= (int)(framesPerBeat * moveDuration)+1;
		int frameDurationStay2 	= (int)(framesPerBeat * stay2Duration)+1;
		int frameDurationEnd	= (int)(framesPerBeat * endDuration)+1;
		int frameDurationAll = frameDurationStart+frameDurationStay1+frameDurationMove+frameDurationStay2+frameDurationEnd;
/*System.out.println("Dura Start:"+frameDurationStart);
System.out.println("Dura Stay1:"+frameDurationStay1);
System.out.println("Dura Move:"+frameDurationMove);
System.out.println("Dura Stay2:"+frameDurationStay2);
System.out.println("Dura End:"+frameDurationEnd);
System.out.println("Dura All:"+frameDurationAll);*/
		double halfFrameDuration = 1.0;
		double distanceX2 = 1.0;
		double distanceY2 = 1.0;
		double distanceStepX2 = 1.0;
		double distanceStepY2 = 1.0;
		double offsetX2 = 1.0;
		double offsetY2 = 1.0;
		double offWide = 0.0;
		double offWide2 = 0.0;
		double distWide1 = 1.0;
		double distWide2 = 1.0;
		double dWideStep1 = 1.0;
		double dWideStep2 = 1.0;

		int savedCurrentFrame = currentFrame;

		for(int phase = 0;phase<5;phase++) {
			if(phase==0) {	// move from start coordinate to stay 1 coordinate. Maybe grow
				halfFrameDuration = (double)frameDurationStart/2;	// calc parameter for start to stay 1 phase
				distanceX2 = stay1X2 - startX2;
				distanceY2 = stay1Y2 - startY2;
				distanceStepX2 = distanceX2/(double)frameDurationStart;
				distanceStepY2 = distanceY2/(double)frameDurationStart;
				offsetX2 = startX2;
				offsetY2 = startY2;
				offWide = startWide;
				offWide2 = startWide2;
				distWide1 = endOfstartWide - startWide;
				distWide2 = endOfstartWide2 - startWide2;
				dWideStep1 = distWide1/(double)frameDurationStart;
				dWideStep2 = distWide2/(double)frameDurationStart;
			} else if(phase==1) {	// stay at stay 1 coordinate. Maybe grow.
				halfFrameDuration = (double)frameDurationStay1/2;	// calc parameter for stay1 to end of stay1 phase
				distanceX2 = 0.0;
				distanceY2 = 0.0;
				distanceStepX2 = 0.0;
				distanceStepY2 = 0.0;
				offsetX2 = stay1X2;
				offsetY2 = stay1Y2;
				offWide = endOfstartWide;
				offWide2 = endOfstartWide2;
				distWide1 = stay1Wide - endOfstartWide;
				distWide2 = stay1Wide2 - endOfstartWide2;
				dWideStep1 = distWide1/(double)frameDurationStay1;
				dWideStep2 = distWide2/(double)frameDurationStay1;
			} else if(phase==2) {	// move from stay 1 coordinate to stay 2 coordinate. Maybe grow
				halfFrameDuration = (double)frameDurationMove/2;	// calc parameter for move from stay1 to stay 2 phase
				distanceX2 = stay2X2 - stay1X2;
				distanceY2 = stay2Y2 - stay1Y2;
				distanceStepX2 = distanceX2/(double)frameDurationMove;
				distanceStepY2 = distanceY2/(double)frameDurationMove;
				offsetX2 = stay1X2;
				offsetY2 = stay1Y2;
				offWide = stay1Wide;
				offWide2 = stay1Wide2;
				distWide1 = moveWide - stay1Wide;
				distWide2 = moveWide2 - stay1Wide2;
				dWideStep1 = distWide1/(double)frameDurationMove;
				dWideStep2 = distWide2/(double)frameDurationMove;
			} else if(phase==3) {	// stay at stay 2 coordinate. maybe grow
				halfFrameDuration = (double)frameDurationStay2/2;	// calc parameter for stay2 to end of stay2 phase
				distanceX2 = 0.0;
				distanceY2 = 0.0;
				distanceStepX2 = 0.0;
				distanceStepY2 = 0.0;
				offsetX2 = stay2X2;
				offsetY2 = stay2Y2;
				offWide = moveWide;
				offWide2 = moveWide2;
				distWide1 = stay2Wide - moveWide;
				distWide2 = stay2Wide2 - moveWide2;
				dWideStep1 = distWide1/(double)frameDurationMove;
				dWideStep2 = distWide2/(double)frameDurationMove;
			} else if(phase==4) {	// move from stay2 to end coordinate. maybe grow or shrink
				halfFrameDuration = (double)frameDurationEnd/2;
				distanceX2 = endX2 - stay2X2;
				distanceY2 = endY2 - stay2Y2;
				distanceStepX2 = distanceX2/(double)frameDurationEnd;
				distanceStepY2 = distanceY2/(double)frameDurationEnd;
				offsetX2 = stay2X2;
				offsetY2 = stay2Y2;
				offWide = stay2Wide;
				offWide2 = stay2Wide2;
				distWide1 = endWide - stay2Wide;
				distWide2 = endWide2 - stay2Wide2;
				dWideStep1 = distWide1/(double)frameDurationMove;
				dWideStep2 = distWide2/(double)frameDurationMove;
			}
		//System.out.println("Dura Half:"+halfFrameDuration);
			// ****** Now the real move calculations.
			// First half of movement
			for(int fr = 0;fr<halfFrameDuration;fr++) {
				if(useRamp) {
					kx2[currentFrame]		= offsetX2 + ((distanceX2/2) * Math.pow((double)fr / halfFrameDuration, 4.0));
					ky2[currentFrame]		= offsetY2 + ((distanceY2/2) * Math.pow((double)fr / halfFrameDuration, 4.0));
					wide1[currentFrame]		= offWide  + ((distWide1/2) * Math.pow((double)fr / halfFrameDuration, 4.0));
					wide2[currentFrame]		= offWide2 + ((distWide2/2) * Math.pow((double)fr / halfFrameDuration, 4.0));
				} else {
					kx2[currentFrame]		= offsetX2 + ((double)fr*distanceStepX2);
					ky2[currentFrame]		= offsetY2 + ((double)fr*distanceStepY2);
					wide1[currentFrame]		= offWide + ((double)fr*dWideStep1);
					wide2[currentFrame]		= offWide2 + ((double)fr*dWideStep2);
				}
				currentFrame++;	// problem with this. must be resetet in the next
			}
			// Second half of movement
			for(int fr = 0;fr<halfFrameDuration;fr++) {
				if(useRamp) {
					kx2[currentFrame]		= offsetX2 + (distanceX2/2) + ((distanceX2/2) * (1-Math.pow(1-((double)fr / halfFrameDuration), 4.0)));
					ky2[currentFrame]		= offsetY2 + (distanceY2/2) + ((distanceY2/2) * (1-Math.pow(1-((double)fr / halfFrameDuration), 4.0)));
					wide1[currentFrame]		= offWide  + (distWide1/2)  + ((distWide1/2)  * (1-Math.pow(1-((double)fr / halfFrameDuration), 4.0)));
					//System.out.println("fr: "+currentFrame+"Wide1Calc: "+wide1[currentFrame]);
					wide2[currentFrame]		= offWide2 + (distWide2/2)  + ((distWide2/2)  * (1-Math.pow(1-((double)fr / halfFrameDuration), 4.0)));
				} else {
					kx2[currentFrame]		= offsetX2 + (halfFrameDuration*distanceStepX2) + ((double)fr*distanceStepX2);
					ky2[currentFrame]		= offsetY2 + (halfFrameDuration*distanceStepY2) + ((double)fr*distanceStepY2);
					wide1[currentFrame]		= offWide  + (halfFrameDuration*dWideStep1) + ((double)fr*dWideStep1);
					//System.out.println("fr: "+currentFrame+"Wide1Calc: "+wide1[currentFrame]);
					wide2[currentFrame]		= offWide2 + (halfFrameDuration*dWideStep2) + ((double)fr*dWideStep2);
				}
				currentFrame++;	// problem with this. must be resetted in the next
			}
		}

		currentFrame = savedCurrentFrame;
		// Static values in this figure
		for(int fr = 0;fr<frameDurationAll;fr++) {
			shapeType[currentFrame]	= shapeTypeP;

			kx1[currentFrame]		= startX1;
			ky1[currentFrame]		= startY1;

			wide3[currentFrame]		= (double)orientation;	// is it a horizontal or vertical figure
			wide4[currentFrame]		= 0.0;	// unused in this figure

			pulse[currentFrame]		= 1.0;	// its a multiplicator
			reverse[currentFrame]		= reversePaint;	// set as called
			dominance[currentFrame]		= false;	// set all as called
			domLayerNr[currentFrame]	= 0;
			domStrength[currentFrame]	= 0.0;
			domFalloff[currentFrame]	= 0.0;
			directionX[currentFrame]	= mdirectionX;
			directionY[currentFrame]	= mdirectionY;
			mode[currentFrame]			= false;
			currentFrame++;
		}
	}

/* Sample call:
		calcInOutRays( 1, 83, 8.0, 4.0, 4.0,
						1340, 600, 1000, -600, 10,  2,
						-800, 600, 1000, -600, 300, 10,
						1340, 600, 1000, -600, 10,  2,
						1, 1, 0.5, 1,
						false, true,
						true, 8, 2.0, 0);
// try to use coordinates for movement from right to left. If you need other direction set the direction parameter
*/
	public void calcInOutRays(int shapeTypeP, double inDuration, double stayDuration, double outDuration,
							double startX1, double startY1, double startX2, double startY2, double swideUp, double swideDown,
							double midX1, double midY1, double midX2, double midY2, double mwideUp, double mwideDown,
							double endX1, double endY1, double endX2, double endY2, double ewideUp, double ewideDown,
							int mdirectionX, int mdirectionY, double pulseStrength, double pulseSpeed,
							boolean reversePaint, boolean useRamp,
							boolean dominanceOn, int layerDom, double strengthDom, double falloffDom) {	// Rendered by PolyRect
		// try to use coordinates for movement from right to left. If you need other direction set the direction parameter
		// duration parameters are in score notation. amount in quarter notes
		double frameDurationIn 		= (framesPerBeat * inDuration)+1;		// NEW VERSION FOR CENTERED COORDINATES, AND MORE PARAMETERS
		double frameDurationStay	= (framesPerBeat * stayDuration)+1;
		double frameDurationOut		= (framesPerBeat * outDuration)+1;

		// ******* In (Forward) calculation
		double halfFrameDuration = frameDurationIn/2;

		double forwardStepUpX = (Math.abs(startX1) + Math.abs(midX1)) / frameDurationIn;
		double forwardStepUpY = (Math.abs(startY1) - Math.abs(midY1)) / frameDurationIn;
		double forwardStepDownX = (Math.abs(startX2) - Math.abs(midX2)) / frameDurationIn;
		double forwardStepDownY = (Math.abs(startY2) - Math.abs(midY2)) / frameDurationIn;
		double wideStepUp = (mwideUp-swideUp) / frameDurationIn;
		double wideStepDown = (mwideDown-swideDown) / frameDurationIn;

		double halfDistXUp = forwardStepUpX * halfFrameDuration;
		double halfDistYUp = forwardStepUpY * halfFrameDuration;
		double halfDistXDown = forwardStepDownX * halfFrameDuration;
		double halfDistYDown = forwardStepDownY * halfFrameDuration;
		double halfWideUp = wideStepUp * halfFrameDuration;
		double halfWideDown = wideStepDown * halfFrameDuration;

		for(int fr = 0;fr<(int)halfFrameDuration;fr++) {	// FIRST HALF  ****************
			shapeType[currentFrame]	= shapeTypeP; 			// Rendered by PolyRect - PolyRect can be used by more than one calc
			if(useRamp) {
				// *** Ease functions by Robert Penner
				// *** http://www.robertpenner.com/easing/
				//      complete distance in pixels  Where we are t    Where have we to go t    easeStrength  // = easeIn
				//easeXUp = halfDistXUp * Math.pow( (double)fr       /   halfFrameDuration,         4.0);
				kx1[currentFrame]		= startX1 - (halfDistXUp * Math.pow((double)fr / halfFrameDuration, 4.0));
				ky1[currentFrame]		= startY1 - (halfDistYUp * Math.pow((double)fr / halfFrameDuration, 4.0));
				kx2[currentFrame]		= startX2 - (halfDistXDown * Math.pow((double)fr / halfFrameDuration, 4.0));
				ky2[currentFrame]		= startY2 - (halfDistYDown * Math.pow((double)fr / halfFrameDuration, 4.0));
			} else {
				kx1[currentFrame]		= startX1 - (fr*forwardStepUpX);
				ky1[currentFrame]		= startY1 - (fr*forwardStepUpY);
				kx2[currentFrame]		= startX2 - (fr*forwardStepDownX);
				ky2[currentFrame]		= startY2 - (fr*forwardStepDownY);
			}

			if(useRamp) {
				wide1[currentFrame]		= swideUp	+ (halfWideUp * Math.pow((double)fr / halfFrameDuration, 4.0));
				wide2[currentFrame]		= swideDown	+ (halfWideDown * Math.pow((double)fr / halfFrameDuration, 4.0));
				wide3[currentFrame]		= 0.0;	// unused in PolyRectRenderer ?
				wide4[currentFrame]		= 0.0;	// unused in PolyRectRenderer ?
			} else {
				wide1[currentFrame]		= swideUp	+ (fr*wideStepUp);
				wide2[currentFrame]		= swideDown	+ (fr*wideStepDown);
				wide3[currentFrame]		= 0.0;	// unused in PolyRectRenderer ?
				wide4[currentFrame]		= 0.0;	// unused in PolyRectRenderer ?
			}

			pulse[currentFrame]		= 1.0;	// its a multiplicator

			reverse[currentFrame]		= reversePaint;	// set as called
			dominance[currentFrame]		= dominanceOn;	// set all as called
			domLayerNr[currentFrame]	= layerDom;
			domStrength[currentFrame]	= strengthDom;
			domFalloff[currentFrame]	= falloffDom;
			directionX[currentFrame]	= mdirectionX;
			directionY[currentFrame]	= mdirectionY;
			mode[currentFrame]			= false;

			currentFrame++;
		}

		for(int fr = 0;fr<(int)halfFrameDuration;fr++) {	// SECOND HALF ****************
			shapeType[currentFrame]	= shapeTypeP; // Rendered by PolyRect - PolyRect can be used by more than one calc
			if(useRamp) {
				//      complete distance in pixels         Where we are t    Where have we to go t    easeStrength
				//easeXUp = (completeDistanceXUp/2) * (1-Math.pow( 1-((double)fr    /   halfFrameDuration),         4.0)); //=easeOut
				kx1[currentFrame]		= startX1 - halfDistXUp - (halfDistXUp * (1-Math.pow( 1-((double)fr / halfFrameDuration), 4.0)));
				ky1[currentFrame]		= startY1 - halfDistYUp - (halfDistYUp * (1-Math.pow( 1-((double)fr / halfFrameDuration), 4.0)));
				kx2[currentFrame]		= startX2 - halfDistXDown - (halfDistXDown * (1-Math.pow( 1-((double)fr / halfFrameDuration), 4.0)));
				ky2[currentFrame]		= startY2 - halfDistYDown - (halfDistYDown * (1-Math.pow( 1-((double)fr / halfFrameDuration), 4.0)));
			} else {
				kx1[currentFrame]		= startX1 - halfDistXUp - (fr*forwardStepUpX);
				ky1[currentFrame]		= startY1 - halfDistYUp - (fr*forwardStepUpY);
				kx2[currentFrame]		= startX2 - halfDistXDown - (fr*forwardStepDownX);
				ky2[currentFrame]		= startY2 - halfDistYDown - (fr*forwardStepDownY);
			}

			if(useRamp) {
				wide1[currentFrame]		= swideUp	+ halfWideUp 	+ (halfWideUp * Math.pow((double)fr / halfFrameDuration, 4.0));
				wide2[currentFrame]		= swideDown	+ halfWideDown	+ (halfWideDown * Math.pow((double)fr / halfFrameDuration, 4.0));
				wide3[currentFrame]		= 0.0;	// unused in PolyRectRenderer ?
				wide4[currentFrame]		= 0.0;	// unused in PolyRectRenderer ?
			} else {
				wide1[currentFrame]		= halfWideUp 	+ swideUp	+ (fr*wideStepUp);
				wide2[currentFrame]		= halfWideDown	+ swideDown	+ (fr*wideStepDown);
				wide3[currentFrame]		= 0.0;	// unused in PolyRectRenderer ?
				wide4[currentFrame]		= 0.0;	// unused in PolyRectRenderer ?
			}

			pulse[currentFrame]		= 1.0;	// its a multiplicator

			reverse[currentFrame]		= reversePaint;	// set as called
			dominance[currentFrame]		= dominanceOn;	// set all as called
			domLayerNr[currentFrame]	= layerDom;
			domStrength[currentFrame]	= strengthDom;
			domFalloff[currentFrame]	= falloffDom;
			directionX[currentFrame]	= mdirectionX;
			directionY[currentFrame]	= mdirectionY;
			mode[currentFrame]			= false;

			currentFrame++;
		}

		// ******* Mid (Stay) calculation

		double sinStep = Math.PI / frameDurationStay;
		double wideRamp = 0;
		for(int fr = 0;fr<(int)frameDurationStay;fr++) {
			shapeType[currentFrame]	= shapeTypeP; // Rendered by PolyRect - PolyRect can be used by more than one calc

			kx1[currentFrame]		= midX1;
			ky1[currentFrame]		= midY1;
			kx2[currentFrame]		= midX2;
			ky2[currentFrame]		= midY2;

			if(fr<(int)(frameDurationStay/2)) {
				wideRamp++;
			} else {
				wideRamp--;
			}

			if(useRamp) {
				wide1[currentFrame]		= mwideUp + wideRamp;
				wide2[currentFrame]		= mwideDown;
				wide3[currentFrame]		= 0.0;	// unused in PolyRectRenderer ?
				wide4[currentFrame]		= 0.0;	// unused in PolyRectRenderer ?
			} else {
				wide1[currentFrame]		= mwideUp;
				wide2[currentFrame]		= mwideDown;
				wide3[currentFrame]		= 0.0;	// unused in PolyRectRenderer ?
				wide4[currentFrame]		= 0.0;	// unused in PolyRectRenderer ?
			}

			pulse[currentFrame]		= 1.0;	// its a multiplicator

			reverse[currentFrame]		= reversePaint;	// set as called
			dominance[currentFrame]		= dominanceOn;	// set all as called
			domLayerNr[currentFrame]	= layerDom;
			domStrength[currentFrame]	= strengthDom;
			domFalloff[currentFrame]	= falloffDom;
			directionX[currentFrame]	= mdirectionX;
			directionY[currentFrame]	= mdirectionY;
			mode[currentFrame]			= false;

			currentFrame++;
		}

		// ******* Out (Backward) calculation
		halfFrameDuration = frameDurationOut/2;

		double backwardStepUpX = (Math.abs(startX1) + Math.abs(midX1)) / frameDurationOut;
		double backwardStepUpY = (Math.abs(startY1) - Math.abs(midY1)) / frameDurationOut;
		double backwardStepDownX = (Math.abs(startX2) - Math.abs(midX2)) / frameDurationOut;
		double backwardStepDownY = (Math.abs(startY2) - Math.abs(midY2)) / frameDurationOut;
		wideStepUp = (mwideUp-ewideUp) / frameDurationOut;
		wideStepDown = (mwideDown-ewideDown) / frameDurationOut;

		halfDistXUp = backwardStepUpX * halfFrameDuration;
		halfDistYUp = backwardStepUpY * halfFrameDuration;
		halfDistXDown = backwardStepDownX * halfFrameDuration;
		halfDistYDown = backwardStepDownY * halfFrameDuration;
		halfWideUp = wideStepUp * halfFrameDuration;
		halfWideDown = wideStepDown * halfFrameDuration;

		for(int fr = 0;fr<(int)halfFrameDuration;fr++) {	// FIRST HALF ****************
			shapeType[currentFrame]	= shapeTypeP; 			// Rendered by PolyRect - PolyRect can be used by more than one calc
			if(useRamp) {
				// *** Ease functions by Robert Penner
				// *** http://www.robertpenner.com/easing/
				//      complete distance in pixels  Where we are t    Where have we to go t    easeStrength  // = easeIn
				//easeXUp = halfDistXUp * Math.pow( (double)fr       /   halfFrameDuration,         4.0);
				kx1[currentFrame]		= midX1 + (halfDistXUp * Math.pow((double)fr / halfFrameDuration, 4.0));
				ky1[currentFrame]		= midY1 + (halfDistYUp * Math.pow((double)fr / halfFrameDuration, 4.0));
				kx2[currentFrame]		= midX2 + (halfDistXDown * Math.pow((double)fr / halfFrameDuration, 4.0));
				ky2[currentFrame]		= midY2 + (halfDistYDown * Math.pow((double)fr / halfFrameDuration, 4.0));
			} else {
				kx1[currentFrame]		= midX1 + (fr*backwardStepUpX);
				ky1[currentFrame]		= midY1 + (fr*backwardStepUpY);
				kx2[currentFrame]		= midX2 + (fr*backwardStepDownX);
				ky2[currentFrame]		= midY2 + (fr*backwardStepDownY);
			}

			if(useRamp) {
				wide1[currentFrame]		= mwideUp	- (halfWideUp * Math.pow((double)fr / halfFrameDuration, 4.0));
				wide2[currentFrame]		= mwideDown	- (halfWideDown * Math.pow((double)fr / halfFrameDuration, 4.0));
				wide3[currentFrame]		= 0.0;	// unused in PolyRectRenderer ?
				wide4[currentFrame]		= 0.0;	// unused in PolyRectRenderer ?
			} else {
				wide1[currentFrame]		= mwideUp	- (fr*wideStepUp);
				wide2[currentFrame]		= mwideDown	- (fr*wideStepDown);
				wide3[currentFrame]		= 0.0;	// unused in PolyRectRenderer ?
				wide4[currentFrame]		= 0.0;	// unused in PolyRectRenderer ?
			}

			pulse[currentFrame]		= 1.0;	// its a multiplicator

			reverse[currentFrame]		= reversePaint;	// set as called
			dominance[currentFrame]		= dominanceOn;	// set all as called
			domLayerNr[currentFrame]	= layerDom;
			domStrength[currentFrame]	= strengthDom;
			domFalloff[currentFrame]	= falloffDom;
			directionX[currentFrame]	= mdirectionX;
			directionY[currentFrame]	= mdirectionY;
			mode[currentFrame]			= false;

			currentFrame++;
		}
		for(int fr = 0;fr<(int)halfFrameDuration;fr++) {	// SECOND HALF  ****************
			shapeType[currentFrame]	= shapeTypeP; // Rendered by PolyRect - PolyRect can be used by more than one calc
			if(useRamp) {
				//      complete distance in pixels         Where we are t    Where have we to go t    easeStrength
				//easeXUp = (completeDistanceXUp/2) * (1-Math.pow( 1-((double)fr    /   halfFrameDuration),         4.0)); //=easeOut
				kx1[currentFrame]		= midX1 + halfDistXUp + (halfDistXUp * (1-Math.pow( 1-((double)fr / halfFrameDuration), 4.0)));
				ky1[currentFrame]		= midY1 + halfDistYUp + (halfDistYUp * (1-Math.pow( 1-((double)fr / halfFrameDuration), 4.0)));
				kx2[currentFrame]		= midX2 + halfDistXDown + (halfDistXDown * (1-Math.pow( 1-((double)fr / halfFrameDuration), 4.0)));
				ky2[currentFrame]		= midY2 + halfDistYDown + (halfDistYDown * (1-Math.pow( 1-((double)fr / halfFrameDuration), 4.0)));
			} else {
				kx1[currentFrame]		= midX1 + halfDistXUp + (fr*backwardStepUpX);
				ky1[currentFrame]		= midY1 + halfDistYUp + (fr*backwardStepUpY);
				kx2[currentFrame]		= midX2 + halfDistXDown + (fr*backwardStepDownX);
				ky2[currentFrame]		= midY2 + halfDistYDown + (fr*backwardStepDownY);
			}

			if(useRamp) {
				wide1[currentFrame]		= mwideUp	- halfWideUp 	- (halfWideUp * Math.pow((double)fr / halfFrameDuration, 4.0));
				wide2[currentFrame]		= mwideDown	- halfWideDown	- (halfWideDown * Math.pow((double)fr / halfFrameDuration, 4.0));
				wide3[currentFrame]		= 0.0;	// unused in PolyRectRenderer ?
				wide4[currentFrame]		= 0.0;	// unused in PolyRectRenderer ?
			} else {
				wide1[currentFrame]		= halfWideUp 	- mwideUp	- (fr*wideStepUp);
				wide2[currentFrame]		= halfWideDown	- mwideDown	- (fr*wideStepDown);
				wide3[currentFrame]		= 0.0;	// unused in PolyRectRenderer ?
				wide4[currentFrame]		= 0.0;	// unused in PolyRectRenderer ?
			}

			pulse[currentFrame]		= 1.0;	// its a multiplicator

			reverse[currentFrame]		= reversePaint;	// set as called
			dominance[currentFrame]		= dominanceOn;	// set all as called
			domLayerNr[currentFrame]	= layerDom;
			domStrength[currentFrame]	= strengthDom;
			domFalloff[currentFrame]	= falloffDom;
			directionX[currentFrame]	= mdirectionX;
			directionY[currentFrame]	= mdirectionY;
			mode[currentFrame]			= false;

			currentFrame++;
		}
	}	// *** calc in out rays end

	public void calcSimpleRect(double duration, boolean moveIt, double startx, double starty, double startw, double starth) {	// Rendered by SimpleRect
		int frameDuration 	= (int)(framesPerBeat * duration);		// NEW VERSION FOR CENTERED COORDINATES, AND MORE PARAMETERS

		// ******* calculation
		double forwardStep  = (originalWidth / 6) / frameDuration;
		double forwardStepY = (originalHeight / 6) / frameDuration;
		double startX1 =  startx;
		double startX2 = startw;	// is x2 not really width
		double startY1 =  starty;
		double startY2 = starth; // is y2 not really height
		for(int fr = 0;fr<frameDuration;fr++) {
			shapeType[currentFrame]	= 99; // Rendered by SimpleRect
			kx1[currentFrame]		= startX1 - (forwardStep*fr);	// testweise mal den Sinus nur auf den Step berechnet
			ky1[currentFrame]		= startY1 + (forwardStepY*fr);
			kx2[currentFrame]		= startX2 + (forwardStep*fr);
			ky2[currentFrame]		= startY2 - (forwardStepY*fr);

			wide1[currentFrame]		= 10.0;
			wide2[currentFrame]		= 10.0;
			wide3[currentFrame]		= 0.0;	// unused in PolyRectRenderer for Ray In Out
			wide4[currentFrame]		= 0.0;	// unused in PolyRectRenderer ?

			pulse[currentFrame]		= 1.0;	// its a multiplicator

			reverse[currentFrame]		= false;	// set as called
			dominance[currentFrame]		= false;	// set all as called
			domLayerNr[currentFrame]	= 0;
			domStrength[currentFrame]	= 1.0;
			domFalloff[currentFrame]	= 1.0;
			directionX[currentFrame]	= 1;
			directionY[currentFrame]	= 1;
			mode[currentFrame]			= false;

			currentFrame++;
		}
	}
	// *** Mask Rendering and Creating Methods ***************************************************************
	public void maskMakerEmpty(int frameNr) {		// Dont save any masks (do nothing)
		// *** shapeType = 0 !!!!
		printMaskCreateFeedback(frameNr, "No mask saved in frame: "+frameNr);
	}

	public void maskMakerWaitEmpty(int frameNr) {				// All Layers visible (upper layer visible alone)
		// *** shapeType = 6 !!!!
		// Delete Background, paint the red door black
		Graphics2D g2 = mask2D;
		if(reverse[frameNr]) {	// select color to paint
			g2.setColor(colBlack);
		} else {
			g2.setColor(colWhite);
		}
		g2.fillRect(0, 0, originalWidth, originalHeight);		// all layers are visible (upper is visible)

		for(int lay = 0;lay<layerCount-1;lay++) {			// save white mask for every layer
			String currMaskFolder = fullWorkingPath+maskFolder+movieNames[lay]+maskSubs+makeLeadingZeros(lay+1,2)+File.separator;
			String currMaskFileName = movieNames[lay]+maskSubs+makeLeadingZeros(lay+1,2)+"-"+makeLeadingZeros(frameNr,5);
			printMaskCreateFeedback(frameNr, currMaskFolder+currMaskFileName+"."+maskFileFormat.toLowerCase());
			saveMaskBufferAsImageFile(currMaskFolder+currMaskFileName);
		}
	}

	public void maskMakerWaitInLayer(int frameNr) {
		// *** shapeType = 3 !!!!
		Graphics2D g2 = mask2D;

		for(int lay = 0;lay<layerCount-1;lay++) {			// save white mask for every layer
			// Delete Background, paint the red door black
			if(reverse[frameNr]) {	// select color to paint
				g2.setColor(colWhite);
			} else {
				g2.setColor(colBlack);
			}
			g2.fillRect(0, 0, originalWidth, originalHeight);		// all layers are visible (upper is visible)

			if(lay > domLayerNr[frameNr]) {
				if(reverse[frameNr]) {	// select color to paint
					g2.setColor(colBlack);
				} else {
					g2.setColor(colWhite);
				}
				g2.fillRect(0, 0, originalWidth, originalHeight);		// all layers are visible (upper is visible)
			}
			String currMaskFolder = fullWorkingPath+maskFolder+movieNames[lay]+maskSubs+makeLeadingZeros(lay+1,2)+File.separator;
			String currMaskFileName = movieNames[lay]+maskSubs+makeLeadingZeros(lay+1,2)+"-"+makeLeadingZeros(frameNr,5);
			printMaskCreateFeedback(frameNr, currMaskFolder+currMaskFileName+"."+maskFileFormat.toLowerCase());
			saveMaskBufferAsImageFile(currMaskFolder+currMaskFileName);
		}
	}

	public void maskMakerAudioSync(int frameNr) {
		// *** shapeType = 7 !!!!
		Graphics2D g2 = mask2D;
		// Delete Background, paint the red door black
		for(int lay = 0;lay<layerCount-1;lay++) {			// save white mask for every layer
			// Delete Background, paint the red door black
			if(reverse[frameNr]) {	// select color to paint
				g2.setColor(colWhite);
			} else {
				g2.setColor(colBlack);
			}
			g2.fillRect(0, 0, originalWidth, originalHeight);		// all layers are visible (upper is visible)

			if(dominance[frameNr]) {
				if(reverse[frameNr]) {	// select color to paint
					g2.setColor(colBlack);
				} else {
					g2.setColor(colWhite);
				}
				g2.fillRect(0, 0, originalWidth, originalHeight);		// all layers are visible (upper is visible)
			}
			String currMaskFolder = fullWorkingPath+maskFolder+movieNames[lay]+maskSubs+makeLeadingZeros(lay+1,2)+File.separator;
			String currMaskFileName = movieNames[lay]+maskSubs+makeLeadingZeros(lay+1,2)+"-"+makeLeadingZeros(frameNr,5);
			printMaskCreateFeedback(frameNr, currMaskFolder+currMaskFileName+"."+maskFileFormat.toLowerCase());
			saveMaskBufferAsImageFile(currMaskFolder+currMaskFileName);
		}
	}

	public void maskMakerSimpleRect(int frameNr) {	// ** Only for testing and debugging
		// ** Standard Start start
		// *** shapeType = 99 !!!!

		Graphics2D g2 = mask2D;

		for(int lay = 0;lay<layerCount-1;lay++) {			// save white mask for every layer
			// Erase Background for every single layer
			if(reverse[frameNr]) {	// select color to paint
				g2.setColor(colWhite);
			} else {
				g2.setColor(colBlack);
			}
			g2.fillRect(0, 0, originalWidth, originalHeight);

			int px1 = (int)((lay/layerCount)*kx1[frameNr]);
			int py1 = (int)((lay/layerCount)*ky1[frameNr]*-1);	// *-1 because coordinate system is headup. Only coordinates need this. Height and Width dont.
			int pwid= (int)((lay/layerCount)*(Math.abs(kx2[frameNr])+Math.abs(kx1[frameNr]))); // In this case negative and positive value must be added up
			int phei= (int)((lay/layerCount)*(Math.abs(ky2[frameNr])+Math.abs(ky1[frameNr])));

			if(reverse[frameNr]) {	// select color to paint
				g2.setColor(colBlack);
			} else {
				g2.setColor(colWhite);
			}
			g2.translate(maskTranslX, maskTranslY);		// is calculated by int translX = (int)((originalWidth*scaleFactor)/2); and so on
			g2.drawRect(px1, py1, pwid, phei);
			g2.translate(-maskTranslX, -maskTranslY);	// reset before paint other stuff that has origin at 0,0

			String currMaskFolder = fullWorkingPath+maskFolder+movieNames[lay]+maskSubs+makeLeadingZeros(lay+1,2)+File.separator;
			String currMaskFileName = movieNames[lay]+maskSubs+makeLeadingZeros(lay+1,2)+"-"+makeLeadingZeros(frameNr,5);
			printMaskCreateFeedback(frameNr, currMaskFolder+currMaskFileName+"."+maskFileFormat.toLowerCase());
			saveMaskBufferAsImageFile(currMaskFolder+currMaskFileName);
		}
	}

	public void maskMakerFaderHorizontal(int frameNr) {
		// *** shapeType = 2 !!!!
		int lx = 0;
		int ly = 0;
		int lw = 0;
		int lh = 0;
		int startLayer 		= wide3[frameNr].intValue();
		int endLayer 		= wide4[frameNr].intValue();
		int localLayerCount = endLayer - startLayer;

		// Delete Background, paint the red door black
		Graphics2D g2 = mask2D;

		// Translate and paint and save layers
		for(int lay = startLayer;lay<endLayer-1;lay++) {
			int theLay = (endLayer-1) - lay;
			String currMaskFolder 	= "";
			String currMaskFileName = "";
			// Erase Background for every single layer
			if(reverse[frameNr]) {	// select color to paint
				g2.setColor(colWhite);
			} else {
				g2.setColor(colBlack);
			}
			g2.fillRect(0, 0, originalWidth, originalHeight);
			// Now translate - after background
			g2.translate(maskTranslX, maskTranslY);		// is calculated by int translX = (int)((originalWidth*scaleFactor)/2); and so on
			// Paint mask
			if(reverse[frameNr]) {	// Paint in reverse order to the files if not reverse
				currMaskFolder = fullWorkingPath+maskFolder+movieNames[lay]+maskSubs+makeLeadingZeros(lay+1,2)+File.separator;
				currMaskFileName = movieNames[lay]+maskSubs+makeLeadingZeros(lay+1,2)+"-"+makeLeadingZeros(frameNr,5);
			} else {
				currMaskFolder = fullWorkingPath+maskFolder+movieNames[theLay-1]+maskSubs+makeLeadingZeros(theLay,2)+File.separator;
				currMaskFileName = movieNames[theLay-1]+maskSubs+makeLeadingZeros(theLay,2)+"-"+makeLeadingZeros(frameNr,5);
			}
			// A simple rect .
			lx = (int)kx1[frameNr];
			ly = (int)(((ky1[frameNr]+(double)originalHeight+wide1[frameNr]) + ((double)lay * wide1[frameNr])))*-1;
			if(ly*-1<(originalHeight/2)) {	// if not mirrored, ly must be corrected
				ly = (originalHeight/2)*-1;
			}
			lh = (int)(originalHeight);
			if(directionY[frameNr]==-1) {	// in case it is mirrored, ly and lh must be corrected
				ly = (int)(((ky1[frameNr]*directionY[frameNr]-wide1[frameNr]) - ((double)lay * wide1[frameNr])))*-1;
				lh = (int)(originalHeight/2)-ly;
			}
			lw = (int)originalWidth;

			if(reverse[frameNr]) {	// select color to paint
				g2.setColor(colBlack);
			} else {
				g2.setColor(colWhite);
			}
			g2.fillRect(lx, ly, lw, lh);
			printMaskCreateFeedback(frameNr, currMaskFolder+currMaskFileName+"."+maskFileFormat.toLowerCase());
			saveMaskBufferAsImageFile(currMaskFolder+currMaskFileName);
			g2.translate(-maskTranslX, -maskTranslY);	// reset before paint other stuff that has origin at 0,0
		}
	}

	public void maskMakerLayerSpot(int frameNr) {
		// *** shapeType = 10 !!!!
		int xp = 0;
		int yp = 0;
		int radp = 0;
		int spottedLayer = domLayerNr[frameNr];
		Graphics2D g2 = mask2D;

		for(int lay = layerCount-1;lay>0;lay--) {
			int theLay = lay-1;
			if(reverse[frameNr]) {
				theLay = (layerCount-1) - lay;
			}
			String currMaskFolder = fullWorkingPath+maskFolder+movieNames[theLay]+maskSubs+makeLeadingZeros(theLay+1,2)+File.separator;
			String currMaskFileName = movieNames[theLay]+maskSubs+makeLeadingZeros(theLay+1,2)+"-"+makeLeadingZeros(frameNr,5);

			if(lay<=spottedLayer) {
				// Delete Background, paint the red door black
				if(reverse[frameNr]) {	// select color to paint
					g2.setColor(colWhite);
				} else {
					g2.setColor(colBlack);
				}
				g2.fillRect(0, 0, originalWidth, originalHeight);
				printMaskCreateFeedback(frameNr, currMaskFolder+currMaskFileName+"."+maskFileFormat.toLowerCase());
				saveMaskBufferAsImageFile(currMaskFolder+currMaskFileName);
			} else {
				// Delete Background, paint the red door black
				if(reverse[frameNr]) {	// select color to paint
					g2.setColor(colWhite);
				} else {
					g2.setColor(colBlack);
				}
				g2.fillRect(0, 0, originalWidth, originalHeight);

				g2.translate(maskTranslX, maskTranslY);		// is calculated by int translX = (int)((originalWidth*scaleFactor)/2); and so on
				radp = (int)((lay-spottedLayer) * (wide1[frameNr]));
				xp = (int)(kx1[frameNr]-(radp/2));		// fillOval Coordinates are in the left upper corner, not the center
				yp = ((int)(ky1[frameNr]+(radp/2))*-1);

				if(reverse[frameNr]) {	// select color to paint
					g2.setColor(colBlack);
				} else {
					g2.setColor(colWhite);
				}
				g2.fillOval(xp, yp, radp, radp);     // Fills the circle
				printMaskCreateFeedback(frameNr, currMaskFolder+currMaskFileName+"."+maskFileFormat.toLowerCase());
				saveMaskBufferAsImageFile(currMaskFolder+currMaskFileName);
				g2.translate(-maskTranslX, -maskTranslY);	// reset before paint other stuff that has origin at 0,0
			}
		}
	}

	public void maskMakerCircleInOut(int frameNr) {
		// *** shapeType = 4 !!!!
		int xp = 0;
		int yp = 0;
		int radp = 0;
		int domVal1 = 0;
		int dominanceWidth1 = 0;
		if(dominance[frameNr]) {
			dominanceWidth1 = (int)(wide1[frameNr] * domStrength[frameNr]);
		}
		Graphics2D g2 = mask2D;
		if(mode[frameNr]) {	// if grow mode, do this. if not, do the other
			int biggestRadius = (int)(wide1[frameNr] * (double)layerCount);
			int growLayCount = (int)((double)biggestRadius / wide2[frameNr]);	// Must have grown minimum the circle width to be painted
			if(growLayCount>layerCount-1) growLayCount = layerCount-1;
				for(int lay = 0;lay<layerCount-1;lay++) {
					int theLay = lay;
					if(!reverse[frameNr]) {
						theLay = (layerCount-1) - lay;
					}

					if(lay<growLayCount) {
						String currMaskFolder = fullWorkingPath+maskFolder+movieNames[theLay-1]+maskSubs+makeLeadingZeros(theLay,2)+File.separator;
						String currMaskFileName = movieNames[theLay-1]+maskSubs+makeLeadingZeros(theLay,2)+"-"+makeLeadingZeros(frameNr,5);

						// Delete Background, paint the red door black
						if(reverse[frameNr]) {	// select color to paint
							g2.setColor(colWhite);
						} else {
							g2.setColor(colBlack);
						}
						g2.fillRect(0, 0, originalWidth, originalHeight);

						g2.translate(maskTranslX, maskTranslY);		// is calculated by int translX = (int)((originalWidth*scaleFactor)/2); and so on
						// No Dominance Layer possible in grow Mode
						radp = (int)((double)biggestRadius - ((double)lay * wide2[frameNr]));
						xp = (int)( kx1[frameNr]-(radp/2));		// fillOval Coordinates are in the left upper corner, not the center
						yp = ((int)( ky1[frameNr]+(radp/2))*-1);

						if(reverse[frameNr]) {	// select color to paint
							g2.setColor(colBlack);
						} else {
							g2.setColor(colWhite);
						}
						g2.fillOval(xp, yp, radp, radp);     // Fills the circle
						printMaskCreateFeedback(frameNr, currMaskFolder+currMaskFileName+"."+maskFileFormat.toLowerCase());
						saveMaskBufferAsImageFile(currMaskFolder+currMaskFileName);
						g2.translate(-maskTranslX, -maskTranslY);	// reset before paint other stuff that has origin at 0,0
					} else {
						String currMaskFolder = fullWorkingPath+maskFolder+movieNames[theLay-1]+maskSubs+makeLeadingZeros(theLay,2)+File.separator;
						String currMaskFileName = movieNames[theLay-1]+maskSubs+makeLeadingZeros(theLay,2)+"-"+makeLeadingZeros(frameNr,5);
						// Delete Background, paint the red door black
						if(reverse[frameNr]) {	// select color to paint
							g2.setColor(colWhite);
						} else {
							g2.setColor(colBlack);
						}
						g2.fillRect(0, 0, originalWidth, originalHeight);

						printMaskCreateFeedback(frameNr, currMaskFolder+currMaskFileName+"."+maskFileFormat.toLowerCase());
						saveMaskBufferAsImageFile(currMaskFolder+currMaskFileName);
					}
				}
			} else {
				for(int lay = layerCount-1;lay>0;lay--) {
					int theLay = lay-1;
					if(reverse[frameNr]) {
						theLay = (layerCount-1) - lay;
					}
					String currMaskFolder = fullWorkingPath+maskFolder+movieNames[theLay]+maskSubs+makeLeadingZeros(theLay+1,2)+File.separator;
					String currMaskFileName = movieNames[theLay]+maskSubs+makeLeadingZeros(theLay+1,2)+"-"+makeLeadingZeros(frameNr,5);
					// Delete Background, paint the red door black
					if(reverse[frameNr]) {	// select color to paint
						g2.setColor(colWhite);
					} else {
						g2.setColor(colBlack);
					}
					g2.fillRect(0, 0, originalWidth, originalHeight);

					g2.translate(maskTranslX, maskTranslY);		// is calculated by int translX = (int)((originalWidth*scaleFactor)/2); and so on


					if(dominance[frameNr]) {
						if(lay>domLayerNr[frameNr]) {
							domVal1 = dominanceWidth1;
						} else {
							domVal1 = 0;
						}
					}
					radp = (int)(lay * (domVal1+wide1[frameNr]));
					xp = (int)(kx1[frameNr]-(radp/2));		// fillOval Coordinates are in the left upper corner, not the center
					yp = ((int)(ky1[frameNr]+(radp/2))*-1);

					if(reverse[frameNr]) {	// select color to paint
						g2.setColor(colBlack);
					} else {
						g2.setColor(colWhite);
					}
					g2.fillOval(xp, yp, radp, radp);     // Fills the circle
					printMaskCreateFeedback(frameNr, currMaskFolder+currMaskFileName+"."+maskFileFormat.toLowerCase());
					saveMaskBufferAsImageFile(currMaskFolder+currMaskFileName);
					g2.translate(-maskTranslX, -maskTranslY);	// reset before paint other stuff that has origin at 0,0
				}
			}
	}

	public void maskMakerQuadBox(int frameNr) {
		// *** shapeType = 4 !!!!
		int xp = 0;
		int yp = 0;
		int radp = 0;
		int domVal1 = 0;
		int dominanceWidth1 = 0;
		if(dominance[frameNr]) {
			dominanceWidth1 = (int)(wide1[frameNr] * domStrength[frameNr]);
		}
		Graphics2D g2 = mask2D;
		if(mode[frameNr]) {	// if grow mode, do this. if not, do the other
			int biggestRadius = (int)(wide1[frameNr] * (double)layerCount);
			int growLayCount = (int)((double)biggestRadius / wide2[frameNr]);	// Must have grown minimum the circle width to be painted
			if(growLayCount>layerCount-1) growLayCount = layerCount-1;
				for(int lay = 0;lay<layerCount-1;lay++) {
					int theLay = lay;
					if(!reverse[frameNr]) {
						theLay = (layerCount-1) - lay;
					}

					if(lay<growLayCount) {
						String currMaskFolder = fullWorkingPath+maskFolder+movieNames[theLay-1]+maskSubs+makeLeadingZeros(theLay,2)+File.separator;
						String currMaskFileName = movieNames[theLay-1]+maskSubs+makeLeadingZeros(theLay,2)+"-"+makeLeadingZeros(frameNr,5);

						// Delete Background, paint the red door black
						if(reverse[frameNr]) {	// select color to paint
							g2.setColor(colWhite);
						} else {
							g2.setColor(colBlack);
						}
						g2.fillRect(0, 0, originalWidth, originalHeight);

						g2.translate(maskTranslX, maskTranslY);		// is calculated by int translX = (int)((originalWidth*scaleFactor)/2); and so on
						// No Dominance Layer possible in grow Mode
						radp = (int)((double)biggestRadius - ((double)lay * wide2[frameNr]));
						xp = (int)( kx1[frameNr]-(radp/2));		// fillOval Coordinates are in the left upper corner, not the center
						yp = ((int)( ky1[frameNr]+(radp/2))*-1);

						if(reverse[frameNr]) {	// select color to paint
							g2.setColor(colBlack);
						} else {
							g2.setColor(colWhite);
						}
						//g2.fillOval(xp, yp, radp, radp);     // Fills the circle
						g2.fillRect(xp,yp,radp,radp);
						printMaskCreateFeedback(frameNr, currMaskFolder+currMaskFileName+"."+maskFileFormat.toLowerCase());
						saveMaskBufferAsImageFile(currMaskFolder+currMaskFileName);
						g2.translate(-maskTranslX, -maskTranslY);	// reset before paint other stuff that has origin at 0,0
					} else {
						String currMaskFolder = fullWorkingPath+maskFolder+movieNames[theLay-1]+maskSubs+makeLeadingZeros(theLay,2)+File.separator;
						String currMaskFileName = movieNames[theLay-1]+maskSubs+makeLeadingZeros(theLay,2)+"-"+makeLeadingZeros(frameNr,5);
						// Delete Background, paint the red door black
						if(reverse[frameNr]) {	// select color to paint
							g2.setColor(colWhite);
						} else {
							g2.setColor(colBlack);
						}
						g2.fillRect(0, 0, originalWidth, originalHeight);

						printMaskCreateFeedback(frameNr, currMaskFolder+currMaskFileName+"."+maskFileFormat.toLowerCase());
						saveMaskBufferAsImageFile(currMaskFolder+currMaskFileName);
					}
				}
			} else {
				for(int lay = layerCount-1;lay>0;lay--) {
					int theLay = lay-1;
					if(reverse[frameNr]) {
						theLay = (layerCount-1) - lay;
					}
					String currMaskFolder = fullWorkingPath+maskFolder+movieNames[theLay]+maskSubs+makeLeadingZeros(theLay+1,2)+File.separator;
					String currMaskFileName = movieNames[theLay]+maskSubs+makeLeadingZeros(theLay+1,2)+"-"+makeLeadingZeros(frameNr,5);
					// Delete Background, paint the red door black
					if(reverse[frameNr]) {	// select color to paint
						g2.setColor(colWhite);
					} else {
						g2.setColor(colBlack);
					}
					g2.fillRect(0, 0, originalWidth, originalHeight);

					g2.translate(maskTranslX, maskTranslY);		// is calculated by int translX = (int)((originalWidth*scaleFactor)/2); and so on


					if(dominance[frameNr]) {
						if(lay>domLayerNr[frameNr]) {
							domVal1 = dominanceWidth1;
						} else {
							domVal1 = 0;
						}
					}
					radp = (int)(lay * (domVal1+wide1[frameNr]));
					xp = (int)(kx1[frameNr]-(radp/2));		// fillOval Coordinates are in the left upper corner, not the center
					yp = ((int)(ky1[frameNr]+(radp/2))*-1);

					if(reverse[frameNr]) {	// select color to paint
						g2.setColor(colBlack);
					} else {
						g2.setColor(colWhite);
					}
					//g2.fillOval(xp, yp, radp, radp);     // Fills the circle
					g2.fillRect(xp,yp,radp,radp);
					printMaskCreateFeedback(frameNr, currMaskFolder+currMaskFileName+"."+maskFileFormat.toLowerCase());
					saveMaskBufferAsImageFile(currMaskFolder+currMaskFileName);
					g2.translate(-maskTranslX, -maskTranslY);	// reset before paint other stuff that has origin at 0,0
				}
			}
	}

	public void maskMakerRectBox(int frameNr) {
		// *** shapeType = 4 !!!!
		int xp = 0;
		int yp = 0;
		int radp = 0;
		int domVal1 = 0;
		int dominanceWidth1 = 0;
		if(dominance[frameNr]) {
			dominanceWidth1 = (int)(wide1[frameNr] * domStrength[frameNr]);
		}
		Graphics2D g2 = mask2D;
		if(mode[frameNr]) {	// if grow mode, do this. if not, do the other
			int biggestRadius = (int)(wide1[frameNr] * (double)layerCount);
			int growLayCount = (int)((double)biggestRadius / wide2[frameNr]);	// Must have grown minimum the circle width to be painted
			if(growLayCount>layerCount-1) growLayCount = layerCount-1;
				for(int lay = 0;lay<layerCount-1;lay++) {
					int theLay = lay;
					if(!reverse[frameNr]) {
						theLay = (layerCount-1) - lay;
					}

					if(lay<growLayCount) {
						String currMaskFolder = fullWorkingPath+maskFolder+movieNames[theLay-1]+maskSubs+makeLeadingZeros(theLay,2)+File.separator;
						String currMaskFileName = movieNames[theLay-1]+maskSubs+makeLeadingZeros(theLay,2)+"-"+makeLeadingZeros(frameNr,5);

						// Delete Background, paint the red door black
						if(reverse[frameNr]) {	// select color to paint
							g2.setColor(colWhite);
						} else {
							g2.setColor(colBlack);
						}
						g2.fillRect(0, 0, originalWidth, originalHeight);

						g2.translate(maskTranslX, maskTranslY);		// is calculated by int translX = (int)((originalWidth*scaleFactor)/2); and so on
						// No Dominance Layer possible in grow Mode
						radp = (int)((double)biggestRadius - ((double)lay * wide2[frameNr]));
						xp = (int)( kx1[frameNr]-(radp/2));		// fillOval Coordinates are in the left upper corner, not the center
						yp = ((int)( ky1[frameNr]+(radp/2))*-1);

						if(reverse[frameNr]) {	// select color to paint
							g2.setColor(colBlack);
						} else {
							g2.setColor(colWhite);
						}
						//g2.fillOval(xp, yp, radp, radp);     // Fills the circle
						g2.fillRect(xp-(radp/2),yp,2*radp,radp);
						printMaskCreateFeedback(frameNr, currMaskFolder+currMaskFileName+"."+maskFileFormat.toLowerCase());
						saveMaskBufferAsImageFile(currMaskFolder+currMaskFileName);
						g2.translate(-maskTranslX, -maskTranslY);	// reset before paint other stuff that has origin at 0,0
					} else {
						String currMaskFolder = fullWorkingPath+maskFolder+movieNames[theLay-1]+maskSubs+makeLeadingZeros(theLay,2)+File.separator;
						String currMaskFileName = movieNames[theLay-1]+maskSubs+makeLeadingZeros(theLay,2)+"-"+makeLeadingZeros(frameNr,5);
						// Delete Background, paint the red door black
						if(reverse[frameNr]) {	// select color to paint
							g2.setColor(colWhite);
						} else {
							g2.setColor(colBlack);
						}
						g2.fillRect(0, 0, originalWidth, originalHeight);

						printMaskCreateFeedback(frameNr, currMaskFolder+currMaskFileName+"."+maskFileFormat.toLowerCase());
						saveMaskBufferAsImageFile(currMaskFolder+currMaskFileName);
					}
				}
			} else {
				for(int lay = layerCount-1;lay>0;lay--) {
					int theLay = lay-1;
					if(reverse[frameNr]) {
						theLay = (layerCount-1) - lay;
					}
					String currMaskFolder = fullWorkingPath+maskFolder+movieNames[theLay]+maskSubs+makeLeadingZeros(theLay+1,2)+File.separator;
					String currMaskFileName = movieNames[theLay]+maskSubs+makeLeadingZeros(theLay+1,2)+"-"+makeLeadingZeros(frameNr,5);
					// Delete Background, paint the red door black
					if(reverse[frameNr]) {	// select color to paint
						g2.setColor(colWhite);
					} else {
						g2.setColor(colBlack);
					}
					g2.fillRect(0, 0, originalWidth, originalHeight);

					g2.translate(maskTranslX, maskTranslY);		// is calculated by int translX = (int)((originalWidth*scaleFactor)/2); and so on


					if(dominance[frameNr]) {
						if(lay>domLayerNr[frameNr]) {
							domVal1 = dominanceWidth1;
						} else {
							domVal1 = 0;
						}
					}
					radp = (int)(lay * (domVal1+wide1[frameNr]));
					xp = (int)(kx1[frameNr]-(radp/2));		// fillOval Coordinates are in the left upper corner, not the center
					yp = ((int)(ky1[frameNr]+(radp/2))*-1);

					if(reverse[frameNr]) {	// select color to paint
						g2.setColor(colBlack);
					} else {
						g2.setColor(colWhite);
					}
					//g2.fillOval(xp, yp, radp, radp);     // Fills the circle
					g2.fillRect(xp-(radp/2),yp,2*radp,radp);
					printMaskCreateFeedback(frameNr, currMaskFolder+currMaskFileName+"."+maskFileFormat.toLowerCase());
					saveMaskBufferAsImageFile(currMaskFolder+currMaskFileName);
					g2.translate(-maskTranslX, -maskTranslY);	// reset before paint other stuff that has origin at 0,0
				}
			}
	}

	public void maskMakerDualCircle(int frameNr) {
		// *** shapeType = 9 !!!!
		int xp = 0;
		int yp = 0;
		int radp = 0;
		int domVal1 = 0;
		int dominanceWidth1 = 0;
		if(dominance[frameNr]) {
			dominanceWidth1 = (int)(wide1[frameNr] * domStrength[frameNr]);
		}
		// same for second circle:
		int xp2 = 0;
		int yp2 = 0;
		int radp2 = 0;
		int domVal2 = 0;
		int dominanceWidth2 = 0;
		if(dominance[frameNr]) {
			dominanceWidth2 = (int)(wide2[frameNr] * domStrength[frameNr]);
		}

		Graphics2D g2 = mask2D;

		for(int lay = layerCount-1;lay>0;lay--) {
			// Paint circle 1
			int theLay = lay-1;
			if(reverse[frameNr]) {
				theLay = (layerCount-1) - lay;
			}
			String currMaskFolder = fullWorkingPath+maskFolder+movieNames[theLay]+maskSubs+makeLeadingZeros(theLay+1,2)+File.separator;
			String currMaskFileName = movieNames[theLay]+maskSubs+makeLeadingZeros(theLay+1,2)+"-"+makeLeadingZeros(frameNr,5);
			// Delete Background, paint the red door black
			if(reverse[frameNr]) {	// select color to paint
				g2.setColor(colWhite);
			} else {
				g2.setColor(colBlack);
			}
			g2.fillRect(0, 0, originalWidth, originalHeight);

			g2.translate(maskTranslX, maskTranslY);		// is calculated by int translX = (int)((originalWidth*scaleFactor)/2); and so on


			if(dominance[frameNr]) {
				if(lay>domLayerNr[frameNr]) {
					domVal1 = dominanceWidth1;
				} else {
					domVal1 = 0;
				}
			}
			radp = (int)(lay * (domVal1+wide1[frameNr]));
			xp = (int)(kx1[frameNr]-(radp/2));		// fillOval Coordinates are in the left upper corner, not the center
			yp = ((int)(ky1[frameNr]+(radp/2))*-1);

			if(reverse[frameNr]) {	// select color to paint
				g2.setColor(colBlack);
			} else {
				g2.setColor(colWhite);
			}
			g2.fillOval(xp, yp, radp, radp);     // Fills the circle

			// Paint circle 2
			if(dominance[frameNr]) {
				if(lay>domLayerNr[frameNr]) {
					domVal2 = dominanceWidth2;
				} else {
					domVal2 = 0;
				}
			}
			radp2 = (int)(lay * (domVal2+wide2[frameNr]));
			xp2 = (int)(kx2[frameNr]-(radp2/2));		// fillOval Coordinates are in the left upper corner, not the center
			yp2 = ((int)(ky2[frameNr]+(radp2/2))*-1);
			// same color. no change needed
			g2.fillOval(xp2, yp2, radp2, radp2);     // Fills the circle

			printMaskCreateFeedback(frameNr, currMaskFolder+currMaskFileName+"."+maskFileFormat.toLowerCase());
			saveMaskBufferAsImageFile(currMaskFolder+currMaskFileName);
			g2.translate(-maskTranslX, -maskTranslY);	// reset before paint other stuff that has origin at 0,0
		}

	}

	public void maskMakerPolyRect(int frameNr) {	// paints one figure per frame. A figure is a polygon with 4 points
		// *** shapeType = 1 !!!!
		int xp[] = new int[4];	// coordinate arrays for a polygon object
		int yp[] = new int[4];
		int domVal1 = 0;
		int domVal2 = 0;
		int dominanceWidth1 = 0;
		int dominanceWidth2 = 0;
		if(dominance[frameNr]) {
			dominanceWidth1 = (int)(wide1[frameNr] * domStrength[frameNr]);
			dominanceWidth2 = (int)(wide2[frameNr] * domStrength[frameNr]);
		}

		Graphics2D g2 = mask2D;

		for(int lay = layerCount-1;lay>0;lay--) {
			int theLay = lay-1;
			if(reverse[frameNr]) {
				theLay = (layerCount-1) - lay;
			}
			String currMaskFolder = fullWorkingPath+maskFolder+movieNames[theLay]+maskSubs+makeLeadingZeros(theLay+1,2)+File.separator;
			String currMaskFileName = movieNames[theLay]+maskSubs+makeLeadingZeros(theLay+1,2)+"-"+makeLeadingZeros(frameNr,5);

			// Delete Background, paint the red door black
			if(reverse[frameNr]) {	// select color to paint
				g2.setColor(colWhite);
			} else {
				g2.setColor(colBlack);
			}
			g2.fillRect(0, 0, originalWidth, originalHeight);

			g2.translate(maskTranslX, maskTranslY);		// is calculated by int translX = (int)((originalWidth*scaleFactor)/2); and so on
			// A simple rect polygon.
			if(dominance[frameNr]) {
				if(lay>domLayerNr[frameNr]) {
					domVal1 = dominanceWidth1;
					domVal2 = dominanceWidth2;
				} else {
					domVal1 = 0;
					domVal2 = 0;
				}
			}
			xp[0]=((int)(kx1[frameNr]) - (int)(lay * (domVal1+wide1[frameNr])))*directionX[frameNr];
			xp[1]=((int)(kx1[frameNr]) + (int)(lay * (domVal1+wide1[frameNr])))*directionX[frameNr];
			xp[2]=((int)(kx2[frameNr]) + (int)(lay * (domVal2+wide2[frameNr])))*directionX[frameNr];
			xp[3]=((int)(kx2[frameNr]) - (int)(lay * (domVal2+wide2[frameNr])))*directionX[frameNr];

			yp[0]=((int)(ky1[frameNr])*-1)*directionY[frameNr];
			yp[1]=((int)(ky1[frameNr])*-1)*directionY[frameNr];
			yp[2]=((int)(ky2[frameNr])*-1)*directionY[frameNr];
			yp[3]=((int)(ky2[frameNr])*-1)*directionY[frameNr];
			int n = 4;

			Polygon p = new Polygon(xp, yp, n);  // This polygon represents a triangle with the above vertices.

			if(reverse[frameNr]) {	// select color to paint
				g2.setColor(colBlack);
			} else {
				g2.setColor(colWhite);
			}

			g2.fillPolygon(p);     // Fills the triangle above.
			printMaskCreateFeedback(frameNr, currMaskFolder+currMaskFileName+"."+maskFileFormat.toLowerCase());
			saveMaskBufferAsImageFile(currMaskFolder+currMaskFileName);
			g2.translate(-maskTranslX, -maskTranslY);	// reset before paint other stuff that has origin at 0,0
		}
	}

	public void maskMakerHorizontThreeStepFader(int frameNr) {
		// *** shapeType = 14 !!!!
		int lx = 0;
		int ly = 0;
		int lw = 0;
		int lh = 0;

		int cv = (int)((originalHeight/2)*-1); 	//compare value if last frame is lower than screen height

		// Delete Background, paint the red door black
		Graphics2D g2 = mask2D;

		// Translate and paint layers
		for(int lay = 0;lay<layerCount-1;lay++) {

			int theLay = (layerCount-1) - lay;
			if(reverse[frameNr]) {
				theLay = lay+1;
			}
			String currMaskFolder = fullWorkingPath+maskFolder+movieNames[theLay-1]+maskSubs+makeLeadingZeros(theLay,2)+File.separator;
			String currMaskFileName = movieNames[theLay-1]+maskSubs+makeLeadingZeros(theLay,2)+"-"+makeLeadingZeros(frameNr,5);

			// Delete Background, paint the red door black
			if(reverse[frameNr]) {	// select color to paint
				g2.setColor(colWhite);
			} else {
				g2.setColor(colBlack);
			}
			g2.fillRect(0, 0, originalWidth, originalHeight);

			g2.translate(maskTranslX, maskTranslY);		// is calculated by int translX = (int)((originalWidth*scaleFactor)/2); and so on

			// A simple rect .
			lx = (int)(kx1[frameNr]);
			ly = (int)(((ky1[frameNr]+(double)originalHeight) - (layerCount*wide1[frameNr]) + ((double)lay * wide1[frameNr]*2) ))*-1;
			if(ly>cv)
				ly = cv;
			lw = (int)((double)originalWidth);
			lh = (int)(((double)originalHeight));
			if(reverse[frameNr]) {	// select color to paint
				g2.setColor(colBlack);
			} else {
				g2.setColor(colWhite);
			}
			g2.fillRect(lx, ly, lw, lh);
			printMaskCreateFeedback(frameNr, currMaskFolder+currMaskFileName+"."+maskFileFormat.toLowerCase());
			saveMaskBufferAsImageFile(currMaskFolder+currMaskFileName);
			g2.translate(-maskTranslX, -maskTranslY);	// reset before paint other stuff that has origin at 0,0
		}
	}

	public void maskMakerCircleRose(int frameNr) {
		// *** shapeType = 15 !!!!
		int startx=0;
		int starty=0;
		int arcwidth = 0;
		int archeight = 0;
		int arcstart = 0;
		int arcrange = 0;
		double darcrange = 0.0;
		double darcstart = 0.0;

		// Delete Background, paint the red door black
		Graphics2D g2 = mask2D;

		for(int lay = 0;lay<layerCount-1;lay++) {
			int theLay = (layerCount-1) - lay;
			if(reverse[frameNr]) {
				theLay = lay+1;
				System.out.println("reverse");
			}
			System.out.println("thelay: "+theLay);
			String currMaskFolder = fullWorkingPath+maskFolder+movieNames[theLay-1]+maskSubs+makeLeadingZeros(theLay,2)+File.separator;
			String currMaskFileName = movieNames[theLay-1]+maskSubs+makeLeadingZeros(theLay,2)+"-"+makeLeadingZeros(frameNr,5);

			// Delete Background, paint the red door black
			if(reverse[frameNr]) {	// select color to paint
				g2.setColor(colWhite);
			} else {
				g2.setColor(colBlack);
			}
			g2.fillRect(0, 0, originalWidth, originalHeight);

			g2.translate(maskTranslX, maskTranslY);		// is calculated by int translX = (int)((originalWidth*scaleFactor)/2); and so on

			arcwidth	= (int)(1.0*wide2[frameNr]);
			archeight	= (int)(1.0*wide2[frameNr]);
			startx		= (int)((kx1[frameNr])-((double)arcwidth/2.0));
			starty		= (int)(((ky1[frameNr]*-1.0)) - ((double)archeight/2.0));
			darcrange 	= (((wide1[frameNr]/(double)layerCount)*((double)layerCount-(double)lay)));
			darcstart 	= (double)darcrange/2.0+90.0;

			if(reverse[frameNr]) {	// select color to paint
				g2.setColor(colBlack);
			} else {
				g2.setColor(colWhite);
			}
			fillPolyArc(g2, startx, starty, arcwidth, archeight, darcstart, darcrange);
			printMaskCreateFeedback(frameNr, currMaskFolder+currMaskFileName+"."+maskFileFormat.toLowerCase());
			saveMaskBufferAsImageFile(currMaskFolder+currMaskFileName);
			g2.translate(-maskTranslX, -maskTranslY);	// reset before paint other stuff that has origin at 0,0
		}
	}

	public void maskMakerPerspectiveRays(int frameNr) {
		// *** shapeType = 11 !!!!
		int xp[] = new int[3];	// coordinate arrays for a polygon object
		int yp[] = new int[3];

		// Delete Background, paint the red door black
		Graphics2D g2 = mask2D;

		for(int lay = layerCount-1;lay>0;lay--) {
			int theLay = lay-1;
			if(reverse[frameNr]) {
				theLay = (layerCount-1) - lay;
			}
			String currMaskFolder = fullWorkingPath+maskFolder+movieNames[theLay]+maskSubs+makeLeadingZeros(theLay+1,2)+File.separator;
			String currMaskFileName = movieNames[theLay]+maskSubs+makeLeadingZeros(theLay+1,2)+"-"+makeLeadingZeros(frameNr,5);

			// Delete Background, paint the red door black
			if(reverse[frameNr]) {	// select color to paint
				g2.setColor(colWhite);
			} else {
				g2.setColor(colBlack);
			}
			g2.fillRect(0, 0, originalWidth, originalHeight);

			g2.translate(maskTranslX, maskTranslY);		// is calculated by int translX = (int)((originalWidth*scaleFactor)/2); and so on

			xp[0]=(int)(kx1[frameNr]) * directionX[frameNr];
			yp[0]=((int)(ky1[frameNr])*-1)*directionY[frameNr];

			if(wide3[frameNr]>0) {	// Horizontal Rays
				xp[1]=(int)(kx2[frameNr]) *directionX[frameNr];
				xp[2]=(int)(kx2[frameNr]) *directionX[frameNr];

				yp[1]=(((int)(ky2[frameNr])+ (int)(lay * (wide1[frameNr])))*-1)*directionY[frameNr];
				yp[2]=(((int)(ky2[frameNr])- (int)(lay * (wide2[frameNr])))*-1)*directionY[frameNr];
			} else {				// Vertical Rays
				xp[1]=((int)(kx2[frameNr]) + (int)(lay * (wide1[frameNr])))*directionX[frameNr];
				xp[2]=((int)(kx2[frameNr]) - (int)(lay * (wide2[frameNr])))*directionX[frameNr];

				yp[1]=((int)(ky2[frameNr])*-1)*directionY[frameNr];
				yp[2]=((int)(ky2[frameNr])*-1)*directionY[frameNr];
			}
			//******************

			int n = 3;

			Polygon p = new Polygon(xp, yp, n);  // This polygon represents a triangle with the above vertices.

			if(reverse[frameNr]) {	// select color to paint
				g2.setColor(colBlack);
			} else {
				g2.setColor(colWhite);
			}

			g2.fillPolygon(p);     // Fills the triangle above.
			printMaskCreateFeedback(frameNr, currMaskFolder+currMaskFileName+"."+maskFileFormat.toLowerCase());
			saveMaskBufferAsImageFile(currMaskFolder+currMaskFileName);
			g2.translate(-maskTranslX, -maskTranslY);	// reset before paint other stuff that has origin at 0,0
		}
	}
	// *** General File (Making) Methods ***************************************************************
	public void setupFileData() {
		//if fileData cannot be read, layerCount will be set to the default in the catch statement
		try {
			int movCounter = 0;
			String whereAmI = System.getProperty("user.dir");
			fullWorkingPath = whereAmI.substring(0, whereAmI.length()-4);
			File folderMovs = new File(fullWorkingPath+movFolder);
			File[] files = folderMovs.listFiles();
			if (files != null) {
				for (int i = 0; i < files.length; i++) {
					if(files[i].getName().endsWith(originalMovieEnding)) {	// MOV or avi, or whatever see header
						movCounter++;
					}
				}
				layerCount	= movCounter;
				greys		= new Color[movCounter];
				movieNames	= new String[movCounter];
				int mvc = 0;
				for (int i = 0; i < files.length; i++) {
					if(files[i].getName().endsWith(originalMovieEnding)) {	// MOV or avi or ...
						movieNames[mvc] = files[i].getName().substring(0,files[i].getName().length()-4);	// if ending longer than .mov, change the value
						mvc++;
					}
				}
			}
		} catch(Exception e) {
			System.out.println("Error analysing the filestructure. Please check foldernames and structure as described in Readme.txt");
			layerCount	= 19;
			greys		= new Color[layerCount];
			movieNames	= new String[layerCount];
			for(int i = 0;i<layerCount;i++) {
				movieNames[i] = "Dummy"+i;
			}
		}
	}
	public String makeLeadingZeros(int theNr, int stellen) {
		String endNumber ="";
		String tempNr = ""+theNr;
		for(int i = 0;i<(stellen-tempNr.length());i++) {	// set amount of leading zeros here
			endNumber = endNumber+"0";
		}
		endNumber = endNumber + tempNr;
		return endNumber;
	}
	public void checkForFolder(String fullPath) {
		File f = new File(fullPath);
		if (!f.exists()){
			f.mkdir(); //create if not exists
		}
	}
	public void saveMaskBufferAsImageFile(String fullFileName) {	// File type is set in Header as bmp or jpg. Comment out what you like
		try{
			File datei = new File(fullFileName+"."+maskFileFormat.toLowerCase());
			ImageIO.write(maskBuffer, maskFileFormat, datei );
		} catch(Exception e){
			System.out.println("Error: "+e);
		}
	}
	public void createTheMasks() {								// Main Mask Creation Controller Method
		final long timeStart = System.currentTimeMillis();
		maskBuffer = new BufferedImage(originalWidth, originalHeight, BufferedImage.TYPE_INT_RGB);
		mask2D = maskBuffer.createGraphics();
		if(lineAlias) {
			mask2D.setRenderingHint( RenderingHints.KEY_TEXT_ANTIALIASING, RenderingHints.VALUE_TEXT_ANTIALIAS_ON);
			mask2D.setRenderingHint(RenderingHints.KEY_ANTIALIASING,RenderingHints.VALUE_ANTIALIAS_ON);
		}
		// Now create the fileStructure if it not exists
		checkForFolder(fullWorkingPath+maskFolder);
		for(int y = 0;y<layerCount-1;y++) {
			String currMaskFolder = fullWorkingPath+maskFolder+movieNames[y]+maskSubs+makeLeadingZeros(y+1,2)+File.separator;
			checkForFolder(currMaskFolder);
		}
		// Instead of a timer we go by cpu time - as fast as possible
		for(int frameNr = 0;frameNr < projectFrames;frameNr++) {
			if(shapeType[frameNr]==1) {	// if shape type = 1 use poly rect renderer
				maskMakerPolyRect(frameNr);			// General Poly Rect
			} else if(shapeType[frameNr]==2) {
				maskMakerFaderHorizontal(frameNr);		// fade Horizontal renderer
			} else if(shapeType[frameNr]==3) {
				maskMakerWaitInLayer(frameNr);		// Wait in Layer renderer
			} else if(shapeType[frameNr]==4) {
				maskMakerCircleInOut(frameNr);		// Simple Circle In Out renderer
			} else if(shapeType[frameNr]==5) {
				maskMakerStartAnimation(frameNr);	// Start Animation Real Time renderer
			} else if(shapeType[frameNr]==6) {
				maskMakerWaitEmpty(frameNr);		// Wait in Empty Layer renderer
			} else if(shapeType[frameNr]==7) {
				maskMakerAudioSync(frameNr);		// Wait in Empty Layer renderer
			} else if(shapeType[frameNr]==8) {
				maskMakerCircleInOut(frameNr);		// Wait in Empty Layer renderer
			} else if(shapeType[frameNr]==9) {
				maskMakerDualCircle(frameNr);		// Wait in Empty Layer renderer
			} else if(shapeType[frameNr]==10) {
				maskMakerLayerSpot(frameNr);		// Wait in Empty Layer renderer
			} else if(shapeType[frameNr]==11) {
				maskMakerPerspectiveRays(frameNr);		// Wait in Empty Layer renderer
			} else if(shapeType[frameNr]==12) {
				maskMakerQuadBox(frameNr);		// Wait in Empty Layer renderer
			} else if(shapeType[frameNr]==13) {
				maskMakerRectBox(frameNr);		// Wait in Empty Layer renderer
			} else if(shapeType[frameNr]==14) {
				maskMakerHorizontThreeStepFader(frameNr);		// Wait in Empty Layer renderer
			} else if(shapeType[frameNr]==15) {
				maskMakerCircleRose(frameNr);		// Wait in Empty Layer renderer
			} else if(shapeType[frameNr]==99) {
				maskMakerSimpleRect(frameNr);		// Simple Rect Renderer
			} else { // no shape calculated for this frame
				maskMakerEmpty(frameNr);				// Empty Frame Renderer
			}
		}

		mask2D.dispose();
		maskBuffer.flush();
		maskBuffer = null;

		final long timeEnd = System.currentTimeMillis();

		int totalSeconde = (int)((timeEnd - timeStart)/1000);
		int theMinute = (int)((double)totalSeconde/(double)60);
		int theSeconde = (int)((double)totalSeconde - ((double)theMinute*(double)60));
		String dura = makeLeadingZeros(theMinute, 2)+":"+makeLeadingZeros(theSeconde, 2);

		printStatus("Masks created up to frame: "+projectFrames+" In: "+dura+" (min:sec)");
	}

	// *** General Paint Methods **************************************************************
	// *** Paint Timing Info in Image
	public void paintTimingInfo(Graphics2D g2d) {	// get Graphics 2D of current ImageBuffer
		g2d.setColor(colYellow);
		g2d.setFont(new Font("SansSerif", Font.PLAIN, 10));

		g2d.drawString("Frame: ", timinginfoX, timinginfoY); // timinginfo coordinate is pre calculated in head area
		g2d.drawString(""+currFrame, timinginfoX+40, timinginfoY);

		int currSeconde = (int)((double)currFrame / 25.0)+1;
		g2d.drawString("Second: ", timinginfoX+80, timinginfoY);
		g2d.drawString(""+currSeconde, timinginfoX+126, timinginfoY);

		int currBeat = (int)((((double)beatsPerMinute/60.0)/framesPerSecond)*(double)currFrame)+1;
		g2d.drawString("Beat: ", timinginfoX+160, timinginfoY);
		g2d.drawString(""+currBeat, timinginfoX+192, timinginfoY);
	}

	// *** Timeline
	public void paintTimeline() {
		int timeLineAreaStart = 80;
		int timelineAreaWidth = prevWidth - timeLineAreaStart-10;

		Graphics2D g2d = (Graphics2D)timeLinePanel.getGraphics();

        // Time Marker
        g2d.setColor(colBlack);

        // Frame Marker on the timeline
        g2d.setFont(new Font("SansSerif", Font.BOLD, 12));
        int x = 5;
        int y = 28;
        g2d.drawString("Frames:", x, y+7);

        double fspace = 200;	// 200 for 180 second video is good
        double pixelPerFrame = (double)timelineAreaWidth / (double)projectFrames;
        double frameMarkerCount = (double)projectFrames / fspace;
        double markerX = 0.0;
        g2d.setFont(new Font("SansSerif", Font.BOLD, 10));
        int tOffset = 2;
        for(int i = 0;i<frameMarkerCount;i++) {
			if(i*fspace<10) {
				tOffset = 2;
			} else if(i*fspace>9 && i*fspace<100) {
				tOffset = 5;
			} else if(i*fspace>99 && i*fspace<1000) {
				tOffset = 9;
			} else {
				tOffset = 12;
			}
			g2d.drawLine(timeLineAreaStart+(int)markerX, y, timeLineAreaStart+(int)markerX, y-4);
			g2d.drawString(""+(int)(i*fspace), timeLineAreaStart+(int)markerX-tOffset, y+9);
			markerX = markerX + (pixelPerFrame*fspace);
		}
		g2d.drawLine(timeLineAreaStart-10, y-8, prevWidth-10, y-8);
		g2d.drawLine(timeLineAreaStart-10, y+13, prevWidth-10, y+13);

        // Second Marker on the timeline
        g2d.setFont(new Font("SansSerif", Font.BOLD, 12));
        x = 5;
        y = 50;
        g2d.drawString("Seconds:", x, y+7);

        double sspace = 5;	// 5 for 180 second video is good
        double projectSeconds = (double)projectFrames/framesPerSecond;
        double pixelPerSecond = (double)timelineAreaWidth / projectSeconds;
        double secondMarkerCount = (double)projectFrames / sspace;
        double smarkerX = 0.0;
        g2d.setFont(new Font("SansSerif", Font.BOLD, 10));
        tOffset = 2;
        for(int i = 0;i<secondMarkerCount;i++) {
			if(i*sspace<10) {
				tOffset = 2;
			} else if(i*sspace>9 && i*sspace<100) {
				tOffset = 5;
			} else if(i*sspace>99 && i*sspace<1000) {
				tOffset = 9;
			} else {
				tOffset = 12;
			}
			g2d.drawLine(timeLineAreaStart+(int)smarkerX, y, timeLineAreaStart+(int)smarkerX, y-4);
			g2d.drawString(""+(int)(i*sspace), timeLineAreaStart+(int)smarkerX-tOffset, y+9);
			smarkerX = smarkerX + (pixelPerSecond*sspace);
		}
		g2d.drawLine(timeLineAreaStart-10, y+13, prevWidth-10, y+13);

        // Beats Marker on the timeline
        g2d.setFont(new Font("SansSerif", Font.BOLD, 12));
        x = 5;
        y = 72;
        g2d.drawString("Beats:", x, y+7);

        double bspace = 16;	// 16 for 180 second video with 140bpm is good
        double beatsPerSecond = (double)beatsPerMinute / 60.0;
        double beatsPerProject = projectSeconds * beatsPerSecond;	// 180 sekunden = 420 beats
        double pixelPerBeat = (double)timelineAreaWidth / beatsPerProject;
        double beatMarkerCount = (double)beatsPerProject / bspace;
        double bmarkerX = 0.0;
        g2d.setFont(new Font("SansSerif", Font.BOLD, 10));
        tOffset = 2;
        for(int i = 0;i<beatMarkerCount;i++) {
			if(i*bspace<10) {
				tOffset = 2;
			} else if(i*bspace>9 && i*bspace<100) {
				tOffset = 5;
			} else if(i*bspace>99 && i*bspace<1000) {
				tOffset = 9;
			} else {
				tOffset = 12;
			}
			g2d.drawLine(timeLineAreaStart+(int)bmarkerX, y, timeLineAreaStart+(int)bmarkerX, y-4);
			g2d.drawString(""+(int)(i*bspace), timeLineAreaStart+(int)bmarkerX-tOffset, y+9);
			bmarkerX = bmarkerX + (pixelPerBeat*bspace);
		}
		g2d.drawLine(timeLineAreaStart-10, y+13, prevWidth-10, y+13);

		updateTimeline();
	}

	public void updateTimeline() {
		int tlHeight = 10;
		int timeLineAreaStart = 80;
		int timelineAreaWidth = prevWidth - timeLineAreaStart-10;

		Graphics2D g2 = (Graphics2D)timeLinePanel.getGraphics();

		//Delete background
		g2.setColor(colAlias);
		g2.fillRect(timeLineAreaStart-5, tlHeight-5, timelineAreaWidth+10, 10);
        // Time Line
        g2.setColor(colBlack);
        g2.drawLine(timeLineAreaStart, tlHeight, prevWidth-10, tlHeight);
        // Calc X Position
        double pixelPerFrame = (double)timelineAreaWidth / (double)projectFrames;
        double currFrameXd = (double)currFrame * pixelPerFrame;
        int currFrameX = (int)currFrameXd;

		// Setup and draw triangle
		int xt[] = new int[3];	// coordinate arrays for a polygon object
		int yt[] = new int[3];
		xt[0]=timeLineAreaStart+currFrameX+3;	xt[1]=timeLineAreaStart+currFrameX-3;	xt[2]=timeLineAreaStart+currFrameX;
		yt[0]=tlHeight-3;	yt[1]=tlHeight-3;	yt[2]=tlHeight+3;
		int n = 3;
		Polygon p = new Polygon(xt, yt, n);  // This polygon represents a triangle
		g2.setColor(colYellow);
		g2.fillPolygon(p);     // Fills the triangle above
		g2.setColor(colBlack);
		g2.drawPolygon(p); 		// Framing
	}

	public void startPaintOriginal() {
		getCurrentDisplayGraphics().setColor(colWhite);
		getCurrentDisplayGraphics().fillRect(0, 0, prevWidth, prevHeight);
		paintOriginalImage();
	}

	public void paintOriginalImage() {
		BufferedImage tempBuffer;
		BufferedImage tmpBuffer;
		File imageFile;

		// Try to use the first still in a folder stills with the same name as the movie files
		imageFile		= new File(System.getProperty("user.dir")+File.separator+".."+File.separator+"stills"+File.separator+movieNames[0]+".jpg");

		if(imageFile.exists()) {	// if this exists ... use it
			printStatus("Image from folder stills loaded: "+movieNames[0]+".jpg");
			absoluteImagePath = System.getProperty("user.dir")+File.separator+".."+File.separator+"stills"+File.separator+movieNames[0]+".jpg";
		} else {
			imageFile		= new File(System.getProperty("user.dir")+File.separator+imageFileName);
			printStatus("Trying to use image with image name from header data: "+System.getProperty("user.dir")+File.separator+imageFileName);
			absoluteImagePath = System.getProperty("user.dir")+File.separator+imageFileName;
		}
		try {
			// *** load image
			tempBuffer				= ImageIO.read(imageFile);			// load original image to tempBuffer
			int imageWidth 				= tempBuffer.getWidth();			// Get original size informations
			int imageHeight 			= tempBuffer.getHeight();
			// *** scale image
			Object hint = RenderingHints.VALUE_INTERPOLATION_BILINEAR;
			int type = (tempBuffer.getTransparency() == Transparency.OPAQUE) ? BufferedImage.TYPE_INT_RGB : BufferedImage.TYPE_INT_ARGB;

			int w,h;
			w = imageWidth;		// use multi-step scale
			h = imageHeight;
			int scaleDirection = 0;
			if(prevWidth>w) scaleDirection = 1;
			do {
				if(scaleDirection == 0) {
					w >>= 1;
					h >>= 1;

					if(h < prevHeight || w < prevWidth) {
						w = prevWidth;
						h = prevHeight;
					}
				}
				if(scaleDirection == 1) {
					w = prevWidth;
					h = prevHeight;
				}

				tmpBuffer = new BufferedImage(w,h, type);
				Graphics2D g2 = tmpBuffer.createGraphics();
				g2.setRenderingHint(RenderingHints.KEY_INTERPOLATION, hint);
				g2.drawImage(tempBuffer,0,0,w,h,null);
				g2.dispose();
				tempBuffer = tmpBuffer;
			} while (w != prevWidth || h != prevHeight);
			// *** scale image

			// *** add (combine/blend) both images
			int ir = 0;	//image  rgb
			int ig = 0;
			int ib = 0;
			int pr = 0;	//pattern rgb
			int pg = 0;
			int pb = 0;
			int dr = 0;	//destination rgb
			int dg = 0;
			int db = 0;
			int drgb = 0;	// pattern rgb
			BufferedImage patternBuffer = getCurrentDisplayBuffer();	// get current displayed pattern
			for(int hs = 0;hs<prevHeight;hs++) {	//step through rows
				for(int ws = 0;ws<prevWidth;ws++) {	//step through columns
					Color ic = new Color(tmpBuffer.getRGB(ws, hs));
					ir = ic.getRed();
					ig = ic.getGreen();
					ib = ic.getBlue();
					Color pc = new Color(patternBuffer.getRGB(ws, hs));
					pr = pc.getRed();
					pg = pc.getGreen();
					pb = pc.getBlue();
					dr = (ir+pr)/2;	//destination rgb
					dg = (ig+pg)/2;
					db = (ib+pb)/2;
					drgb = (dr << 16) | (dg << 8) | db;
					tempBuffer.setRGB(ws, hs, drgb);
				}
			}
			// ** free ressources
			tmpBuffer.flush();
			tmpBuffer = null;
			// **** Paint image
			Graphics2D g2d = getTheGraphics();
			g2d.drawImage(tempBuffer,0,0,prevWidth,prevHeight,null);
			switchBuffer();
			tempBuffer.flush();
			tempBuffer = null;
		} catch( IOException ex ) {
			printStatus("Bildladefehler: "+System.getProperty("user.dir")+"/"+imageFileName);
		}
	}

	// *** more general helper methods
	public void printStatus(String theStat) {
		labStatus.setText("Status: "+theStat);
	}

	public void printTiming(int theCurrentFrame) {
		if(showTimingInfo) {
			System.out.println("CurrentFrame: "+theCurrentFrame+" - At Beat: "+(double)theCurrentFrame/framesPerBeat);
			System.out.println("       next Beat would be: "+(double)(theCurrentFrame+1)/framesPerBeat);
		}
	}

	public void printMaskCreateFeedback(int frNr, String fileName) {
		if(showMaskCreateFeedback) {
			int totalSeconde = (int)((double)frNr / 25.0)+1;
			int currMinute = (int)((double)totalSeconde/(double)60);
			int currSeconde = (int)((double)totalSeconde - ((double)currMinute*(double)60));
			System.out.println(makeLeadingZeros(currMinute, 2)+":"+makeLeadingZeros(currSeconde, 2)+" saved: "+fileName);
		}
	}

	public void printDataDebugRect(int frNr) {

			System.out.println("x1:"+kx1[frNr]);
			System.out.println("y1:"+ky1[frNr]);
			System.out.println("x2:"+kx2[frNr]);
			System.out.println("y2:"+ky2[frNr]);

			System.out.println("w1:"+wide1[frNr]);
			System.out.println("w2:"+wide2[frNr]);
			System.out.println("w3:"+wide3[frNr]);
			System.out.println("w4:"+wide4[frNr]);

			System.out.println("pulse:"+pulse[frNr]);

			System.out.println("rev:"+reverse[frNr]);
			System.out.println("dom:"+dominance[frNr]);
			System.out.println("domlay:"+domLayerNr[frNr]);
			System.out.println("domstr:"+domStrength[frNr]);
			System.out.println("domfall:"+domFalloff[frNr]);
			System.out.println("dx:"+directionX[frNr]);
			System.out.println("dy:"+directionY[frNr]);
	}
	public void selectAFile() {
		JFileChooser fileChooser;
		//fileChooser = new JFileChooser(lastSinglePath);
		fileChooser = new JFileChooser();
		fileChooser.addChoosableFileFilter(new javax.swing.filechooser.FileFilter() {
			public boolean accept(File file) {
				String filename = file.getName();
				return filename.toUpperCase().endsWith("JPG") ||
						filename.toUpperCase().endsWith("JEPG") ||
						filename.toUpperCase().endsWith("BMP") ||
						filename.toUpperCase().endsWith("GIF") ||
						filename.toUpperCase().endsWith("PNG") ||
						file.isDirectory();
			}
			public String getDescription() {
				return "*.jpg, *.bmp, *.gif, *.png";
			}
		});
		fileChooser.showOpenDialog(this);
		File selFile = fileChooser.getSelectedFile();
		if (selFile != null) {
			//selectedImageAbsolute = selFile.getAbsolutePath();
			printStatus("Filename: "+selFile.getAbsolutePath());
			//selectedFile = selFile;
			//sendNewFileName();
			//lastSinglePath = selectedImageAbsolute;
		} else {
			//Do nothing
		}
	}
	// *** Programm dependent code end
	// *** Now only GUI Control and creation code
	// ********************* Button Control **********************
	public void doButtPreviewAction() {
		buttPreview.setEnabled(false);
		buttPause.setEnabled(true);
		buttBack.setEnabled(false);
		buttCass.setEnabled(false);
		buttCreateMask.setEnabled(false);
		buttOriginal.setEnabled(false);
		playPreview();
	}

	public void doButtPauseAction() {
		buttPreview.setEnabled(true);
		buttPause.setEnabled(false);
		buttBack.setEnabled(true);
		buttCass.setEnabled(true);
		buttCreateMask.setEnabled(true);
		buttOriginal.setEnabled(true);
		timer.cancel();
		timer.purge();
	}

	public void doButtBackAction() {
		resetPreview();
	}

	public void doButtInfoAction() {
		if(infoOn) {
			infoOn = false;
		} else {
			infoOn = true;
		}
	}

	public void doButtCreateMaskAction() {
		createTheMasks();
	}

	public void doButtOriginalAction() {
		paintOriginalImage();
	}

	public void doButtExitAction() {
		quitProgramm();
	}
	public void doButtCassAction() { // buttCass
		CoordinateFrame cassWindow = new CoordinateFrame(absoluteImagePath);
		cassWindow.pack();
		cassWindow.centerWindow();
		cassWindow.setVisible(true);
	}
	// ********************* Action Listeners *****************
	public void setupListeningComponents() {
		// **JButton Listeners
		ActionListener buttPreviewListener = new ActionListener() {	// JButton
			public void actionPerformed(ActionEvent e) {
				doButtPreviewAction();
			}
		};
		buttPreview.addActionListener(buttPreviewListener);
		buttPreview.setToolTipText("Play Animation");

		// **JButton Listeners
		ActionListener buttPauseListener = new ActionListener() {	// JButton
			public void actionPerformed(ActionEvent e) {
				doButtPauseAction();
			}
		};
		buttPause.addActionListener(buttPauseListener);
		buttPause.setToolTipText("Pause Animation");

		// **JButton Listeners
		ActionListener buttBackListener = new ActionListener() {	// JButton
			public void actionPerformed(ActionEvent e) {
				doButtBackAction();
			}
		};
		buttBack.addActionListener(buttBackListener);
		buttBack.setToolTipText("Jump to Animation Start");

		// **JButton Listeners
		ActionListener buttInfoListener = new ActionListener() {	// JButton
			public void actionPerformed(ActionEvent e) {
				doButtInfoAction();
			}
		};
		buttInfo.addActionListener(buttInfoListener);
		buttInfo.setToolTipText("Display Timing Info in Image");

		// **JButton Listeners
		ActionListener buttCreateMaskListener = new ActionListener() {	// JButton
			public void actionPerformed(ActionEvent e) {
				doButtCreateMaskAction();
			}
		};
		buttCreateMask.addActionListener(buttCreateMaskListener);
		buttCreateMask.setToolTipText("Create the Mask File Sequences");

		// **JButton Listeners
		ActionListener buttOriginalListener = new ActionListener() {	// JButton
			public void actionPerformed(ActionEvent e) {
				doButtOriginalAction();
			}
		};
		buttOriginal.addActionListener(buttOriginalListener);
		buttOriginal.setToolTipText("Show Original on current Frame");

		// **JButton Listeners
		ActionListener buttExitListener = new ActionListener() {	// JButton
			public void actionPerformed(ActionEvent e) {
				doButtExitAction();
			}
		};
		buttExit.addActionListener(buttExitListener);
		buttExit.setToolTipText("Quit Programm");

		// **JButton Listeners
		ActionListener buttCassListener = new ActionListener() {	// JButton
			public void actionPerformed(ActionEvent e) {
				doButtCassAction();
			}
		};
		buttCass.addActionListener(buttCassListener);
		buttCass.setToolTipText("Coordinate Calculation Assistance");
	}

	public void setupComponentInstances() {
		timeDimPanel 	= new JPanel();
		previewPanel	= new JPanel();
		timeLinePanel	= new JPanel();
		labStatus		= new JLabel();
		buttPreview		= new JButton();
		buttPause		= new JButton();
		buttBack		= new JButton();
		buttInfo		= new JButton();
		buttCreateMask	= new JButton();
		buttOriginal	= new JButton();
		buttExit		= new JButton();
		buttCass		= new JButton();
	}

	public void setupComponentInitContent() {
		labStatus.setText("Status:");
		buttPreview.setText("> Play");
		buttPause.setText("|| Pause");
		buttBack.setText("|<< Back");
		buttInfo.setText("Info on/off");
		buttCreateMask.setText("Create Mask");
		buttOriginal.setText("Original");
		buttExit.setText("Exit");
		buttCass.setText("CAss");
	}

	public void setupComponentStyles() {
		//timeDimPanel.setBackground(new Color(255,255,255));
		//timeDimPanel.setForeground(Color.blue);
		//timeDimPanel.setBorder(new BevelBorder(BevelBorder.RAISED));
		//timeDimPanel.setBorder(new EtchedBorder());
		//timeDimPanel.setBorder(new TitledBorder("Name"));

		//previewPanel.setBackground(new Color(255,255,255));
		//previewPanel.setForeground(Color.blue);
		//previewPanel.setBorder(new BevelBorder(BevelBorder.RAISED));
		previewPanel.setBorder(new EtchedBorder());
		//previewPanel.setBorder(new TitledBorder("Name"));
		previewPanel.setPreferredSize(new Dimension(prevWidth,prevHeight));
		previewPanel.setSize(prevWidth, prevHeight);
		previewPanel.setMinimumSize(new Dimension(prevWidth, prevHeight));
		previewPanel.setMaximumSize(new Dimension(prevWidth, prevHeight));

		//timeLinePanel.setBackground(new Color(255,255,255));
		//timeLinePanel.setForeground(Color.blue);
		//timeLinePanel.setBorder(new BevelBorder(BevelBorder.RAISED));
		timeLinePanel.setBorder(new EtchedBorder());
		//timeLinePanel.setBorder(new TitledBorder("Name"));
		timeLinePanel.setPreferredSize(new Dimension(prevWidth,prevTimeHeight));
		timeLinePanel.setSize(prevWidth, prevTimeHeight);
		timeLinePanel.setMinimumSize(new Dimension(prevWidth, prevTimeHeight));
		timeLinePanel.setMaximumSize(new Dimension(prevWidth, prevTimeHeight));

		labStatus.setFont(labStatus.getFont().deriveFont(Font.PLAIN));
		//labStatus.setFont(labStatus.getFont().deriveFont(10f));
		labStatus.setHorizontalAlignment(JLabel.LEFT);
		//labStatus.setBackground(new Color(255,255,255));
		//labStatus.setForeground(Color.blue);

		//buttPreview.setFont(buttPreview.getFont().deriveFont(Font.PLAIN));
		//buttPreview.setFont(buttPreview.getFont().deriveFont(10f));
		//buttPreview.setBackground(new Color(255,255,255));
		//buttPreview.setForeground(Color.blue);
		//buttPreview.setEnabled(false);

		//buttPause.setFont(buttPause.getFont().deriveFont(Font.PLAIN));
		//buttPause.setFont(buttPause.getFont().deriveFont(10f));
		//buttPause.setBackground(new Color(255,255,255));
		//buttPause.setForeground(Color.blue);
		buttPause.setEnabled(false);

		//buttCreateMask.setFont(buttCreateMask.getFont().deriveFont(Font.PLAIN));
		//buttCreateMask.setFont(buttCreateMask.getFont().deriveFont(10f));
		buttCreateMask.setBackground(new Color(192,255,192));
		//buttCreateMask.setForeground(Color.blue);
		//buttCreateMask.setEnabled(false);

		//buttOriginal.setFont(buttOriginal.getFont().deriveFont(Font.PLAIN));
		//buttOriginal.setFont(buttOriginal.getFont().deriveFont(10f));
		//buttOriginal.setBackground(new Color(255,255,255));
		//buttOriginal.setForeground(Color.blue);
		//buttOriginal.setEnabled(false);

		//buttExit.setFont(buttExit.getFont().deriveFont(Font.PLAIN));
		//buttExit.setFont(buttExit.getFont().deriveFont(10f));
		buttExit.setBackground(new Color(255,192,192));
		//buttExit.setForeground(Color.blue);
		//buttExit.setEnabled(false);
	}

	public void makeLayout() {
		int guiRand = 20;
		gbConstraints = new GridBagConstraints();
		timeDimPanel.setLayout(new GridBagLayout());

		gbConstraints.gridx = 0;
		gbConstraints.gridy = 0;
		gbConstraints.gridwidth = 1;
		gbConstraints.gridheight = 1;
		gbConstraints.weightx = 0.0;
		gbConstraints.weighty = 0.0;
		gbConstraints.insets = new Insets(guiRand, guiRand, guiRand, guiRand);		// top, left, bottom, right
		gbConstraints.anchor = GridBagConstraints.LINE_START;
		gbConstraints.fill = GridBagConstraints.NONE;
		timeDimPanel.add(buttPreview, gbConstraints);

		gbConstraints.gridx = 1;
		gbConstraints.gridy = 0;
		gbConstraints.gridwidth = 1;
		gbConstraints.gridheight = 1;
		gbConstraints.weightx = 0.0;
		gbConstraints.weighty = 0.0;
		//gbConstraints.insets = new Insets(20, 20, 0, 20);		// top, left, bottom, right
		gbConstraints.anchor = GridBagConstraints.LINE_START;
		gbConstraints.fill = GridBagConstraints.NONE;
		timeDimPanel.add(buttPause, gbConstraints);

		gbConstraints.gridx = 2;
		gbConstraints.gridy = 0;
		gbConstraints.gridwidth = 1;
		gbConstraints.gridheight = 1;
		gbConstraints.weightx = 0.0;
		gbConstraints.weighty = 0.0;
		//gbConstraints.insets = new Insets(20, 20, 0, 20);		// top, left, bottom, right
		gbConstraints.anchor = GridBagConstraints.LINE_START;
		gbConstraints.fill = GridBagConstraints.NONE;
		timeDimPanel.add(buttBack, gbConstraints);

		gbConstraints.gridx = 3;
		gbConstraints.gridy = 0;
		gbConstraints.gridwidth = 1;
		gbConstraints.gridheight = 1;
		gbConstraints.weightx = 0.0;
		gbConstraints.weighty = 0.0;
		//gbConstraints.insets = new Insets(20, 20, 0, 20);		// top, left, bottom, right
		gbConstraints.anchor = GridBagConstraints.LINE_START;
		gbConstraints.fill = GridBagConstraints.NONE;
		timeDimPanel.add(buttInfo, gbConstraints);

		gbConstraints.gridx = 4;
		gbConstraints.gridy = 0;
		gbConstraints.gridwidth = 1;
		gbConstraints.gridheight = 1;
		gbConstraints.weightx = 0.0;
		gbConstraints.weighty = 0.0;
		//gbConstraints.insets = new Insets(20, 20, 0, 20);		// top, left, bottom, right
		gbConstraints.anchor = GridBagConstraints.LINE_START;
		gbConstraints.fill = GridBagConstraints.NONE;
		timeDimPanel.add(buttOriginal, gbConstraints);


		gbConstraints.gridx = 5;
		gbConstraints.gridy = 0;
		gbConstraints.gridwidth = 1;
		gbConstraints.gridheight = 1;
		gbConstraints.weightx = 0.0;
		gbConstraints.weighty = 0.0;
		//gbConstraints.insets = new Insets(20, 20, 0, 20);		// top, left, bottom, right
		gbConstraints.anchor = GridBagConstraints.LINE_START;
		gbConstraints.fill = GridBagConstraints.NONE;
		timeDimPanel.add(buttCass, gbConstraints);


		gbConstraints.gridx = 6;
		gbConstraints.gridy = 0;
		gbConstraints.gridwidth = 1;
		gbConstraints.gridheight = 1;
		gbConstraints.weightx = 0.0;
		gbConstraints.weighty = 0.0;
		//gbConstraints.insets = new Insets(20, 20, 0, 20);		// top, left, bottom, right
		gbConstraints.anchor = GridBagConstraints.LINE_START;
		gbConstraints.fill = GridBagConstraints.NONE;
		timeDimPanel.add(buttCreateMask, gbConstraints);

		gbConstraints.gridx = 7;
		gbConstraints.gridy = 0;
		gbConstraints.gridwidth = 1;
		gbConstraints.gridheight = 1;
		gbConstraints.weightx = 0.0;
		gbConstraints.weighty = 0.0;
		//gbConstraints.insets = new Insets(20, 20, 0, 20);		// top, left, bottom, right
		gbConstraints.anchor = GridBagConstraints.LINE_START;
		gbConstraints.fill = GridBagConstraints.NONE;
		timeDimPanel.add(buttExit, gbConstraints);

		gbConstraints.gridx = 0;
		gbConstraints.gridy = 1;
		gbConstraints.gridwidth = 8;
		gbConstraints.gridheight = 1;
		gbConstraints.weightx = 1.0;
		gbConstraints.weighty = 0.0;
		gbConstraints.insets = new Insets(guiRand, guiRand, 2, guiRand);		// top, left, bottom, right
		gbConstraints.anchor = GridBagConstraints.FIRST_LINE_START;
		gbConstraints.fill = GridBagConstraints.BOTH;
		timeDimPanel.add(previewPanel, gbConstraints);

		gbConstraints.gridx = 0;
		gbConstraints.gridy = 2;
		gbConstraints.gridwidth = 8;
		gbConstraints.gridheight = 1;
		gbConstraints.weightx = 1.0;
		gbConstraints.weighty = 0.0;
		gbConstraints.insets = new Insets(2, guiRand, guiRand, guiRand);		// top, left, bottom, right
		gbConstraints.anchor = GridBagConstraints.FIRST_LINE_START;
		gbConstraints.fill = GridBagConstraints.BOTH;
		timeDimPanel.add(timeLinePanel, gbConstraints);

		gbConstraints.gridx = 0;
		gbConstraints.gridy = 3;
		gbConstraints.gridwidth = 8;
		gbConstraints.gridheight = 1;
		gbConstraints.weightx = 1.0;
		gbConstraints.weighty = 0.0;
		gbConstraints.insets = new Insets(guiRand, guiRand, guiRand, guiRand);		// top, left, bottom, right
		gbConstraints.anchor = GridBagConstraints.FIRST_LINE_START;
		gbConstraints.fill = GridBagConstraints.HORIZONTAL;
		timeDimPanel.add(labStatus, gbConstraints);
	}

	// *******************************************************************************************
	// ******** Because these code portions are so long, they are located at the end of this file
	// ********
	// ********                             REAL TIME RENDERERS
	// ********
	// *******************************************************************************************

	public void shapeRendererStartAnimation(int frameNr) {
		// Special Real Time Renderer. Here the whole animation is not only rendered but also calculated
		// MUST HAVE MORE THAN 12 LAYERS!!! IF NOT THIS FIGURE NOT WORKS!!!!
		// *** shapeType = 5 !!!!
		int lx = 1;
		int ly = 1;
		int lw = 1;
		int lh = 1;

		int lxl = 1;		// for fade out
		int lyl = 1;
		int lwl = 1;
		int lhl = 1;

		// Delete Background, paint the red door black
		Graphics2D g2 = getTheGraphics();
		if(reverse[frameNr]) {	// select color to paint background. Start Ani is starting White
			g2.setColor(colBlack);
		} else {
			g2.setColor(colWhite);
		}
		g2.fillRect(0, 0, prevWidth, prevHeight);

		g2.translate(translX, translY);		// is calculated by int translX = (int)((originalWidth*scaleFactor)/2); and so on
//4
		// *** ***StartFigure - 2 Quarters, one layer fades over basic layer
		int fr = domLayerNr[frameNr]; // is the local frame nr, stored in this abused field
		double firstDuration = 4.0;	// *** duration in beats
		double firstDualMoveDuration		= (framesPerBeat * firstDuration);	// duration in frames
		double distanceStep = (originalWidth/2)/firstDualMoveDuration;
		if(fr > 0 && fr <= firstDualMoveDuration) {
			// we are inside first figure of this animation, do animation stuff
			// paint layers
			for(int lay = layerCount;lay>0;lay--) {
				lx = (int)(scaleFactor * ((originalWidth/2)-(fr*distanceStep))  );
				ly = (int)(scaleFactor * -1 * (originalHeight/2));
				lw = (int)(scaleFactor * originalWidth);
				lh = (int)(scaleFactor * originalHeight);
				if(reverse[frameNr]) {	// select color to paint
					g2.setColor(greys[layerCount-lay-1]);
				} else {
					g2.setColor(greys[lay-1]);
				}
				g2.fillRect(lx, ly, lw, lh);
			}
		}
//8
		// *** ***First Wait - 4 Quarters, last pattern stands still for view and wait
		double secondDuration = 8.0;	// *** duration in beats
		double secondWaitDuration		= (framesPerBeat * secondDuration)+1;	// duration in frames
		int startFr = (int)(firstDualMoveDuration);
		int endFr	= startFr+(int)secondWaitDuration;
		if(fr > startFr && fr <= endFr) {
			// do the wait
			// paint layers
			for(int lay = layerCount;lay>0;lay--) {
				lx = 0;
				ly = (int)(scaleFactor * -1 * (originalHeight/2));
				lw = (int)(scaleFactor * originalWidth);
				lh = (int)(scaleFactor * originalHeight);
				if(reverse[frameNr]) {	// select color to paint
					g2.setColor(greys[layerCount-lay-1]);
				} else {
					g2.setColor(greys[lay-1]);
				}
				g2.fillRect(lx, ly, lw, lh);
			}
		}
//4
		// *** ***  Second Figure - 2 Quarters, two layer move from the middle over basic layer
		double thirdDuration = 4.0;	// *** duration in beats
		double thirdMultiMoveDuration		= (framesPerBeat * thirdDuration);	// duration in frames
		distanceStep = ((originalWidth/4)/thirdMultiMoveDuration)/2;
		int localFr = fr-(int)(firstDualMoveDuration+secondWaitDuration);
		startFr = (int)(firstDualMoveDuration+secondWaitDuration)-1;
		endFr	= startFr+(int)thirdMultiMoveDuration;
		if(fr > startFr && fr <= endFr) {
			for(int lay = 0;lay<layerCount;lay++) {

				if(lay <= layerCount/3) {		// other layers
					lx = (int)(scaleFactor * 0-(localFr*distanceStep)  );
					ly = (int)(scaleFactor * -1 * (originalHeight/2));
					lw = (int)(scaleFactor * originalWidth);
					lh = (int)(scaleFactor * originalHeight);
				} else if(lay > layerCount/3 && lay <= (layerCount/3*2)) {		// other layers
					lx = 0;
					ly = (int)(scaleFactor * -1 * (originalHeight/2));
					lw = (int)(scaleFactor * originalWidth);
					lh = (int)(scaleFactor * originalHeight);
				} else {		// other layers
					lx = (int)(scaleFactor * 0+(localFr*distanceStep)  );
					ly = (int)(scaleFactor * -1 * (originalHeight/2));
					lw = (int)(scaleFactor * originalWidth);
					lh = (int)(scaleFactor * originalHeight);
				}
				if(reverse[frameNr]) {	// select color to paint
					g2.setColor(greys[lay-1]);
				} else {
					g2.setColor(greys[layerCount-lay-1]);
				}
				g2.fillRect(lx, ly, lw, lh);
			}
		}

//8
		// *** ***  Second Wait - 4 Quarters, last pattern stands still for view and wait
		double fourthDuration = 8.0;	// *** duration in beats
		double fourthWaitDuration		= (framesPerBeat * fourthDuration)+1;	// duration in frames
		localFr = fr-(int)(firstDualMoveDuration+secondWaitDuration+thirdMultiMoveDuration);
		startFr = (int)(firstDualMoveDuration+secondWaitDuration+thirdMultiMoveDuration)-2;
		endFr	= startFr+(int)fourthWaitDuration;
		if(fr > startFr && fr <= endFr) {
			for(int lay = 0;lay<layerCount;lay++) {
				if(lay <= layerCount/3) {		// other layers
					lx = (int)(scaleFactor * (0-(originalWidth/4))  );
					ly = (int)(scaleFactor * -1 * (originalHeight/2));
					lw = (int)(scaleFactor * originalWidth);
					lh = (int)(scaleFactor * originalHeight);
				} else if(lay > layerCount/3 && lay <= (layerCount/3*2)) {		// third quarter layers
					lx = 0;
					ly = (int)(scaleFactor * -1 * (originalHeight/2));
					lw = (int)(scaleFactor * originalWidth);
					lh = (int)(scaleFactor * originalHeight);
				} else {		// other layers
					lx = (int)(scaleFactor * (0+(originalWidth/4)));
					ly = (int)(scaleFactor * -1 * (originalHeight/2));
					lw = (int)(scaleFactor * originalWidth);
					lh = (int)(scaleFactor * originalHeight);
				}
				if(reverse[frameNr]) {	// select color to paint
					g2.setColor(greys[lay-1]);
				} else {
					g2.setColor(greys[layerCount-lay-1]);
				}
				g2.fillRect(lx, ly, lw, lh);
			}
		}
//4
		// *** ***  Third Figure - x Quarters, 8 layer move from on the screen
		double fifthDuration = 4.0;	// *** duration in beats
		double fifthMultiMoveDuration		= (framesPerBeat * fifthDuration);	// duration in frames
		distanceStep = ((originalWidth/12)/fifthMultiMoveDuration);
		localFr = fr-(int)(firstDualMoveDuration+secondWaitDuration+thirdMultiMoveDuration+fourthWaitDuration)+2;
		startFr = (int)(firstDualMoveDuration+secondWaitDuration+thirdMultiMoveDuration+fourthWaitDuration)-3;
		endFr	= startFr+(int)fifthMultiMoveDuration;
		if(fr > startFr && fr <= endFr) {
			for(int lay = 0;lay<layerCount;lay++) {
				if(lay <= layerCount/9*1) {		// other layers
					lx = (int)(scaleFactor * (0-(originalWidth/4)-((double)localFr*(double)distanceStep)));
					ly = (int)(scaleFactor * -1 * (originalHeight/2));
					lw = (int)(scaleFactor * originalWidth);
					lh = (int)(scaleFactor * originalHeight);
				} else if(lay > layerCount/9*1 && lay <= (layerCount/9*2)) {		// other layers
					lx = (int)(scaleFactor * (0-(originalWidth/4)));
					ly = (int)(scaleFactor * -1 * (originalHeight/2));
					lw = (int)(scaleFactor * originalWidth);
					lh = (int)(scaleFactor * originalHeight);
				} else if(lay > layerCount/9*2 && lay <= (layerCount/9*3)) {		// more layer
					lx = (int)(scaleFactor * (0-(originalWidth/4)+((double)localFr*(double)distanceStep)));
					ly = (int)(scaleFactor * -1 * (originalHeight/2));
					lw = (int)(scaleFactor * originalWidth);
					lh = (int)(scaleFactor * originalHeight);

				} else if(lay > layerCount/9*3 && lay <= (layerCount/9*4)) {		// more layer
					lx = (int)(scaleFactor * (0-((double)localFr*(double)distanceStep)));
					ly = (int)(scaleFactor * -1 * (originalHeight/2));
					lw = (int)(scaleFactor * originalWidth);
					lh = (int)(scaleFactor * originalHeight);
				} else if(lay > layerCount/9*4 && lay <= (layerCount/9*5)) {		// other layers
					lx = (int)(scaleFactor * 0);
					ly = (int)(scaleFactor * -1 * (originalHeight/2));
					lw = (int)(scaleFactor * originalWidth);
					lh = (int)(scaleFactor * originalHeight);
				} else if(lay > layerCount/9*5 && lay <= (layerCount/9*6)) {		// more layer
					lx = (int)(scaleFactor * (0+((double)localFr*(double)distanceStep)));
					ly = (int)(scaleFactor * -1 * (originalHeight/2));
					lw = (int)(scaleFactor * originalWidth);
					lh = (int)(scaleFactor * originalHeight);

				} else if(lay > layerCount/9*6 && lay <= (layerCount/9*7)) {		// more layer
					lx = (int)(scaleFactor * (0+(originalWidth/4)-((double)localFr*(double)distanceStep)));
					ly = (int)(scaleFactor * -1 * (originalHeight/2));
					lw = (int)(scaleFactor * originalWidth);
					lh = (int)(scaleFactor * originalHeight);
				} else if(lay > layerCount/9*7 && lay <= (layerCount/9*8)) {		// other layers
					lx = (int)(scaleFactor * (0+(originalWidth/4)));
					ly = (int)(scaleFactor * -1 * (originalHeight/2));
					lw = (int)(scaleFactor * originalWidth);
					lh = (int)(scaleFactor * originalHeight);

				} else {		// other layers
					lx = (int)(scaleFactor * (0+(originalWidth/4)+((double)localFr*(double)distanceStep)));
					ly = (int)(scaleFactor * -1 * (originalHeight/2));
					lw = (int)(scaleFactor * originalWidth);
					lh = (int)(scaleFactor * originalHeight);
				}
				if(reverse[frameNr]) {	// select color to paint
					g2.setColor(greys[lay-1]);
				} else {
					g2.setColor(greys[layerCount-lay-1]);
				}
				g2.fillRect(lx, ly, lw, lh);
			}
		}
//4
		// *** ***  Third Wait - x Quarters, 8 layer move from on the screen
		double sixthDuration = 4.0;	// *** duration in beats
		double sixthMultiMoveDuration		= (framesPerBeat * sixthDuration)+1;	// duration in frames
		distanceStep = ((originalWidth/12)/sixthMultiMoveDuration)/2;
		localFr = fr-(int)(firstDualMoveDuration+secondWaitDuration+thirdMultiMoveDuration+fourthWaitDuration+fifthMultiMoveDuration);
		startFr = (int)(firstDualMoveDuration+secondWaitDuration+thirdMultiMoveDuration+fourthWaitDuration+fifthMultiMoveDuration)-3;
		endFr	= startFr+(int)sixthMultiMoveDuration;
		if(fr > startFr && fr <= endFr) {

			for(int lay = 0;lay<layerCount;lay++) {
				if(lay <= layerCount/9*1) {		// other layers
					lx = (int)(scaleFactor * (0-(originalWidth/4)-(originalWidth/12)));
					ly = (int)(scaleFactor * -1 * (originalHeight/2));
					lw = (int)(scaleFactor * originalWidth);
					lh = (int)(scaleFactor * originalHeight);
				} else if(lay > layerCount/9*1 && lay <= (layerCount/9*2)) {		// other layers
					lx = (int)(scaleFactor * (0-(originalWidth/4)));
					ly = (int)(scaleFactor * -1 * (originalHeight/2));
					lw = (int)(scaleFactor * originalWidth);
					lh = (int)(scaleFactor * originalHeight);
				} else if(lay > layerCount/9*2 && lay <= (layerCount/9*3)) {		// more layer
					lx = (int)(scaleFactor * (0-(originalWidth/4)+(originalWidth/12)));
					ly = (int)(scaleFactor * -1 * (originalHeight/2));
					lw = (int)(scaleFactor * originalWidth);
					lh = (int)(scaleFactor * originalHeight);

				} else if(lay > layerCount/9*3 && lay <= (layerCount/9*4)) {		// more layer
					lx = (int)(scaleFactor * (0-(originalWidth/12)));
					ly = (int)(scaleFactor * -1 * (originalHeight/2));
					lw = (int)(scaleFactor * originalWidth);
					lh = (int)(scaleFactor * originalHeight);
				} else if(lay > layerCount/9*4 && lay <= (layerCount/9*5)) {		// other layers
					lx = (int)(scaleFactor * 0);
					ly = (int)(scaleFactor * -1 * (originalHeight/2));
					lw = (int)(scaleFactor * originalWidth);
					lh = (int)(scaleFactor * originalHeight);
				} else if(lay > layerCount/9*5 && lay <= (layerCount/9*6)) {		// more layer
					lx = (int)(scaleFactor * (0+(originalWidth/12)));
					ly = (int)(scaleFactor * -1 * (originalHeight/2));
					lw = (int)(scaleFactor * originalWidth);
					lh = (int)(scaleFactor * originalHeight);

				} else if(lay > layerCount/9*6 && lay <= (layerCount/9*7)) {		// more layer
					lx = (int)(scaleFactor * (0+(originalWidth/4)-(originalWidth/12)));
					ly = (int)(scaleFactor * -1 * (originalHeight/2));
					lw = (int)(scaleFactor * originalWidth);
					lh = (int)(scaleFactor * originalHeight);
				} else if(lay > layerCount/9*7 && lay <= (layerCount/9*8)) {		// other layers
					lx = (int)(scaleFactor * (0+(originalWidth/4)));
					ly = (int)(scaleFactor * -1 * (originalHeight/2));
					lw = (int)(scaleFactor * originalWidth);
					lh = (int)(scaleFactor * originalHeight);

				} else {		// other layers
					lx = (int)(scaleFactor * (0+(originalWidth/4)+(originalWidth/12)));
					ly = (int)(scaleFactor * -1 * (originalHeight/2));
					lw = (int)(scaleFactor * originalWidth);
					lh = (int)(scaleFactor * originalHeight);
				}
				if(reverse[frameNr]) {	// select color to paint
					g2.setColor(greys[lay-1]);
				} else {
					g2.setColor(greys[layerCount-lay-1]);
				}
				g2.fillRect(lx, ly, lw, lh);
			}
		}
//4
		// *** ***  Fourth Figure - Last spread on screen before disappearing
		double seventhDuration = 4.0;	// *** duration in beats
		double seventhMultiMoveDuration		= (framesPerBeat * seventhDuration);	// duration in frames
		distanceStep = ((originalWidth/12)/seventhMultiMoveDuration)/2;
		localFr = fr-(int)(firstDualMoveDuration+secondWaitDuration+thirdMultiMoveDuration+fourthWaitDuration+fifthMultiMoveDuration+sixthMultiMoveDuration)+3;
		startFr = (int)(firstDualMoveDuration+secondWaitDuration+thirdMultiMoveDuration+fourthWaitDuration+fifthMultiMoveDuration+sixthMultiMoveDuration)-4;
		endFr	= startFr+(int)seventhMultiMoveDuration;
		if(fr > startFr && fr <= endFr) {
			// Unfortunatly we first have to calculate all x start and end positions for this figure
			int lpos[] = new int[layerCount];	// last position in wait of figure three
			int epos[] = new int[layerCount];	// end position for this figure

			int hcor = 25; // Horizontal correction value to keep centered line (neeeds to be adjusted. 25 is for 19 layers
			for(int lay = 0;lay<layerCount;lay++) {
				if(lay <= layerCount/9*1) {		// other layers
					lpos[lay] = (int)(scaleFactor * (0-(originalWidth/4)-(originalWidth/12)));
					epos[lay] = hcor+(int)(scaleFactor * ((lay * originalWidth/layerCount)-(originalWidth/2)));
				} else if(lay > layerCount/9*1 && lay <= (layerCount/9*2)) {		// other layers
					lpos[lay] = (int)(scaleFactor * (0-(originalWidth/4)));
					epos[lay] = hcor+(int)(scaleFactor * ((lay * originalWidth/layerCount)-(originalWidth/2)));
				} else if(lay > layerCount/9*2 && lay <= (layerCount/9*3)) {		// more layer
					lpos[lay] = (int)(scaleFactor * (0-(originalWidth/4)+(originalWidth/12)));
					epos[lay] = hcor+(int)(scaleFactor * ((lay * originalWidth/layerCount)-(originalWidth/2)));

				} else if(lay > layerCount/9*3 && lay <= (layerCount/9*4)) {		// more layer
					lpos[lay] = (int)(scaleFactor * (0-(originalWidth/12)));
					epos[lay] = hcor+(int)(scaleFactor * ((lay * originalWidth/layerCount)-(originalWidth/2)));
				} else if(lay > layerCount/9*4 && lay <= (layerCount/9*5)) {		// other layers
					lpos[lay] = (int)(scaleFactor * 0);
					epos[lay] = hcor+(int)(scaleFactor * ((lay * originalWidth/layerCount)-(originalWidth/2)));
				} else if(lay > layerCount/9*5 && lay <= (layerCount/9*6)) {		// more layer
					lpos[lay] = (int)(scaleFactor * (0+(originalWidth/12)));
					epos[lay] = hcor+(int)(scaleFactor * ((lay * originalWidth/layerCount)-(originalWidth/2)));

				} else if(lay > layerCount/9*6 && lay <= (layerCount/9*7)) {		// more layer
					lpos[lay] = (int)(scaleFactor * (0+(originalWidth/4)-(originalWidth/12)));
					epos[lay] = hcor+(int)(scaleFactor * ((lay * originalWidth/layerCount)-(originalWidth/2)));
				} else if(lay > layerCount/9*7 && lay <= (layerCount/9*8)) {		// other layers
					lpos[lay] = (int)(scaleFactor * (0+(originalWidth/4)));
					epos[lay] = hcor+(int)(scaleFactor * ((lay * originalWidth/layerCount)-(originalWidth/2)));

				} else {		// other layers
					lpos[lay] = (int)(scaleFactor * (0+(originalWidth/4)+(originalWidth/12)));
					epos[lay] = hcor+(int)(scaleFactor * ((lay * originalWidth/layerCount)-(originalWidth/2)));
				}
			}

			for(int lay = 0;lay<layerCount;lay++) {	// Now paint all layers
				lx = (int)((double)lpos[lay]+(double)localFr*(((double)epos[lay]-(double)lpos[lay])/seventhMultiMoveDuration));
				ly = (int)(scaleFactor * -1 * (originalHeight/2));
				lw = (int)(scaleFactor * originalWidth);
				lh = (int)(scaleFactor * originalHeight);

				if(reverse[frameNr]) {	// select color to paint
					g2.setColor(greys[lay-1]);
				} else {
					g2.setColor(greys[layerCount-lay-1]);
				}
				g2.fillRect(lx, ly, lw, lh);
			}
		}
//8
		// *** ***  Fourth Wait - All Layers are spread wide and wait a bit now
		double eigthDuration = 8.0;	// *** duration in beats
		double eigthMultiMoveDuration		= (framesPerBeat * eigthDuration)+2;	// duration in frames
		localFr = fr-(int)(firstDualMoveDuration+secondWaitDuration+thirdMultiMoveDuration+fourthWaitDuration+fifthMultiMoveDuration+sixthMultiMoveDuration+seventhMultiMoveDuration);
		startFr = (int)(firstDualMoveDuration+secondWaitDuration+thirdMultiMoveDuration+fourthWaitDuration+fifthMultiMoveDuration+sixthMultiMoveDuration+seventhMultiMoveDuration)-5;
		endFr	= startFr+(int)eigthMultiMoveDuration+6; // 6 is a timing correction
		if(fr > startFr && fr <= endFr) {
			int epos[] = new int[layerCount];	// end position for this figure
			int hcor = 25; // Horizontal correction value to keep centered line (neeeds to be adjusted. 25 is for 19 layers
			for(int lay = 0;lay<layerCount;lay++) {
				epos[lay] = hcor+(int)(scaleFactor * ((lay * originalWidth/layerCount)-(originalWidth/2)));
			}

			for(int lay = 0;lay<layerCount;lay++) {	// Now paint all layers
				lx = epos[lay];
				ly = (int)(scaleFactor * -1 * (originalHeight/2));
				lw = (int)(scaleFactor * originalWidth);
				lh = (int)(scaleFactor * originalHeight);

				if(reverse[frameNr]) {	// select color to paint
					g2.setColor(greys[lay-1]);
				} else {
					g2.setColor(greys[layerCount-lay-1]);
				}
				g2.fillRect(lx, ly, lw, lh);
			}
		}
//4
		// *** ***  Start Animation - Fade Out Figure - All stripes go black
		// pre Figure Calculation (to detect if we are in Figure)
		double ninethDuration = 4.0;	// *** duration in beats
		double ninethMultiMoveDuration		= (framesPerBeat * ninethDuration)+1;	// duration in frames

		localFr = fr-(int)(firstDualMoveDuration+secondWaitDuration+thirdMultiMoveDuration+fourthWaitDuration+fifthMultiMoveDuration+sixthMultiMoveDuration+seventhMultiMoveDuration+eigthMultiMoveDuration)+4;
		startFr = (int)(firstDualMoveDuration+secondWaitDuration+thirdMultiMoveDuration+fourthWaitDuration+fifthMultiMoveDuration+sixthMultiMoveDuration+seventhMultiMoveDuration+eigthMultiMoveDuration)-5;
		endFr	= startFr+(int)ninethMultiMoveDuration; // + is a timing correction

		if(fr > startFr && fr <= endFr) {
			double moveDistance = originalWidth/ninethMultiMoveDuration;

			int epos[] = new int[layerCount];	// end position for this figure
			int hcor = 25; // Horizontal correction value to keep centered line (neeeds to be adjusted. 25 is for 19 layers
			for(int lay = 0;lay<layerCount;lay++) {
				epos[lay] = hcor+(int)(scaleFactor * ((lay * originalWidth/layerCount)-(originalWidth/2)));
			}

			for(int lay = 0;lay<layerCount;lay++) {	// Now paint all layers
				lx = epos[lay] - (int)(scaleFactor * (localFr * moveDistance));
				ly = (int)(scaleFactor * -1 * (originalHeight/2));
				lw = (int)(scaleFactor * originalWidth*2);
				lh = (int)(scaleFactor * originalHeight);

				if(reverse[frameNr]) {	// select color to paint
					g2.setColor(greys[lay-1]);
				} else {
					g2.setColor(greys[layerCount-lay-1]);
				}
				g2.fillRect(lx, ly, lw, lh);
			}
		}

		g2.translate(-translX, -translY);	// reset before paint other stuff that has origin at 0,0
		// do the rest
		if(infoOn) {
			paintTimingInfo(g2);
		}
		switchBuffer();
	}

	// *******************************************************************************************
	// ******** Because these code portions are so long, they are located at the end of this file
	// ********
	// ********                            REAL TIME MASK MAKERS
	// ********
	// *******************************************************************************************

	public void maskMakerStartAnimation(int frameNr) {
		// Special Real Time Mask maker. Here the whole animation is not only written to mask file but also calculated
		// MUST HAVE MORE THAN 12 LAYERS!!! IF NOT THIS FIGURE NOT WORKS!!!!
		// *** shapeType = 5 !!!!
		int lx = 1;
		int ly = 1;
		int lw = 1;
		int lh = 1;

		int lxl = 1;		// for fade out
		int lyl = 1;
		int lwl = 1;
		int lhl = 1;

		Graphics2D g2 = mask2D;
		g2.translate(maskTranslX, maskTranslY);		// is calculated by int translX = (int)((originalWidth*scaleFactor)/2); and so on
//4
		// *** ***StartFigure - 2 Quarters, one layer fades over basic layer
		int fr = domLayerNr[frameNr]; // is the local frame nr, stored in this abused field
		double firstDuration = 4.0;	// *** duration in beats
		double firstDualMoveDuration		= (framesPerBeat * firstDuration);	// duration in frames
		double distanceStep = (originalWidth/2)/firstDualMoveDuration;
		if(fr > 0 && fr <= firstDualMoveDuration) {
			// we are inside first figure of this animation, do animation stuff
			// paint layers
			for(int lay = layerCount-1;lay>0;lay--) {
				//** Pre block
				int theLay = lay-1;
				if(reverse[frameNr]) {
					theLay = (layerCount-1) - lay;
				}
				String currMaskFolder = fullWorkingPath+maskFolder+movieNames[theLay]+maskSubs+makeLeadingZeros(theLay+1,2)+File.separator;
				String currMaskFileName = movieNames[theLay]+maskSubs+makeLeadingZeros(theLay+1,2)+"-"+makeLeadingZeros(frameNr,5);

				// Delete Background, paint the red door black
				if(reverse[frameNr]) {	// select color to paint background. Start Ani is starting White
					g2.setColor(colBlack);
				} else {
					g2.setColor(colWhite);
				}
				g2.fillRect(originalWidth/2*-1, originalHeight/2*-1, originalWidth, originalHeight);
				// ** Pre Block end

				lx = (int)(((originalWidth/2)-(fr*distanceStep))  );
				ly = (int)(-1 * (originalHeight/2));
				lw = (int)(originalWidth);
				lh = (int)(originalHeight);

				if(reverse[frameNr]) {	// select color to paint
					g2.setColor(colWhite);
				} else {
					g2.setColor(colBlack);
				}
				g2.fillRect(lx, ly, lw, lh);

				// ** Past Paint 2 lines
				printMaskCreateFeedback(frameNr, currMaskFolder+currMaskFileName+"."+maskFileFormat.toLowerCase());
				saveMaskBufferAsImageFile(currMaskFolder+currMaskFileName);
			}
		}
//8
		// *** ***First Wait - 4 Quarters, last pattern stands still for view and wait
		double secondDuration = 8.0;	// *** duration in beats
		double secondWaitDuration		= (framesPerBeat * secondDuration)+1;	// duration in frames
		int startFr = (int)(firstDualMoveDuration);
		int endFr	= startFr+(int)secondWaitDuration;
		if(fr > startFr && fr <= endFr) {
			// do the wait
			// paint layers
			for(int lay = layerCount-1;lay>0;lay--) {
				//** Pre block
				int theLay = lay-1;
				if(reverse[frameNr]) {
					theLay = (layerCount-1) - lay;
				}
				String currMaskFolder = fullWorkingPath+maskFolder+movieNames[theLay]+maskSubs+makeLeadingZeros(theLay+1,2)+File.separator;
				String currMaskFileName = movieNames[theLay]+maskSubs+makeLeadingZeros(theLay+1,2)+"-"+makeLeadingZeros(frameNr,5);

				// Delete Background, paint the red door black
				if(reverse[frameNr]) {	// select color to paint background. Start Ani is starting White
					g2.setColor(colBlack);
				} else {
					g2.setColor(colWhite);
				}
				g2.fillRect(originalWidth/2*-1, originalHeight/2*-1, originalWidth, originalHeight);
				// ** Pre Block end

				lx = 0;
				ly = (int)(-1 * (originalHeight/2));
				lw = (int)(originalWidth);
				lh = (int)(originalHeight);

				// Color Select and FileWrite for Mask
				if(reverse[frameNr]) {	// select color to paint
					g2.setColor(colWhite);
				} else {
					g2.setColor(colBlack);
				}
				g2.fillRect(lx, ly, lw, lh);

				// ** Past Paint 2 lines
				printMaskCreateFeedback(frameNr, currMaskFolder+currMaskFileName+"."+maskFileFormat.toLowerCase());
				saveMaskBufferAsImageFile(currMaskFolder+currMaskFileName);
			}
		}
//4
		// *** ***  Second Figure - 2 Quarters, two layer move from the middle over basic layer
		double thirdDuration = 4.0;	// *** duration in beats
		double thirdMultiMoveDuration		= (framesPerBeat * thirdDuration);	// duration in frames
		distanceStep = ((originalWidth/4)/thirdMultiMoveDuration);
		int localFr = fr-(int)(firstDualMoveDuration+secondWaitDuration);
		startFr = (int)(firstDualMoveDuration+secondWaitDuration)-1;
		endFr	= startFr+(int)thirdMultiMoveDuration;

		if(fr > startFr && fr <= endFr) {
			for(int lay = 0;lay<layerCount-1;lay++) {
				//** Pre block
				int theLay = lay;
				if(reverse[frameNr]) {
					theLay = (layerCount) - lay;
				}
				String currMaskFolder = fullWorkingPath+maskFolder+movieNames[theLay]+maskSubs+makeLeadingZeros(theLay+1,2)+File.separator;
				String currMaskFileName = movieNames[theLay]+maskSubs+makeLeadingZeros(theLay+1,2)+"-"+makeLeadingZeros(frameNr,5);

				// Delete Background, paint the red door black
				if(reverse[frameNr]) {	// select color to paint background. Start Ani is starting White
					g2.setColor(colBlack);
				} else {
					g2.setColor(colWhite);
				}
				g2.fillRect(originalWidth/2*-1, originalHeight/2*-1, originalWidth, originalHeight);
				// ** Pre Block end

				if(lay <= layerCount/3) {		// other layers
					lx = (int)(0-(localFr*distanceStep)  );
					ly = (int)(-1 * (originalHeight/2));
					lw = (int)(originalWidth);
					lh = (int)(originalHeight);
				} else if(lay > layerCount/3 && lay <= (layerCount/3*2)) {		// other layers
					lx = 0;
					ly = (int)(-1 * (originalHeight/2));
					lw = (int)(originalWidth);
					lh = (int)(originalHeight);
				} else {		// other layers
					lx = (int)(0+(localFr*distanceStep)  );
					ly = (int)(-1 * (originalHeight/2));
					lw = (int)(originalWidth);
					lh = (int)(originalHeight);
				}

				// Color Select and FileWrite for Mask
				if(reverse[frameNr]) {	// select color to paint
					g2.setColor(colWhite);
				} else {
					g2.setColor(colBlack);
				}
				g2.fillRect(lx, ly, lw, lh);

				// ** Past Paint 2 lines
				printMaskCreateFeedback(frameNr, currMaskFolder+currMaskFileName+"."+maskFileFormat.toLowerCase());
				saveMaskBufferAsImageFile(currMaskFolder+currMaskFileName);
			}
		}

//8
		// *** ***  Second Wait - 4 Quarters, last pattern stands still for view and wait
		double fourthDuration = 8.0;	// *** duration in beats
		double fourthWaitDuration		= (framesPerBeat * fourthDuration)+1;	// duration in frames
		localFr = fr-(int)(firstDualMoveDuration+secondWaitDuration+thirdMultiMoveDuration);
		startFr = (int)(firstDualMoveDuration+secondWaitDuration+thirdMultiMoveDuration)-2;
		endFr	= startFr+(int)fourthWaitDuration;
		if(fr > startFr && fr <= endFr) {
			for(int lay = 0;lay<layerCount-1;lay++) {
				//** Pre block
				int theLay = lay;
				if(reverse[frameNr]) {
					theLay = (layerCount) - lay;
				}
				String currMaskFolder = fullWorkingPath+maskFolder+movieNames[theLay]+maskSubs+makeLeadingZeros(theLay+1,2)+File.separator;
				String currMaskFileName = movieNames[theLay]+maskSubs+makeLeadingZeros(theLay+1,2)+"-"+makeLeadingZeros(frameNr,5);

				// Delete Background, paint the red door black
				if(reverse[frameNr]) {	// select color to paint background. Start Ani is starting White
					g2.setColor(colBlack);
				} else {
					g2.setColor(colWhite);
				}
				g2.fillRect(originalWidth/2*-1, originalHeight/2*-1, originalWidth, originalHeight);
				// ** Pre Block end

				if(lay <= layerCount/3) {		// other layers
					lx = (int)((0-(originalWidth/4))  );
					ly = (int)(-1 * (originalHeight/2));
					lw = (int)(originalWidth);
					lh = (int)(originalHeight);
				} else if(lay > layerCount/3 && lay <= (layerCount/3*2)) {		// third quarter layers
					lx = 0;
					ly = (int)(-1 * (originalHeight/2));
					lw = (int)(originalWidth);
					lh = (int)(originalHeight);
				} else {		// other layers
					lx = (int)((0+(originalWidth/4)));
					ly = (int)(-1 * (originalHeight/2));
					lw = (int)(originalWidth);
					lh = (int)(originalHeight);
				}

				// Color Select and FileWrite for Mask
				if(reverse[frameNr]) {	// select color to paint
					g2.setColor(colWhite);
				} else {
					g2.setColor(colBlack);
				}
				g2.fillRect(lx, ly, lw, lh);

				// ** Past Paint 2 lines
				printMaskCreateFeedback(frameNr, currMaskFolder+currMaskFileName+"."+maskFileFormat.toLowerCase());
				saveMaskBufferAsImageFile(currMaskFolder+currMaskFileName);
			}
		}
//4
		// *** ***  Third Figure - x Quarters, 8 layer move from on the screen
		double fifthDuration = 4.0;	// *** duration in beats
		double fifthMultiMoveDuration		= (framesPerBeat * fifthDuration);	// duration in frames
		distanceStep = ((originalWidth/12)/fifthMultiMoveDuration);
		localFr = fr-(int)(firstDualMoveDuration+secondWaitDuration+thirdMultiMoveDuration+fourthWaitDuration)+2;
		startFr = (int)(firstDualMoveDuration+secondWaitDuration+thirdMultiMoveDuration+fourthWaitDuration)-3;
		endFr	= startFr+(int)fifthMultiMoveDuration;
		if(fr > startFr && fr <= endFr) {
			for(int lay = 0;lay<layerCount-1;lay++) {
				//** Pre block
				int theLay = lay;
				if(reverse[frameNr]) {
					theLay = (layerCount) - lay;
				}
				String currMaskFolder = fullWorkingPath+maskFolder+movieNames[theLay]+maskSubs+makeLeadingZeros(theLay+1,2)+File.separator;
				String currMaskFileName = movieNames[theLay]+maskSubs+makeLeadingZeros(theLay+1,2)+"-"+makeLeadingZeros(frameNr,5);

				// Delete Background, paint the red door black
				if(reverse[frameNr]) {	// select color to paint background. Start Ani is starting White
					g2.setColor(colBlack);
				} else {
					g2.setColor(colWhite);
				}
				g2.fillRect(originalWidth/2*-1, originalHeight/2*-1, originalWidth, originalHeight);
				// ** Pre Block end

				if(lay <= layerCount/9*1) {		// other layers
					lx = (int)((0-(originalWidth/4)-((double)localFr*(double)distanceStep)));
					ly = (int)(-1 * (originalHeight/2));
					lw = (int)(originalWidth);
					lh = (int)(originalHeight);
				} else if(lay > layerCount/9*1 && lay <= (layerCount/9*2)) {		// other layers
					lx = (int)((0-(originalWidth/4)));
					ly = (int)(-1 * (originalHeight/2));
					lw = (int)(originalWidth);
					lh = (int)(originalHeight);
				} else if(lay > layerCount/9*2 && lay <= (layerCount/9*3)) {		// more layer
					lx = (int)((0-(originalWidth/4)+((double)localFr*(double)distanceStep)));
					ly = (int)(-1 * (originalHeight/2));
					lw = (int)(originalWidth);
					lh = (int)(originalHeight);

				} else if(lay > layerCount/9*3 && lay <= (layerCount/9*4)) {		// more layer
					lx = (int)((0-((double)localFr*(double)distanceStep)));
					ly = (int)(-1 * (originalHeight/2));
					lw = (int)(originalWidth);
					lh = (int)(originalHeight);
				} else if(lay > layerCount/9*4 && lay <= (layerCount/9*5)) {		// other layers
					lx = (int)(0);
					ly = (int)(-1 * (originalHeight/2));
					lw = (int)(originalWidth);
					lh = (int)(originalHeight);
				} else if(lay > layerCount/9*5 && lay <= (layerCount/9*6)) {		// more layer
					lx = (int)((0+((double)localFr*(double)distanceStep)));
					ly = (int)(-1 * (originalHeight/2));
					lw = (int)(originalWidth);
					lh = (int)(originalHeight);

				} else if(lay > layerCount/9*6 && lay <= (layerCount/9*7)) {		// more layer
					lx = (int)((0+(originalWidth/4)-((double)localFr*(double)distanceStep)));
					ly = (int)(-1 * (originalHeight/2));
					lw = (int)(originalWidth);
					lh = (int)(originalHeight);
				} else if(lay > layerCount/9*7 && lay <= (layerCount/9*8)) {		// other layers
					lx = (int)((0+(originalWidth/4)));
					ly = (int)(-1 * (originalHeight/2));
					lw = (int)(originalWidth);
					lh = (int)(originalHeight);
				//} else if(lay > layerCount/9*3 && lay <= (layerCount/9*4)) {		// more layer

				} else {		// other layers
					lx = (int)((0+(originalWidth/4)+((double)localFr*(double)distanceStep)));
					ly = (int)(-1 * (originalHeight/2));
					lw = (int)(originalWidth);
					lh = (int)(originalHeight);
				}

				// Color Select and FileWrite for Mask
				if(reverse[frameNr]) {	// select color to paint
					g2.setColor(colWhite);
				} else {
					g2.setColor(colBlack);
				}
				g2.fillRect(lx, ly, lw, lh);

				// ** Past Paint 2 lines
				printMaskCreateFeedback(frameNr, currMaskFolder+currMaskFileName+"."+maskFileFormat.toLowerCase());
				saveMaskBufferAsImageFile(currMaskFolder+currMaskFileName);
			}

		}
//4
		// *** ***  Third Wait - x Quarters, 8 layer move from on the screen
		double sixthDuration = 4.0;	// *** duration in beats
		double sixthMultiMoveDuration		= (framesPerBeat * sixthDuration)+1;	// duration in frames
		distanceStep = ((originalWidth/12)/sixthMultiMoveDuration)/2;
		localFr = fr-(int)(firstDualMoveDuration+secondWaitDuration+thirdMultiMoveDuration+fourthWaitDuration+fifthMultiMoveDuration);
		startFr = (int)(firstDualMoveDuration+secondWaitDuration+thirdMultiMoveDuration+fourthWaitDuration+fifthMultiMoveDuration)-3;
		endFr	= startFr+(int)sixthMultiMoveDuration;
		if(fr > startFr && fr <= endFr) {

			for(int lay = 0;lay<layerCount-1;lay++) {
				//** Pre block
				int theLay = lay;
				if(reverse[frameNr]) {
					theLay = (layerCount) - lay;
				}
				String currMaskFolder = fullWorkingPath+maskFolder+movieNames[theLay]+maskSubs+makeLeadingZeros(theLay+1,2)+File.separator;
				String currMaskFileName = movieNames[theLay]+maskSubs+makeLeadingZeros(theLay+1,2)+"-"+makeLeadingZeros(frameNr,5);

				// Delete Background, paint the red door black
				if(reverse[frameNr]) {	// select color to paint background. Start Ani is starting White
					g2.setColor(colBlack);
				} else {
					g2.setColor(colWhite);
				}
				g2.fillRect(originalWidth/2*-1, originalHeight/2*-1, originalWidth, originalHeight);
				// ** Pre Block end

				if(lay <= layerCount/9*1) {		// other layers
					lx = (int)( (0-(originalWidth/4)-(originalWidth/12)));
					ly = (int)( -1 * (originalHeight/2));
					lw = (int)( originalWidth);
					lh = (int)( originalHeight);
				} else if(lay > layerCount/9*1 && lay <= (layerCount/9*2)) {		// other layers
					lx = (int)( (0-(originalWidth/4)));
					ly = (int)( -1 * (originalHeight/2));
					lw = (int)( originalWidth);
					lh = (int)( originalHeight);
				} else if(lay > layerCount/9*2 && lay <= (layerCount/9*3)) {		// more layer
					lx = (int)( (0-(originalWidth/4)+(originalWidth/12)));
					ly = (int)( -1 * (originalHeight/2));
					lw = (int)( originalWidth);
					lh = (int)( originalHeight);

				} else if(lay > layerCount/9*3 && lay <= (layerCount/9*4)) {		// more layer
					lx = (int)( (0-(originalWidth/12)));
					ly = (int)( -1 * (originalHeight/2));
					lw = (int)( originalWidth);
					lh = (int)( originalHeight);
				} else if(lay > layerCount/9*4 && lay <= (layerCount/9*5)) {		// other layers
					lx = (int)( 0);
					ly = (int)( -1 * (originalHeight/2));
					lw = (int)( originalWidth);
					lh = (int)( originalHeight);
				} else if(lay > layerCount/9*5 && lay <= (layerCount/9*6)) {		// more layer
					lx = (int)( (0+(originalWidth/12)));
					ly = (int)( -1 * (originalHeight/2));
					lw = (int)( originalWidth);
					lh = (int)( originalHeight);

				} else if(lay > layerCount/9*6 && lay <= (layerCount/9*7)) {		// more layer
					lx = (int)( (0+(originalWidth/4)-(originalWidth/12)));
					ly = (int)( -1 * (originalHeight/2));
					lw = (int)( originalWidth);
					lh = (int)( originalHeight);
				} else if(lay > layerCount/9*7 && lay <= (layerCount/9*8)) {		// other layers
					lx = (int)( (0+(originalWidth/4)));
					ly = (int)( -1 * (originalHeight/2));
					lw = (int)( originalWidth);
					lh = (int)( originalHeight);

				} else {		// other layers
					lx = (int)( (0+(originalWidth/4)+(originalWidth/12)));
					ly = (int)( -1 * (originalHeight/2));
					lw = (int)( originalWidth);
					lh = (int)( originalHeight);
				}

				// Color Select and FileWrite for Mask
				if(reverse[frameNr]) {	// select color to paint
					g2.setColor(colWhite);
				} else {
					g2.setColor(colBlack);
				}
				g2.fillRect(lx, ly, lw, lh);

				// ** Past Paint 2 lines
				printMaskCreateFeedback(frameNr, currMaskFolder+currMaskFileName+"."+maskFileFormat.toLowerCase());
				saveMaskBufferAsImageFile(currMaskFolder+currMaskFileName);
			}
		}
//4
		// *** ***  Fourth Figure - Last spread on screen before disappearing
		double seventhDuration = 4.0;	// *** duration in beats
		double seventhMultiMoveDuration		= (framesPerBeat * seventhDuration);	// duration in frames
		distanceStep = ((originalWidth/12)/seventhMultiMoveDuration)/2;
		localFr = fr-(int)(firstDualMoveDuration+secondWaitDuration+thirdMultiMoveDuration+fourthWaitDuration+fifthMultiMoveDuration+sixthMultiMoveDuration)+3;
		startFr = (int)(firstDualMoveDuration+secondWaitDuration+thirdMultiMoveDuration+fourthWaitDuration+fifthMultiMoveDuration+sixthMultiMoveDuration)-4;
		endFr	= startFr+(int)seventhMultiMoveDuration;
		if(fr > startFr && fr <= endFr) {
			// Unfortunatly we first have to calculate all x start and end positions for this figure
			int lpos[] = new int[layerCount];	// last position in wait of figure three
			int epos[] = new int[layerCount];	// end position for this figure

			int hcor = 25; // Horizontal correction value to keep centered line (neeeds to be adjusted. 25 is for 19 layers
			for(int lay = 0;lay<layerCount-1;lay++) {
				if(lay <= layerCount/9*1) {		// other layers
					lpos[lay] = (int)( (0-(originalWidth/4)-(originalWidth/12)));
					epos[lay] = hcor+(int)( ((lay * originalWidth/layerCount)-(originalWidth/2)));
				} else if(lay > layerCount/9*1 && lay <= (layerCount/9*2)) {		// other layers
					lpos[lay] = (int)( (0-(originalWidth/4)));
					epos[lay] = hcor+(int)( ((lay * originalWidth/layerCount)-(originalWidth/2)));
				} else if(lay > layerCount/9*2 && lay <= (layerCount/9*3)) {		// more layer
					lpos[lay] = (int)( (0-(originalWidth/4)+(originalWidth/12)));
					epos[lay] = hcor+(int)( ((lay * originalWidth/layerCount)-(originalWidth/2)));

				} else if(lay > layerCount/9*3 && lay <= (layerCount/9*4)) {		// more layer
					lpos[lay] = (int)( (0-(originalWidth/12)));
					epos[lay] = hcor+(int)( ((lay * originalWidth/layerCount)-(originalWidth/2)));
				} else if(lay > layerCount/9*4 && lay <= (layerCount/9*5)) {		// other layers
					lpos[lay] = (int)( 0);
					epos[lay] = hcor+(int)( ((lay * originalWidth/layerCount)-(originalWidth/2)));
				} else if(lay > layerCount/9*5 && lay <= (layerCount/9*6)) {		// more layer
					lpos[lay] = (int)( (0+(originalWidth/12)));
					epos[lay] = hcor+(int)( ((lay * originalWidth/layerCount)-(originalWidth/2)));

				} else if(lay > layerCount/9*6 && lay <= (layerCount/9*7)) {		// more layer
					lpos[lay] = (int)( (0+(originalWidth/4)-(originalWidth/12)));
					epos[lay] = hcor+(int)( ((lay * originalWidth/layerCount)-(originalWidth/2)));
				} else if(lay > layerCount/9*7 && lay <= (layerCount/9*8)) {		// other layers
					lpos[lay] = (int)( (0+(originalWidth/4)));
					epos[lay] = hcor+(int)( ((lay * originalWidth/layerCount)-(originalWidth/2)));

				} else {		// other layers
					lpos[lay] = (int)( (0+(originalWidth/4)+(originalWidth/12)));
					epos[lay] = hcor+(int)( ((lay * originalWidth/layerCount)-(originalWidth/2)));
				}
			}

			for(int lay = 0;lay<layerCount-1;lay++) {	// Now paint all layers
				//** Pre block
				int theLay = lay;
				if(reverse[frameNr]) {
					theLay = (layerCount) - lay;
				}
				String currMaskFolder = fullWorkingPath+maskFolder+movieNames[theLay]+maskSubs+makeLeadingZeros(theLay+1,2)+File.separator;
				String currMaskFileName = movieNames[theLay]+maskSubs+makeLeadingZeros(theLay+1,2)+"-"+makeLeadingZeros(frameNr,5);

				// Delete Background, paint the red door black
				if(reverse[frameNr]) {	// select color to paint background. Start Ani is starting White
					g2.setColor(colBlack);
				} else {
					g2.setColor(colWhite);
				}
				g2.fillRect(originalWidth/2*-1, originalHeight/2*-1, originalWidth, originalHeight);
				// ** Pre Block end

				lx = (int)((double)lpos[lay]+(double)localFr*(((double)epos[lay]-(double)lpos[lay])/seventhMultiMoveDuration));
				ly = (int)(-1 * (originalHeight/2));
				lw = (int)(originalWidth);
				lh = (int)(originalHeight);

				// Color Select and FileWrite for Mask
				if(reverse[frameNr]) {	// select color to paint
					g2.setColor(colWhite);
				} else {
					g2.setColor(colBlack);
				}
				g2.fillRect(lx, ly, lw, lh);

				// ** Past Paint 2 lines
				printMaskCreateFeedback(frameNr, currMaskFolder+currMaskFileName+"."+maskFileFormat.toLowerCase());
				saveMaskBufferAsImageFile(currMaskFolder+currMaskFileName);
			}


		}
//8
		// *** ***  Fourth Wait - All Layers are spread wide and wait a bit now
		double eigthDuration = 8.0;	// *** duration in beats
		double eigthMultiMoveDuration		= (framesPerBeat * eigthDuration)+2;	// duration in frames
		localFr = fr-(int)(firstDualMoveDuration+secondWaitDuration+thirdMultiMoveDuration+fourthWaitDuration+fifthMultiMoveDuration+sixthMultiMoveDuration+seventhMultiMoveDuration);
		startFr = (int)(firstDualMoveDuration+secondWaitDuration+thirdMultiMoveDuration+fourthWaitDuration+fifthMultiMoveDuration+sixthMultiMoveDuration+seventhMultiMoveDuration)-5;
		endFr	= startFr+(int)eigthMultiMoveDuration+6; // 6 is a timing correction
		if(fr > startFr && fr <= endFr) {
			int epos[] = new int[layerCount];	// end position for this figure
			int hcor = 25; // Horizontal correction value to keep centered line (neeeds to be adjusted. 25 is for 19 layers
			for(int lay = 0;lay<layerCount-1;lay++) {
				epos[lay] = hcor+(int)(((lay * originalWidth/layerCount)-(originalWidth/2)));
			}

			for(int lay = 0;lay<layerCount-1;lay++) {	// Now paint all layers
				//** Pre block
				int theLay = lay;
				if(reverse[frameNr]) {
					theLay = (layerCount) - lay;
				}
				String currMaskFolder = fullWorkingPath+maskFolder+movieNames[theLay]+maskSubs+makeLeadingZeros(theLay+1,2)+File.separator;
				String currMaskFileName = movieNames[theLay]+maskSubs+makeLeadingZeros(theLay+1,2)+"-"+makeLeadingZeros(frameNr,5);

				// Delete Background, paint the red door black
				if(reverse[frameNr]) {	// select color to paint background. Start Ani is starting White
					g2.setColor(colBlack);
				} else {
					g2.setColor(colWhite);
				}
				g2.fillRect(originalWidth/2*-1, originalHeight/2*-1, originalWidth, originalHeight);
				// ** Pre Block end

				lx = epos[lay];
				ly = (int)(-1 * (originalHeight/2));
				lw = (int)(originalWidth);
				lh = (int)(originalHeight);

				// Color Select and FileWrite for Mask
				if(reverse[frameNr]) {	// select color to paint
					g2.setColor(colWhite);
				} else {
					g2.setColor(colBlack);
				}
				g2.fillRect(lx, ly, lw, lh);

				// ** Past Paint 2 lines
				printMaskCreateFeedback(frameNr, currMaskFolder+currMaskFileName+"."+maskFileFormat.toLowerCase());
				saveMaskBufferAsImageFile(currMaskFolder+currMaskFileName);
			}
		}
//4
		// *** ***  Start Animation - Fade Out Figure - All stripes go black
		// pre Figure Calculation (to detect if we are in Figure)
		double ninethDuration = 4.0;	// *** duration in beats
		double ninethMultiMoveDuration		= (framesPerBeat * ninethDuration)+1;	// duration in frames

		localFr = fr-(int)(firstDualMoveDuration+secondWaitDuration+thirdMultiMoveDuration+fourthWaitDuration+fifthMultiMoveDuration+sixthMultiMoveDuration+seventhMultiMoveDuration+eigthMultiMoveDuration)+4;
		startFr = (int)(firstDualMoveDuration+secondWaitDuration+thirdMultiMoveDuration+fourthWaitDuration+fifthMultiMoveDuration+sixthMultiMoveDuration+seventhMultiMoveDuration+eigthMultiMoveDuration)-5;
		endFr	= startFr+(int)ninethMultiMoveDuration; // + is a timing correction

		if(fr > startFr && fr <= endFr) {
			double moveDistance = originalWidth/ninethMultiMoveDuration;

			int epos[] = new int[layerCount];	// end position for this figure
			int hcor = 25; // Horizontal correction value to keep centered line (neeeds to be adjusted. 25 is for 19 layers
			for(int lay = 0;lay<layerCount-1;lay++) {
				epos[lay] = hcor+(int)(((lay * originalWidth/layerCount)-(originalWidth/2)));
			}

			for(int lay = 0;lay<layerCount-1;lay++) {	// Now paint all layers
				//** Pre block
				int theLay = lay;
				if(reverse[frameNr]) {
					theLay = (layerCount) - lay;
				}
				String currMaskFolder = fullWorkingPath+maskFolder+movieNames[theLay]+maskSubs+makeLeadingZeros(theLay+1,2)+File.separator;
				String currMaskFileName = movieNames[theLay]+maskSubs+makeLeadingZeros(theLay+1,2)+"-"+makeLeadingZeros(frameNr,5);

				// Delete Background, paint the red door black
				if(reverse[frameNr]) {	// select color to paint background. Start Ani is starting White
					g2.setColor(colBlack);
				} else {
					g2.setColor(colWhite);
				}
				g2.fillRect(originalWidth/2*-1, originalHeight/2*-1, originalWidth, originalHeight);
				// ** Pre Block end

				lx = epos[lay] - (int)((localFr * moveDistance));
				ly = (int)(-1 * (originalHeight/2));
				lw = (int)(originalWidth*2);
				lh = (int)(originalHeight);

				// Color Select and FileWrite for Mask
				if(reverse[frameNr]) {	// select color to paint
					g2.setColor(colWhite);
				} else {
					g2.setColor(colBlack);
				}
				g2.fillRect(lx, ly, lw, lh);

				// ** Past Paint 2 lines
				printMaskCreateFeedback(frameNr, currMaskFolder+currMaskFileName+"."+maskFileFormat.toLowerCase());
				saveMaskBufferAsImageFile(currMaskFolder+currMaskFileName);
			}
		}
		g2.translate(-maskTranslX, -maskTranslY);	// reset before paint other stuff that has origin at 0,0
	}
	// ****** End of Mask Making Methods

}	// End of Class